/**
* Title: 		 Boudoir
* Description:	 Main API reference for the Boudoir library
* Copyright:     Copyright (c) 2013
* Company:       Morpheus Quality Technologies
* Creation date: 30-10-2013
* Last modified: 13-12-2013
* @author 		 Stefano Adriani
* @version 		 1.0
*/

package adriani.boudoir ;

import adriani.boudoir.db.Database ;
import adriani.boudoir.app.GeneralAnalyzer ;
import adriani.boudoir.logging.WSLogger ;
import adriani.boudoir.util.ErrorMessage ;
import adriani.boudoir.util.KeywordsManager ;

/** 
* To allow creation of multiple instances of this library in the same context (each one with his own configuration),
* every <code>Boudoir</code> instance creates his own <code>Settings</code> object, his <code>WSlogger</code>, 
* his  <code>Database</code> object etc.<br/>
* Such objects will be singleton only within the same specific library instance: this means that in the same context 
* there could be many different <code>Boudoir</code> instance, each owning 
* his own <code>Settings</code> object, his own <code>WSlogger</code> object etc.
*/	
public class Boudoir {
	
  	/**	Logical name (or string ID) identifying this <code>Boudoir</code> instance.							*/
	protected 		 			String				name ;
	
	/** Reference to the library's configuration (the <code>Settings</code> class).							*/	
	protected 		 			Settings 			settings ;
	
	/** Reference to the library's logging system (the <code>WSLogger</code> interface).					*/	
	protected 		 			WSLogger 			wsLogger ;
	
	/** Reference to the library's database layer (the <code>Database</code> class).						*/	
	protected 		 			Database 			database ;
	
	/** Reference to class managing resources storing list of custom keywords (optional).					*/	
	protected 		 			KeywordsManager 	keywordsManager ;
	
	/** Reference to the object collecting usage statistics (the <code>GeneralAnalyzer</code> interface).	*/	
	protected 		 			GeneralAnalyzer 	analyzer ;
	
	/** Reference to the object collecting usage statistics (the <code>GeneralAnalyzer</code> interface).	*/	
	protected 		 			ErrorMessage 		errorMessage ;

	/** Retrieving the logger instance to be used for this class.											*/
	protected  					WSLogger			logger 				= null ;
	
  	/** Beginning of each logging message produced by this class.											*/
	public static final			String				DEFAULT_NAME 		= "bd" ;
	
  	/** Beginning of each logging message produced by this class.											*/
	protected static 			String				leadingName 		= "Boudoir - " ;
	
	/**
	* Standard constructor, loading default configuration.
	*/
	public Boudoir() {
	
		this(DEFAULT_NAME) ;
	} 

	/**
	* Advanced constructor, loading custom configuration.<br/>
	* By setting a logical name to this Boudoir instance we allow multiple Boudoir objects
	* to be used in the same context. Each instance, along with his own objects and configuration,
	* will be completely identified by his logical name.<br/><br/>
	* 
	* Such approach allows also to put Boudoir's configuration files in the same directory,
	* since each Boudoir instance will access only files matching his logical name.<br/> 
	* For example, choosing the logical name "foo" implies that the <em>Boudoir-foo</em> 
	* instance will access following files:<br/>
	* 
	* <ul>
	* <li><code>foo_config.properties</code> as main configuration file</li>
	* <li><code>foo_data.properties</code> as data or keywords list file</li>
	* <li><code>foo_log4j.properties</code> as Log4j main configuration file</li>
	* <li><code>foo_log4j.xml</code> as Log4j alternate configuration file</li>
	* <li><code>foo_i18n.properties</code> as internationalization file (labels and message)</li>
	* </ul>
	* 
	* @param name the logic name to be used to identify this Boudoir instance
	*/
	public Boudoir(String name) {
	
		// Remark: order of these commands is very important: DO NOT CHANGE WITHOUT DEEP IMPACT ANALYSIS		
		this.name 				= name ;
		this.settings 			= new Settings(this) ;
		this.wsLogger 			= new WSLogger(this) ;
		this.database			= new Database(this) ;
		this.keywordsManager	= new KeywordsManager(this) ;
		this.errorMessage		= new ErrorMessage(this) ;
		init() ;
		
		if (settings.DEBUG) {
			StringBuffer message = new StringBuffer ("\n") ;
			message.append("***************************************************\n") ;
			message.append("Boudoir ID  : ") ;
			message.append(name) ;			
			message.append("\n") ;
			if (settings.VERBOSE) {
				message.append("Object ref  : ") ;
				message.append(this) ;
				message.append("\n") ;
			}
			message.append("***************************************************\n") ;
			System.out.println( message.toString() ) ;
		}
	}
	
// *********************************************************************************************************
// **********************************			INITIALIZING		****************************************
// *********************************************************************************************************
	
	/**
	* Initializes all instances wrapped by this class, such as: the <code>logger</code> object, 
	* the <code>serviceAnalyzer</code> object, etc...<br/><br/>
	* 
	* Notice that such instances are created (not initialized) within the constructor of this class.
	*/
	public void init() {
		
    	// 1. loading the proper logger for this class		
		this.logger = wsLogger.getLogger(Boudoir.class) ;
		
		// 2. Optional: preparing the analysis environment (if required)
		if (settings.ANALYZE) this.analyzer = new GeneralAnalyzer(this) ;
		
		// 3. Loading library keywords (if used)
		keywordsManager.loadKeywords(this.settings.KEYWORDS) ;
		//keywordsManager.printKeywords() ;
	}	
	
// *********************************************************************************************************
// **********************************			ACCESSORS			****************************************
// *********************************************************************************************************

	/**
	* Getter method.
	*/
	public  String getName() 							{			return this.name ;					}
	
	/**
	* Getter method.
	*/
	public  Settings getConfiguration() 				{			return this.settings ;				}

	/**
	* Getter method.
	*/
	public  WSLogger getLogger() 						{			return this.wsLogger ;				}
	
	/**
	* Getter method.
	*/
	public  Database getDatabase() 						{			return this.database ;				}
	
	/**
	* Getter method.
	*/
	public  KeywordsManager getKeywordsManager() 		{			return this.keywordsManager ;		}
	
	/**
	* Getter method.
	*/
	public  GeneralAnalyzer getAnalyzer() 				{			return this.analyzer ;				}
	
	/**
	* Getter method.
	*/
	public  ErrorMessage getMessages() 					{			return this.errorMessage ;			}
	
// *********************************************************************************************************
// **********************************			MUTATORS			****************************************
// *********************************************************************************************************
	
	/**
	* Setter method.
	*/
	public void setAnalyzer(GeneralAnalyzer analyzer)	{			this.analyzer = analyzer ;			}
	
} // End of class