/**
* Title: 		 Settings
* Description:	 Class wrapping the basic library configuration
* Copyright:     Copyright (c) 2013
* Company:       Morpheus Quality Technologies
* Creation date: 06-12-2013
* Last modified: 11-12-2013
* @author 		 Stefano Adriani
* @version 		 1.0
*/

package adriani.boudoir ;

import adriani.boudoir.util.Utilities ;
import adriani.boudoir.logging.WSLogger ;

import java.util.ResourceBundle ;

/**
* Class containing the "basic" values defining boolean flags, constants and logical names.
* By "basic" values we mean the general purposes properties, regarding the whole library
* and not a specific feature. For example, <em>Database</em> configuration or <em>Logger</em>
* are not loaded here, but are loaded from the specific class (e.g. the <code>Database</code> class
* or the <code>WSLogger</code> class).
*/
public class Settings {
	
	/** Reference to the top-level <code>Boudoir</code> instance.											*/
	public 		 				Boudoir 			boudoir ;
	
	/** Reference to the library's logging system (the <code>WSLogger</code> class).						*/	
	protected 		 			WSLogger 			logger ;
	
  	/** Beginning of each logging message produced by this class.											*/
	protected static 			String				leadingName 		= "Settings - " ;
	
// ******************************************************************************************************
// **************************				LIBRARY VERSION					*****************************
// ******************************************************************************************************
	
	/** 
	* Last released version of the library. This version number should be synchronized with:<br/><br/>  
	* 
	* - build.xml file (top)<br/> 
	* - change.log text file (top)<br/>
	*/
	public 		 	String		VERSION		= "demo" ;
	
	/** Name of this project, used to build at run-time all other logical names.						*/
	public 		 	String		NAME		= "boudoir" ;					// Default value
	
// ******************************************************************************************************
// **************************				LOGICAL FLAGS					*****************************
// ******************************************************************************************************
	
	/** If <code>true</code> debug messages are printed on standard output.								*/
	public 			boolean		DEBUG		= true ;						// Default value

	/** If <code>true</code> very detailed debug messages are printed on the standard output (verbose).	*/
	public 			boolean		VERBOSE		= true ;						// Default value

	/** If <code>true</code> statistical analysis data are collected during execution.					*/
	public 			boolean		ANALYZE		= false ;						// Default value

	/** If <code>true</code> the remote manager feature is enabled.										*/
	public 		 	boolean		MANAGER		= true ;						// Default value

	/** If <code>true</code> all SQL statements execution times will be logged into a custom log file.	*/
	public 		 	boolean		BENCH		= false ;						// Default value

// ******************************************************************************************************
// **************************				LOGGER SETTINGS					*****************************
// ******************************************************************************************************

  	/** If <code>true</code> the application is configured to use the Log4j mechanism.					*/
	public 		 	boolean 	LOG4J		= false ;						// Default value

  	/** If <code>true</code> the application is configured to use the <code>SystemLogger</code> logger.	*/
	public 			boolean 	JAVA		= true ;						// Default value

	/** 
	* Prefix name of the Log4j configuration file: not used at the moment (place-holder).
	* This field can be useful if the log configuration name must be qualified (eg. "com.company.bundle").
	* Usually we have two options:<br/>
	* 1. Keeping qualified name in the src folder, allowing the IDE to automatically find them after a build
	*    (not used for this project).<br/>
	* 2. Keeping qualified name in a resources folder: this imply updating the CLASSPATH in order to find
	*    the configuration files (used by this project).<br/>
	*/
	public 		 	String		LOG_PREFIX	= "" ;

	/** Specific name of the Log4j configuration file (extension is "properties").						*/
	public 			String		LOG_CONFIG	= "bd_log4j" ;

// ******************************************************************************************************
// **************************			SYSTEM LOGGER SETTINGS				*****************************
// ******************************************************************************************************

	/** 
	* Parent directory for the log files (used only by the <code>SystemLogger</code> class).
  	* This value is usually overwritten by the configuration file. 			
	*/
	public 			String		LOGDIR		= "logs" ;						// Default value

	/** 
	* Rotation of the logger file (minutes) (used only by the <code>SystemLogger</code> class).
  	* This value is usually overwritten by the configuration file. 		
	*/
	public  		int			ROTATE 		= 480 ;							// Default value
	
	/** Logic name of the logger object (used only by the <code>SystemLogger</code> class).				*/
	public  		String		LOGGER		= NAME + ".logger" ;			// Logic name

	/** Name of the performance analyzer object (used only by the <code>SystemLogger</code> class).		*/
	public 			String		BENCHMARK 	= NAME + ".benchmark" ;			// Logic name

	/** Name of the possible extensions to be parsed by the <code>SystemLogger</code> (if used).		*/
	public static	String[]	ARCHIVE 	= {".class", ".jar", ".aar", ".war"} ;

// ******************************************************************************************************
// **************************				PROPERTIES						*****************************
// ******************************************************************************************************

	/** File name of the main configuration Resource Bundle (extension is "properties").				*/
	public 		 	String		CONFIG		= "_config"  ;

	/** File name of the message i18n Resource Bundle (extension is "properties").						*/
	public 		 	String		MESSAGES	= "bd_i18n" ;

	/** File name of the logic keywords Resource Bundle (extension is "properties").					*/
	public 		 	String		KEYWORDS	= "bd_data" ;

// ******************************************************************************************************
// **************************			   REMOTE MANAGER					*****************************
// ******************************************************************************************************
	
  	/** 
  	* The hard-coded pattern used to recognize remote manager commands (e.g. "000000").				
  	* This value is usually overwritten by the configuration file. 
  	*/
	public 		 	String 		COMMAND		= "000000" ;						// Default value
	
// ******************************************************************************************************
// ******************************************************************************************************
// **************************			   INITIALIZING						*****************************
// ******************************************************************************************************
// ******************************************************************************************************
	
	/**
	* Standard constructor, bounding this class to a single <code>Boudoir</code> instance.
	* 
	* @param boudoir the <code>Boudoir</code> instance "owning" this object.
	*/
	public Settings(Boudoir boudoir) {
		
		this.boudoir = boudoir ;
		init() ;
	} 
	
	/**
	* Initializes all fields depending by the Boudoir's logicalname.
	* 
	*/
	public void init() {
		
		this.CONFIG 	= boudoir.getName() + "_config" ;
		this.MESSAGES 	= boudoir.getName() + "_i18n" ;
		this.KEYWORDS 	= boudoir.getName() + "_data" ;
		this.LOG_CONFIG	= boudoir.getName() + "_log4j" ;
		this.LOGGER		= boudoir.getName() + ".logger" ;
		this.BENCHMARK	= boudoir.getName() + ".benchmark" ;
		loadGeneralProperties() ;		
	}
	
// **************************************************************************************************************
// **********************************		LOAD RESOURCE BUNDLE		*****************************************
// **************************************************************************************************************
	
	/**
	* Load data from the resource bundle into this class fields.<br/><br/> 
	*  
	* Actual implementation of the get bundle mechanism is implemented by the delegated <code>getBundle()</code> 
	* method: in this way we allow all classes to load bundles using the same method.
	* 
	* @see GeneralApplication#loadBundle()
	*/
	public void loadGeneralProperties() {

		// 1. Loading data from the configuration.properties file
		ResourceBundle rb	= Utilities.loadBundle(this.CONFIG) ;
		if (rb != null) {
			
			// Parsing boolean properties
			String verbose		= rb.getString("library.verbose").trim() ;
			String debug 		= rb.getString("library.debug").trim() ;
			String analyze		= rb.getString("library.analyze").trim() ;
			String manager		= rb.getString("library.manager").trim() ;
			String benchmark	= rb.getString("library.benchmark").trim() ;
			this.VERBOSE		= verbose.equalsIgnoreCase("YES") ;
			this.DEBUG			= debug.equalsIgnoreCase("YES") ;
			this.ANALYZE		= analyze.equalsIgnoreCase("YES") ;
			this.MANAGER		= manager.equalsIgnoreCase("YES") ;
			this.BENCH			= benchmark.equalsIgnoreCase("YES") ;
			
			// Parsing string properties
			String prefix		= rb.getString("manager.prefix").trim() ;
			this.COMMAND		= prefix ;
			
			// Showing status
			if (this.VERBOSE) System.out.println(leadingName + "Loaded configuration file: " + this.CONFIG + ".properties") ;
			
		} else {
			System.err.println(leadingName + "Using default configuration properties") ;
		}

		// 2. Printing general configuration on STDOUT
		if (this.VERBOSE) {
			String blank = "" ;
			System.out.print(leadingName + "Using flags: " 			+ (this.DEBUG?"Debug ":blank)) ;
			System.out.print((this.VERBOSE?"Verbose ":blank) 		+ (this.ANALYZE?"Analyze ":blank)) ;
			System.out.print((this.MANAGER?"Manager ":blank)		+ (this.BENCH?"Benchmark ":blank)) ;
			System.out.println() ;
		}
	}
	
} // End of class