/**
* Title: 		 GeneralAnalyzer
* Description:	 General purpose parent class for the service analyzer feature
* Copyright:     Copyright (c) 2013
* Company:       Morpheus Quality Technologies
* Creation date: 06-12-2013
* Last modified: 13-12-2013
* @author 		 Stefano Adriani
* @version 		 1.0
*/

package adriani.boudoir.app ;

import adriani.boudoir.Boudoir;
import adriani.boudoir.Settings ;
import adriani.boudoir.logging.WSLogger;
import adriani.boudoir.util.ErrorMessage;

import java.util.* ;

/**
* This class is used to collect and analyze performance data. Average data are shown between square brackets.
* Following data are collected:<br/><br/>
*
* <ol>
* <li><strong>Total calls to the service:</strong> number of service invocations.</li>
* <li><strong>Calls still running:</strong> number of calls still hung up (included current one).</li>
* <li><strong>Calls interrupted by a macro error:</strong> number of times an Exception has been thrown at the macro level.</li>
* <li><strong>Calls to the service that provided no result:</strong> number of times the logic did not find any data.</li>
* <li><strong>Calls stopped because the database was missing:</strong> number of times a database has not been found.</li>
* <li><strong>Calls interrupted by a logic error:</strong> invocations not completed for some error at the logical level.</li>
* <li><strong>Calls completed with a validation error:</strong> invocations that produced invalid output data.</li>
* <li><strong>Calls completed returning valid data:</strong> invocations that provided a complete valid response.</li>
* <li><strong>Calls due to a command code (with valid data):</strong> invocations used to manage the service
* (which however produced valid output), without considering the current call.</li>
* </ol>
*
* The total number of calls (1st point) should be equal the sum of all other seven entries.<br/><br/>
*
* <ol>
* <li><strong>Execution completed at macro level:</strong> number of times the macro level completed execution.</li>
* <li><strong>Number of tries at the DAO level:</strong> number of times we executed at least one DAO call at the macro level.</li>
* <li><strong>Number of time measurements:</strong> number of times we counted an execution time.</li>
* <li><strong>Total execution time:</strong> total execution time measured by the chronwatch.</li>
* <li><strong>Total SQL queries:</strong> total number of queries executed by the DAO layer.</li>
*
 The first point is used to evaluate the average execution time. The second point is used to evaluate
* the average number of SQL queries for each service call (considering also the invocations stopped after just
* one query to the database).
*
* <ol>
* <li><strong>SQL queries without problems:</strong> number of successfully executed SQL queries.</li>
* <li><strong>SQL queries with no result:</strong> number of queries with no results (no record found).</li>
* <li><strong>SQL statement in error state:</strong> number of queries that generated an SQL error.</li>
* <li><strong>DAO errors:</strong> number of DAO execution not completed and not handled.</li>
* </ol>
*/
public class GeneralAnalyzer {
	
	/** Reference to the top-level <code>Boudoir</code> instance.											*/
	public 		 			Boudoir 			boudoir ;

	/**
	* The hash table tracking all the threads that are collecting results.<br>
	* By this mechanism, if two (or more) different threads invoke the <code>startCall()</code> method
	* before ending the call, we can acknowledge the identity of the caller, in order to match each
	* <code>endCall()</code> method with the proper starting time.<br/>
	* The hash table contains the pairs: caller thread (key) and starting time (value). Subclass should
	* use the <code>table.size()</code> method to acknowledge the number of calls still hanging up.
	*/
	protected 	 			Hashtable 	table ;

  	/** List of all measured execution times.																*/
	protected 				ArrayList 	list ;

  	/** Total number of times this service has been invoked.												*/
	protected 				int 		calls 				= 0 ;

  	/** Number of calls to the web service ended without problems.											*/
	protected 				int 		no_error 			= 0 ;

	/** Number of errors at the main <code>ServiceLogic</code> level.										*/
	protected 				int 		logic_error 		= 0 ;

	/** Number of errors at the intermediate <code>ServiceMacro</code> level.								*/
	protected 				int 		macro_error 		= 0 ;

  	/** Number of errors due to an <code>ADBBean</code> exception.											*/
	protected 				int 		validation_error 	= 0 ;

  	/** Number of times the web service logic found no data at all.											*/
	protected 				int 		no_result 			= 0 ;

  	/** Number of calls to the web service failed because there was no database connection.					*/
	protected 				int 		no_db 				= 0 ;

  	/** Number of SQL statement executed without problems.													*/
	protected 				int 		sql_query 			= 0 ;

  	/** Number of SQL statement that generated an SQL error.												*/
	protected 				int 		sql_error 			= 0 ;

  	/** Number of SQL SELECT query performed with result "NO RECORD".										*/
	protected 				int 		no_record 			= 0 ;

  	/** Number of errors not properly handlef during execution of a DBA method.								*/
	protected 				int 		dao_error 			= 0 ;

  	/**
	* Number calls not executed because the input parameters was a command for the <code>manageService()</code>
	* method of the <code>ServiceLogic()</code> class.
	*/
	protected 				int 		command 			= 0 ;

  	/** The timestamp of the startup time of the web service (i.e. the time of the first call).				*/
	protected 				Date 		startup ;
	
  	/** The day of the month of the last automatic dump.													*/
	protected 				int 		last_dump ;

  	/** Beginning of each logging message produced by this class.											*/
	protected static 		String 		leadingName 		= "GeneralAnalyzer - " ;

	/** Retrieving the logger instance to be used for this class (static).									*/
	protected static 		WSLogger	logger 				= null ;

	/**
	* Standard constructor. This constructor is called implicitly by any "empty constructor" of any subclass
	* of this <code>GeneralAnalyzer</code> class (i.e. without the need to write <code>super()</code>).<br/>
	* Hence here we instantiate all the general purposes object, such as:<br/>
	* - the <code>Hashtable</code> that will store the threads references.<br/>
	* - the timestamp attribute storing the startup time of the service (i.e. the first call time).<br/>
	*/
	public GeneralAnalyzer() {

		this.startup			= new Date() ;
		this.table 				= new Hashtable() ;
		Calendar 	calendar 	= Calendar.getInstance() ;
		this.last_dump			= calendar.get(Calendar.DAY_OF_MONTH) ;
		//System.out.println(leadingName + "GeneralAnalyzer startup : " + startup + " (" + this.last_dump + ")") ;
	}
	
	/**
	* Standard constructor, bounding this class to a single <code>Boudoir</code> instance.
	* 
	* @param boudoir the <code>Boudoir</code> instance "owning" this object
    */
    public GeneralAnalyzer(Boudoir boudoir) {

    	this() ;
    	this.boudoir 	= boudoir ;
		this.list 		= new ArrayList(100) ;    	
    	this.logger 	= boudoir.getLogger().getLogger(GeneralAnalyzer.class) ;
    }	

	/**
	* Collect a generic single <code>ServiceDAO</code> event and increment the related counter, using the integer
	* code values defined in the <code>ErrorMessage</code> class.
	*
	* @param code the integer code describing the status of the DBA after execution
	* @see ErrorMessage
	*/
	public void collect(int code) {

		switch(code) {
			case ErrorMessage.DAO 			: 	addDAOError() ;		break ;
			case ErrorMessage.SQL 			: 	addSQLError() ;		break ;
			case ErrorMessage.NO_RECORD		: 	addNoRecord() ;		break ;
			case ErrorMessage.OK			: 	addSQLQuery() ;		break ;
		}
	}

// ***********************************************************************************************************
// **********************************			ANALYSIS			******************************************
// ***********************************************************************************************************

	/**
	* Produce a <code>StringBuffer</code> object containing all the analysis data already formatted and
	* ready for printing.<br/>
	*
	* @return a string buffer object containing the formatted results of the analysis
	*/
	public StringBuffer getResults() {

		// Counting the number of queries executed
		// Notice that the no_customer case is already counted by the no_record case
		long total = sql_query + no_record + sql_error + dao_error ;

		// Getting the total execution time of the calls completed successfully
		int sum = 0 ;
		Object[] times = this.list.toArray() ;
		for (int i = 0 ; i < times.length ; i++) sum += ((Integer) times[i]).intValue() ;

		// Getting the number of calls without errors
		int complete = calls - (no_db + macro_error ) ;

		// Getting the number of times we begun accessing the macro level
		int macro = calls - no_db ;

		// Formatting analysis results
		StringBuffer report = new StringBuffer() ;
		report.append("Total calls to the service    : " + calls		+ "\n") ;
		report.append("Number of calls still hung up : " + table.size()	+ "\n") ;
		report.append("Time of first service call    : " + startup		+ "\n") ;
		report.append("\n") ;
		report.append("Calls interrupted by a macro error                   : " + 	macro_error 		+ "\n") ;
		report.append("Calls to the service that provided no result         : " + 	no_result 			+ "\n") ;
		report.append("Calls stopped because the database was missing       : " + 	no_db 				+ "\n") ;
		report.append("Calls interrupted by a logic error                   : " + 	logic_error 		+ "\n") ;
		report.append("Calls completed with a validation error              : " + 	validation_error	+ "\n") ;
		report.append("Calls completed returning valid data                 : " + 	no_error 			+ "\n") ;
		report.append("Calls due to a command code (excluding current)      : " + 	command	 			+ "\n") ;
		report.append("\n") ;
		if (complete > 0) {
			report.append("Execution completed at macro level  : " + complete		+ "\n") ;
			report.append("Number of tries at the macro level  : " + macro 			+ "\n") ;
			report.append("Number of time measurements         : " + times.length 	+ "\n") ;
			report.append("Total execution time                : " + sum 			+ " ms [" + (sum/complete) + " ms]\n") ;
			report.append("Total SQL queries                   : " + total 			+ " [" + (total/macro) + "]\n") ;
			report.append("\n") ;
		}
		if (total > 0) {
			report.append("SQL queries without problems        : " + sql_query + " [" + (100*sql_query/total) + "%]\n") ;
			report.append("SQL queries with no records         : " + no_record + " [" + (100*no_record/total) + "%]\n") ;
			report.append("SQL statement in error state        : " + sql_error + " [" + (100*sql_error/total) + "%]\n") ;
			report.append("DAO unexpected errors (not handled) : " + dao_error + "\n") ;
			report.append("\n") ;
		}
		return report ;
	}

	/**
	* Print the analysis results produced by the <code>getResults()</code> method on the standard output.
	*
	* @see #getResults()
	*/
	public void printResults() {

		System.out.println() ;
		System.out.println(leadingName + "ANALYSIS RESULT:\n" + getResults()) ;
	}

	/**
	* Log the analysis results produced by the <code>getResults()</code> method in the LOG file.
	*
	* @see #getResults()
	*/
	public void logResults() {

			logger.logInfo(leadingName + "ANALYSIS RESULT:\n\n" + getResults()) ;
	}

// ***********************************************************************************************************
// **********************************			CHRONWATCH			******************************************
// ***********************************************************************************************************

	/**
	* Start the chronwatch at the beginning of one call service.
	* This method should be called only once by the main <code>ServiceLogic</code>.
	*/
	public void startCall() {

		// Incrementing calls counter (the thread doesn't matter)
		this.calls++ ;
		// Storing starting time for current thread
		Long start = new Long( System.currentTimeMillis() ) ;
		Thread thread = Thread.currentThread() ;
		this.table.put(thread, start) ;
	}

	/**
	* Stop the chronwatch after the end of one call service, even if the call generated an error.
	* If the chronwatch is stopped after an execution error, the error code is identified using the static
	* values defined in the <code>ErrorMessage</code> class.<br/>
	* This method should be called only once at the end of the service call. This means that this method
	* could be called by the <code>ServiceLogic</code> class or by a <code>ServiceMacro</code>,
	* depending by the particular "logic branch" executed at run-time.<br/><br/>
	*
	* <strong>Remark:</strong> implementation of this method is left to the child class because the range of
	* Possible error codes depends by the web service feature.
	*
	* @param code the integer code describing the event or error
	* @see ErrorMessage
	*/
	public void endCall(int code) {

		// 1. Incrementing counters
		switch(code) {
			case ErrorMessage.MACRO				: addMacroError() ;			break ;
			case ErrorMessage.LOGIC				: addLogicError() ;			break ;
			case ErrorMessage.COMMAND			: addCommand() ;			break ;
			case ErrorMessage.NO_DB 			: addNoDatabase() ;			break ;
			case ErrorMessage.VALIDATION		: addValidationError() ;	break ;
			case ErrorMessage.OK				: addNoError() ;			break ;
			case ErrorMessage.NO_RESULT			: addNoResult() ;			break ;
		}

		// 2. Finding the starting time for the current thread
		Thread current 	= Thread.currentThread() ;
		Long start		= (Long) this.table.get(current) ;

		// 3.1 Stopping chronwatch only if actually started
		if (start != null) {
			// Evaluating elapsed time
			int time = (int) (System.currentTimeMillis() - start.longValue()) ;
			// Adding the time measure only if the call was successfully
			if (code == ErrorMessage.OK) this.list.add(time) ;
			if (boudoir.getConfiguration().VERBOSE) {
				System.out.println(leadingName + "Call execution time  : " + time + " ms") ;
				System.out.println(leadingName + "Call execution thread: " + current) ;
			}
			// Remembering that we stopped the chronwatch
			this.table.remove(current) ;
		// 3.2 If the thread is not in the hash table, something went wrong
		} else {
			logger.logWarning(leadingName + "No starting time for thread " + current) ;
		}
		
		// 4. Checking if we must automatically dump all statistics (daily dump)
		Calendar 	calendar 	= Calendar.getInstance() ;
		int 		day 		= calendar.get(Calendar.DAY_OF_MONTH) ;
		if (day != this.last_dump) {
			this.logResults() ;
			this.last_dump = day ;
			if (boudoir.getConfiguration().VERBOSE) System.out.println(leadingName + "Set new last dump day : " + day) ;
		}
	}

// **********************************************************************************************************
// **********************************			MUTATORS		*********************************************
// ***********************************************************************************************************

	/**
	* Mutator method.
	*/
	public void addNoError() 				{	this.no_error++ ;											}

	/**
	* Mutator method.
	*/
	public void addCommand() 				{	this.command++ ;											}

	/**
	* Mutator method.
	*/
	public void addLogicError() 			{	this.logic_error++ ;										}

	/**
	* Mutator method.
	*/
	public void addValidationError() 		{	this.validation_error++ ;									}

	/**
	* Mutator method.
	*/
	public void addNoResult() 				{	this.no_result++ ;											}

	/**
	* Mutator method.
	*/
	public void addNoDatabase()				{	this.no_db++ ;												}

	/**
	* Mutator method.
	*/
	public void addMacroError() 			{	this.macro_error++ ;										}

	/**
	* Mutator method.
	*/
	public void addSQLQuery() 				{	this.sql_query++ ;											}

	/**
	* Mutator method.
	*/
	public void addNoRecord() 				{	this.no_record++ ;											}

	/**
	* Mutator method.
	*/
	public void addSQLError() 				{	this.sql_error++ ;											}

	/**
	* Mutator method.
	*/
	public void addDAOError() 				{	this.dao_error++ ;											}
	
	/**
	* Mutator method.
	*/
	public void setLastDump(int value)		{	this.last_dump = value ;									}

// **********************************************************************************************************
// **********************************			ACCESSORS			*****************************************
// ***********************************************************************************************************

	/**
	* Mutator method.
	*/
	public int getNoError() 				{	return this.no_error ;										}

	/**
	* Accessor method.
	*/
	public int getCommand() 				{	return this.command ;										}

	/**
	* Accessor method.
	*/
	public int getLogicError() 				{	return this.logic_error ;									}

	/**
	* Accessor method.
	*/
	public int getValidationError()			{	return this.validation_error ;								}

	/**
	* Accessor method.
	*/
	public int getNoResult()				{	return this.no_result ;										}

	/**
	* Mutator method.
	*/
	public int getNoDatabase() 				{	return this.no_db ;											}

	/**
	* Accessor method.
	*/
	public int getMacroError() 				{	return this.macro_error ;									}

	/**
	* Mutator method.
	*/
	public int getSQLQuery() 				{	return this.sql_query ;										}

	/**
	* Accessor method.
	*/
	public int getNoRecord() 				{	return this.no_record ;										}

	/**
	* Accessor method.
	*/
	public int getSQLError() 				{	return this.sql_error ;										}

	/**
	* Accessor method.
	*/
	public int getDAOError() 				{	return this.dao_error ;										}

	/**
	* Accessor method.
	*/
	public int getCalls() 					{	return this.calls ;											}

	/**
	* Accessor method.
	*/
	public Date getStartup() 				{	return this.startup ;										}
	
	/**
	* Accessor method.
	*/
	public int getLastDump() 				{	return this.last_dump ;										}
	
} // End of class