/**
* Title: 		 GeneralApplication (legacy name: GeneralService)
* Description:	 The parent class for all classes implementing a service business logic
* Copyright:     Copyright (c) 2013
* Company:       Morpheus Quality Technologies
* Creation date: 29-10-2013
* Last modified: 13-12-2013 
* @author 		 Stefano Adriani
* @version 		 1.0
*/

package adriani.boudoir.app ;

import adriani.boudoir.* ;
import adriani.boudoir.logging.LoggerManager;
import adriani.boudoir.logging.WSLogger;
import adriani.boudoir.util.* ;

import java.util.* ;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
* The parent class for all the classes implementing the service business logic: <code>ServiceLogic</code>
* and <code>ServiceMacro</code>. At the moment this class offers following features for a general
* web service class:<br/><br/>
*
* <ol>
* <li>Loads the general configuration from the properties file.</li>
* <li>Static configuration (on start time) of the required logger objects.</li>
* <li>Load keywords from the corresponding properties file (if required).</li>
* <li>Implementation of the remote command manager feature.</li>
* </ol>
*/
public class GeneralApplication {
	
	/** Reference to the top-level <code>Boudoir</code> instance.											*/
	public		 			Boudoir 			boudoir ;
	
	/** Static regex pattern used to check if input argument is a remote manager command.					*/
	public 					Pattern 		PREFIX ;

	/** Static final <code>String</code> values stored here to improve performances (empty string: "").		*/
	public static final 	String				EMPTY				= "" ;
	/** Static final <code>String</code> values stored here to improve performances (blank string: " ").	*/
	public static final 	String				BLANK				= " " ;
	/** Static final <code>String</code> values stored here to improve performances (string "0").			*/
	public static final 	StringBuffer		ZERO				= new StringBuffer("0") ;
	
	/** Retrieving the logger instance to be used for this class.											*/
	protected 				WSLogger			logger 				= null ;

  	/** Beginning of each logging message produced by this class.											*/
	protected static 		String 				leadingName 		= "GeneralApplication - " ;
	
	/**
	* Empty constructor, allowing to use some general purposes concrete methods (i.e. not static), 
	* without the need of a creating <code>Boudoir</code> instance.<br/><br/>
	* 
	* <strong>Remark</strong>: usage of this constructor is not suggested, since it creates
	* an incomplete <code>GeneralApplication</code> (e.g. without a logger object):
	* use <code>GeneralApplication(Boudoir boudoir)</code> instead.
	* 
	* @see GeneralApplication(Boudoir boudoir)
    */
    public GeneralApplication() {

    }	

	/**
	* Standard constructor, bounding this class to a single <code>Boudoir</code> instance.
	* 
	* @param boudoir the <code>Boudoir</code> instance "owning" this object
    */
    public GeneralApplication(Boudoir boudoir) {

    	this.boudoir 	= boudoir ;
    	this.logger 	= boudoir.getLogger().getLogger(GeneralApplication.class) ;
    }	

	/**
	* Method useful for debug and testing, the verify the scope of the web service objects.
    */
    public void finalize() {

		if (boudoir.getConfiguration().VERBOSE) System.out.println(leadingName + "Finalized object: " + this) ;
	}
	
// **************************************************************************************************************
// **********************************			ACCESSORS		*************************************************
// **************************************************************************************************************

	/**
	* Return the <code>Boudoir</code> instance representing the Boudoir library used by the 
	* application.
	*
	* @return a reference to the <code>Boudoir</code> object
    */
    public Boudoir getBoudoir()			 							{		return boudoir ; 					}
    
// **************************************************************************************************************
// ***********************************			SERVICE MANAGER			*****************************************
// **************************************************************************************************************

	/**
	* Prints in the LOG file some information about the library configuration.<br/>
	* Generally speaking, the library configuration depends by actual usage, hence users SHOULD NOT
	* rely on this method, but should instead override it in the subclass (the class extending the
	* <code>GeneralApplication</code>).
	*/
	public void showConfiguration() {
		
		Settings settings = this.boudoir.getConfiguration() ;
		
		// Preparing output message
		StringBuffer message = new StringBuffer("\nLIBRARY CONFIGURATION (") ;
		message.append(Settings.class + ")\n\n") ;
		message.append("Version           : " + settings.VERSION	+ "\n") ;
		message.append("Debug flag        : " + settings.DEBUG 		+ "\n") ;
		message.append("Verbose flag      : " + settings.VERBOSE 	+ "\n") ;
		message.append("Analyze flag      : " + settings.ANALYZE 	+ "\n") ;
		message.append("Manager flag      : " + settings.MANAGER 	+ "\n") ;
		message.append("Benchmark flag    : " + settings.BENCH 		+ "\n") ;
		message.append("\n") ;
		message.append("Logger system         : " + (settings.JAVA?"Java":EMPTY) + (settings.LOG4J?"Log4j":EMPTY) + "\n") ;
		message.append("Log path              : " + logger.getParentDir() 		+ "\n") ;
		message.append("Log rotation time     : " + logger.getRotationTime()	+ "\n") ;
		if (settings.JAVA || settings.BENCH){
			message.append("Java Logger directory : " + settings.LOGDIR + "\n") ;
			if (settings.JAVA) message.append	("Java Logger name      : " + settings.LOGGER + "\n") ;
			if (settings.BENCH) message.append	("Benchmark logger      : " + settings.BENCHMARK + "\n") ;
		}
		message.append("\n") ;
		message.append("Configuration file             : " + settings.CONFIG 			+ ".properties\n"	) ;
		message.append("Log4j configuration file       : " + settings.LOG_CONFIG 		+ ".properties\n"	) ;
		message.append("Multilanguage file             : " + settings.MESSAGES 			+ ".properties\n"	) ;
		logger.logInfo(message.toString()) ;
	}

	/**
	* Allow to configure the service dynamically, without reloading the AAR package on the application server.<br/>
	* Following commands are available:<br/><br/>
	* <ul>
	* <li>1: Log service configuration.</li>
	* <li>2: Log system properties and runtime data.</li>
	* <li>3: Log the list of threads in this group.</li>
	* <li>4: Log metadata information about the database connection.</li>
	* <li>5: Log result of the <code>SimpleAnalyzer</code> class.</li>
	* <li>6: Switch the <code>Settings.ANALYZE</code> flag on/off.</li>
	* <li>7: Switch the <code>Settings.VERBOSE</code> flag on/off.</li>
	* <li>8: Switch the <code>Settings.DEBUG</code> flag on/off.</li>
	* <li>9: Switch the <code>Settings.MANAGER</code> flag on/off.</li>
	* <li>10: Switch the <code>Settings.BENCH</code> flag on/off (and open/close the related logger).</li>
	* <li>11: Log the status of the JDBC connection pool.</li>
	* </ul><br/><br/>
	*
	* <strong>Remark:</strong> this method can be invoked only when the service in the MANAGER mode.
	* Hence, when this method is called using the option number 9 the MANAGER mode is turned off and
	* this method can not be invoked again!
	* This behavior can be used to "close" this access when security issues are required.
	*
	* @param key the integer command code as described above
	*/
	public boolean manageCommand(int key) {
		
		boolean found = true ;

		switch(key) {
			case 1  :		showConfiguration() ;		// Abstract method: see the ServiceMacro class
							break ;
			case 2  :		showJavaProperties() ;
							//showRuntime() ;
							break ;
			case 3  :		new ThreadManager(this.boudoir).logList() ;
							break ;
			case 4  :		boudoir.getDatabase().logMetaData() ;
							break ;
			case 5  :		// Can throw unpredictable exceptions if the analysis goes wrong
							if (boudoir.getAnalyzer() != null) try {
								boudoir.getAnalyzer().logResults() ;	
							} catch(Exception e){ 
								e.printStackTrace() ;				
							} else {
								logger.logWarning(leadingName + "ANALYZER IS OFF") ;
							}
							break ;
			case 6  :		boudoir.getConfiguration().ANALYZE 	= ! boudoir.getConfiguration().ANALYZE ;
							logger.logInfo(leadingName + "Analyzer mode: " + boudoir.getConfiguration().ANALYZE) ;
							break ;
			case 7  :		boudoir.getConfiguration().VERBOSE 	= ! boudoir.getConfiguration().VERBOSE ;
							logger.logInfo(leadingName + "Verbose mode: " + boudoir.getConfiguration().VERBOSE) ;
							break ;
			case 8  :		boudoir.getConfiguration().DEBUG 	= ! boudoir.getConfiguration().DEBUG ;
							logger.logInfo(leadingName + "Debug mode: " + boudoir.getConfiguration().DEBUG) ;
							break ;
			case 9  :		boudoir.getConfiguration().MANAGER 	= ! boudoir.getConfiguration().MANAGER ;
							logger.logInfo(leadingName + "Service manager: " + boudoir.getConfiguration().MANAGER) ;
							break ;
			case 10 :		boudoir.getConfiguration().BENCH 	= ! boudoir.getConfiguration().BENCH ;
							// Next line takes care also to initialize the benchmark logger (if necessary) 
							LoggerManager benchLogger = this.boudoir.getLogger().getBenchmarkLogger() ;
							if (boudoir.getConfiguration().BENCH) {
								logger.logInfo(leadingName + "Activated benchmark: " + benchLogger) ;
							} else {
								this.boudoir.getLogger().closeBenchmark() ;
								logger.logInfo(leadingName + "Closed benchmark: " + benchLogger) ;
							}
							break ;
			case 11  :		boudoir.getDatabase().logPoolsData() ;
							break ;
			default	:		found = false ;
							logger.logInfo(leadingName + "Invalid command code") ;
		}
		
		return found ;
	}

	/**
	* Overloaded version of the <code>manageCommand(int key)</code> method.
	*
	* @param value string version of the integer remote command code
	* @see manageCommand(int key)
	*/
	public boolean manageCommand(String value) {
		
		// Checking if PREFIX is already set
		if (this.PREFIX == null) {
			String		string		= "^" +  this.boudoir.getConfiguration().COMMAND + ".+" ;
			this.PREFIX	 			= Pattern.compile(string, Pattern.CASE_INSENSITIVE) ;
		}
		
		// Parsing input command
		Matcher matcher = PREFIX.matcher(value) ;
		if (matcher.matches()) {
			int 	key 	= 0 ;				
			int 	length 	= this.boudoir.getConfiguration().COMMAND.length() ;
			String 	command	= value.substring(length, value.length()) ;
			try 				{ key = Integer.parseInt(command.trim()) ;	}
			catch(Exception e) 	{ key = 0 ;									}
			// Forwarding action to actual method
			return this.manageCommand(key) ;
			
		} else {
			return false ;
		}
	}

	/**
	* Prints in the LOG file (or on STDOUT, if enable) all the Java properties set at run time. 
	* This method should be used only for testing or debug.
	*/
	protected void showJavaProperties() {

		// 1. Initializing
		Properties 			properties 	= System.getProperties() ;
		Enumeration<String>	e 			= (Enumeration<String>) properties.propertyNames() ;
		// Placeholder: not used at the moment
		//Runtime 			runtime 	= Runtime.getRuntime() ;
		//ClassLoader 		loader 		= getClass().getClassLoader() ;
		
		// 2. Evaluating max column size
		int max 	= 0 ;
		int size 	= 0 ;
		while (e.hasMoreElements()) {
			size++ ;
			String key = e.nextElement() ;
			if (key.length() > max) max = key.length() ; 
		}
		
		// 3. Producing string buffer
						e 			= (Enumeration<String>) properties.propertyNames() ;		// Resetting iterator position
		StringBuffer 	message 	= new StringBuffer("\nPRINTING " + size + " JAVA PROPERTIES:\n\n") ;
		while (e.hasMoreElements()) {
			String 			key 	= e.nextElement() ;
			StringBuffer 	line 	= new StringBuffer(key) ;
			while ( line.length() < max) line.append(" ") ;
			message.append(line + ": " + properties.getProperty(key) + "\n") ;
		}
		
		// 3. Optional properties (not always set): for the time being just one
		message.append("\nOTHER PROPERTIES:\n\n") ;		
		String[] keys 	= new String[] { "Oracle TSN names" 	} ;
		String[] props 	= new String[] { "oracle.net.tns_admin" } ;
		for (int i = 0 ; i < keys.length ; i++) {
			StringBuffer line = new StringBuffer(keys[i]) ;
			while ( line.length() < max) line.append(" ") ;
			message.append(line + ": " + properties.getProperty(props[i]) + "\n") ;			
		}
		
		// 5. Showing result
		logger.logInfo(message.toString()) ;
	}

    /**
	* Produce a short description list for remote manager commands accepted by the web service when the
	* MANAGER feature is activated. This information are produced in text format, and are ready to be
	* printed on a <code>OutputStream</code> object (usually STDOUT).
	*
	* @return a printable string containing the list of available remote commands
   */
    public StringBuffer getCommandsList() {

		String 			prefix 	= this.boudoir.getConfiguration().COMMAND ;
		StringBuffer 	message = new StringBuffer("\nREMOTE MANAGER HELP:\n") ;

		message.append("\n") ;
		message.append("Usage: java -jar boudoir.jar command\n") ;
		message.append("or     java -cp %CLASSPATH% adriani.boudoir.SimpleClient command\n") ;
		message.append("\n") ;
		message.append("Example : java -jar boudoir.jar " + prefix + "1\n") ;
		message.append("\n") ;
		message.append("For the complete commands list see below.\n") ;
		message.append("\n") ;
		message.append("Remote manager commands can be used also when invoking the library from command line,\n") ;
		message.append("in such scenario the command must be passed as command line argument,\n") ;
		message.append("with the '" + prefix + "' prefix before the command.\n") ;
		message.append("See documentation for details.\n") ;
		message.append("\n") ;

		message.append("REMOTE MANAGER COMMANDS LIST:\n") ;
		message.append("\n") ;
		message.append(prefix + "1  : Log service general configuration (and version)\n") ;
		message.append(prefix + "2  : Log Java system properties and some runtime data\n") ;
		message.append(prefix + "3  : Log the list of all threads in this group\n") ;
		message.append(prefix + "4  : Log metadata information about the database connection\n") ;
		message.append(prefix + "5  : Log result of the serviceAnalyzer class\n") ;
		message.append(prefix + "6  : Switch the Settings.ANALYZE flag on/off\n") ;
		message.append(prefix + "7  : Switch the Settings.VERBOSE flag on/off\n") ;
		message.append(prefix + "8  : Switch the Settings.DEBUG flag on/off\n") ;
		message.append(prefix + "9  : Switch the Settings.MANAGER flag on/off\n") ;
		message.append(prefix + "10 : Switch the Settings.BENCH flag on/off (and open/close the related logger)\n") ;
		message.append(prefix + "11 : Log the status of the JDBC connection pool\n") ;

		return message ;
	}

// **************************************************************************************************************
// **************************************		UTILITIES			*********************************************
// **************************************************************************************************************

   /**
    * Stops this thread for few milliseconds.<br/>
    *
    * @param time the time (in milliseconds) that this thread must be stopped
    */
    public void delay(long time) {

		try {
			Thread.currentThread().sleep(time) ;
		} catch (InterruptedException interrupted) {
			logger.logSevere(leadingName + ": error while sleeping") ;
		  	interrupted.printStackTrace() ;
		}
    }

	/**
	* Parse the string value of the input <code>String</code> into an integer value.
	* The <code>no_value</code> parameter represent the default integer value to be returned if
	* the parsing is not succesfull.
	*
	* @param string the input string object
	* @param no_value the int value to be returned if conversion fails
	* @return the integer value of the input string (or the <code>no_value</code> number)
	*/
	protected int valueOfString(String string, int no_value) {

		int result = no_value ;
		try 				{	result = Integer.parseInt( string.trim() ) ;								}
		catch(Exception e) 	{ 	result = no_value ;															}
		return result ;
	}
	
 	/**
 	* Helper method.
 	*
 	* @param first a <code>StringBuffer</code> object
 	* @param second another <code>StringBuffer</code> object
 	* @return <code>true</code> if the two string object contains the same text
 	*/
 	public boolean compareSB(StringBuffer first, StringBuffer second) {

 		String 	string = null ;
 		boolean result = false ;
 		if (first != null) 	string = first.toString() ;
 		if (string != null) result = string.equalsIgnoreCase( second.toString() ) ;
 		return result ;
 	}

} // End of class