/**
* Title: 		 Database
* Description:	 The class wrapping the database management
* Copyright:     Copyright (c) 2013
* Company:       Morpheus Quality Technologies
* Creation date: 06-12-2013
* Last modified: 11-02-2014
* @author 		 Stefano Adriani
* @version 		 1.0
*/

package adriani.boudoir.db ;

import adriani.boudoir.* ;
import adriani.boudoir.util.* ;
import adriani.boudoir.app.GeneralAnalyzer ;
import adriani.boudoir.app.GeneralApplication ;
import adriani.boudoir.logging.WSLogger ;

import java.sql.* ;
import java.util.* ;

/**
* The data-bridge class used to retrieve and return JDBC connection from the database. To obtain a connection,
* users should use a code block like this:<br/>
*
* <pre><code>
*  		Connection connection = Database.getMockConnection(targetDB) ;
*		// Perform your SQL operations here
*		Database.returnMockConnection(targetDB, this.mock[targetDB]) ;
* </code></pre><br/>
*
* By this mechanism this class will always return a <code>MockConnection</code>, which in general is not
* actually a mock connection, but it's just a wrapper for a normal java SQL connection.<br/><br/>
*
* Moreover this class holds some static reference labels and parameters common to all databases.<br/>
* Example: the <code>Database.DB_01</code>, <code>Database.DB_02</code> labels and the
* <em>Database.DB_NUMBER</em> parameter.
*/
 public class Database {
	 
	/** Reference to the top-level <code>Boudoir</code> instance.											*/
	public 		 			Boudoir 				boudoir ;

  	/** Reference code for the DB_01 database.																*/
	public static final	 	int 					DB_01 			= 0 ;
  	/** Reference code for the DB_02 database (Placeholder)													*/
	public static final 	int 					DB_02 			= 1 ;
  	/** Total number of databases handled by this class.													*/
	public 					int 					DB_NUMBER		= 1 ;
	
  	/** Connection data (loaded from configuration file).													*/
	protected 				String[]				alias ;
  	/** Connection data (loaded from configuration file).													*/
	protected 				String[]				address ;
  	/** Connection data (loaded from configuration file).													*/
	protected 				String[]				user ;
  	/** Connection data (loaded from configuration file).													*/
	protected				String[]				password ;

	/**
	* Specifies if connections are retrieved from a pure JDBC pool (the <code>JDBCConnectionPool</code> class)
	* or from a mocked connection pool (the <code>MockConnectionPool</code> class).
	*/
	protected 				boolean					mock ;

	/** Array of JDBC connection pool for the databases handled by this class.								*/
	protected 				JDBCConnectionPool[]	pool ;

	/** Sleep time for the cleaning thread for ALL the JDBC pools (in seconds).								*/
	protected 				int 					timeout ;

	/** Maximum allowed number of connections for ALL the JDBC pools.										*/
	protected 		 		int 					maximum ;

	/** Name of the current JDBC driver (for testing or debug).												*/
	protected 				String 					driver ;

	/** The JDBC prefix used to build the complete connect string.											*/
	public 		  			String 					jdbc_prefix ;

	/** Absolute path of the <code>tnsnames.ora</code> (if used).											*/
	protected 				String 					tnsnames ;

	/** Name of the Java Property where to set the reference to the <code>tnsnames.ora</code> file.			*/
	protected 				String 					tns_property	= "oracle.net.tns_admin" ;

	/** Beginning of each logging message produced by this class.											*/
	protected 		 		String 					leadingName		= "Database - " ;

	/** Retrieving the logger instance to be used for this class.											*/
	protected 		 		WSLogger				logger ;

	/**
	* Standard constructor, bounding this class to a single <code>Boudoir</code> instance.
	* 
	* @param boudoir the <code>Boudoir</code> instance "owning" this object.
    */
    public Database(Boudoir boudoir) {
    	
    	this.boudoir 	= boudoir ;
    	this.init(null, null) ;
    }
    	
	/**
	* Initialize this database object, with respect of configuration of the 
	* current <code>Boudoir</code> instance. Initialization takes care of loading connect strings
	* from configuration files and creation of necessary JDBC pools.<br/><br/>
	* 
	* <strong>Remark:</strong> user and password are optional. If passed as null values, the 
	* Initialization will use values retrieved from the configuration file.<br/>
	* On the opposite, if input values are not null, they will be used to override connection
	* strings for ALL databases.
	* 
	* @param login user name used to login to the database (optional)
	* @param pwd user's password used to login to the database (optional)
    */
    public void init(String login, String pwd) {

    	this.logger 	= boudoir.getLogger().getLogger(GeneralAnalyzer.class) ;

		// 1. Defining arrays (global variables)
		alias			= new String[DB_NUMBER] ;
		user			= new String[DB_NUMBER] ;
		password		= new String[DB_NUMBER] ;
		address			= new String[DB_NUMBER] ;
		// 1.1 Defining arrays (local variables)
		String[] host 			= new String[DB_NUMBER] ;
		String[] schema 		= new String[DB_NUMBER] ;

		// 2. Common DEFAULT values (overridden by properties values)
		timeout					= 900 ;
		maximum					= 100 ;
		driver					= "oracle.jdbc.driver.OracleDriver" ;
		tnsnames				= "/home/wsadm/tnsnames.ora" ;

		// 3. Connection DEFAULT values (overridden by properties values)
		alias		[0]			= "my_schema" ;
		user		[0]			= "user" ;
		password	[0]			= "password" ;
		address		[0]			= "jdbc:oracle:thin:@192.168.1.10:1521:db_name_01" ;

		// 4. Parsing properties from resource bundle
		ResourceBundle rb		= Utilities.loadBundle(this.boudoir.getConfiguration().CONFIG) ;
		if (rb != null) {
		
			// 4.1 Retrieving general properties
			driver				= rb.getString("jdbc.driver").trim() ;
			jdbc_prefix			= rb.getString("jdbc.prefix").trim() ;

			// 4.2 Retrieving properties for the DB_01 database
			for (int i = 0 ; i < DB_NUMBER ; i++) {
				int index = i + 1 ;
				host		[i]		= rb.getString("db_0" + index + ".host"		).trim() ;
				schema		[i]		= rb.getString("db_0" + index + ".schema"	).trim() ;
				user		[i]		= rb.getString("db_0" + index + ".user"		).trim() ;
				password	[i]		= rb.getString("db_0" + index + ".password"	).trim() ;
				alias		[i]		= rb.getString("db_0" + index + ".alias"	).trim() ;
				address		[i]		= jdbc_prefix + host[i] + "/" + schema[i] ;
				
				// Some hard-coded examples (keep as reference)
				//address	[i]		= "jdbc:postgresql://" + host[i] +"/" + schema[i] ;
				//address	[i]		= "jdbc:oracle:thin:@" + host[i] +":" + schema[i] ;
				
				// 4.3 Considering if we must override values read from the configuration file
				if (login 	!= null)	user[i]		= login ;	
				if (pwd 	!= null)	password[i]	= pwd ;	
			}
			
			// 4.4 Parsing boolean properties
			String mocked		= rb.getString("pool.mocked").trim() ;
			mock				= mocked.equalsIgnoreCase("YES") ;

			// 4.5 Parsing integer properties
			String value		= rb.getString("pool.timeout") ;
			try 				{ timeout = Integer.parseInt(value.trim()) ;	}
			catch(Exception e) 	{ timeout = 900 ;								}	// Default value (15 minutes)
			value				= rb.getString("pool.maximum") ;
			try 				{ maximum = Integer.parseInt(value.trim()) ;	}
			catch(Exception e) 	{ maximum = 100 ;								}	// Default value

			// 4.6 Setting system properties for tns system
			tnsnames = rb.getString("oracle.tnsnames") ;
			System.setProperty(tns_property, tnsnames) ;
			//System.setProperty("oracle.net.tns_admin", tnsnames) ;
			if (boudoir.getConfiguration().VERBOSE) System.out.println(leadingName + "Loaded configuration file: " + boudoir.getConfiguration().CONFIG + ".properties") ;

		} else {
			System.err.println(leadingName + "Using default configuration properties") ;
		}
		
		// 5. Create the JDBC pool (standard or mocked)
		initPool() ;
	}
    
	/**
	* 
	*/
	public void initPool() {
		
		// 1. Creating the array of proper connection pool types (pure JDBC or mocked).
		if (mock) {
			pool = new MockConnectionPool[DB_NUMBER] ;
			logger.logInfo(leadingName + "using mocked JDBC connection pool") ;
		} else {
			pool = new JDBCConnectionPool[DB_NUMBER] ;
			logger.logInfo(leadingName + "using pure (not mocked) JDBC connection pool") ;
		}

		// 2. Creating the static reference to the necessary JDBCConnectionPool objects.
		// Note that we register the JDBC driver only when creating the 1st pool.
		for (int i = 0 ; i < DB_NUMBER ; i++) {
			if (i == 0) {
				if (mock) 	pool[i] = new MockConnectionPool(boudoir, driver, address[i], user[i], password[i], timeout) ;
				else		pool[i] = new JDBCConnectionPool(boudoir, driver, address[i], user[i], password[i], timeout) ;
			} else {
				if (mock) 	pool[i] = new MockConnectionPool(boudoir, address[i], user[i], password[i], timeout) ;
				else 		pool[i] = new JDBCConnectionPool(boudoir, address[i], user[i], password[i], timeout) ;
			}
			// 2.1 Configuring the pool
			pool[i].setTnsName(alias[i]) ;
			pool[i].setPoolSize(maximum) ;
		}
		
		// Keep for testing or debug
		//System.out.println(leadingName + ">>>>>>>>>>>> created pool for boudoir: " + boudoir) ;		
	}

// **********************************************************************************************************
// *****************************			  ACCESSORS				*****************************************
// **********************************************************************************************************

	/**
	* Specifies if connections are retrieved from the pure JDBC pool (i.e. the <code>JDBCConnectionPool</code>)
	* or from the <code>MockConnectionPool</code>. Notice that all other information about the pool (timeout,
	* maximum limit etc.) can be retrieved by inspecting the proper pool instance as obtained by the method
	* <code>getJDBCPool(int targetDB)</code> method.
	*
	* @return <code>true</code> if this class relies on an instance of the <code>MockConnectionPool</code>,
	* <code>false</code> if the standard <code>JDBCConnectionPool</code> is used
	*/
	public boolean usesMockedPool() 								{		return mock ;					}

	/**
	* Return the <code>JDBCConnectionPool</code> object managing all the JDBC connections to the database
	* specified in input, according to the static fields reference defined by this class.<br/>
	* The returned pool can be a pure <code>JDBCConnectionPool</code> object or another valid subclass,
	* for example a <code>MockConnectionPool</code> class.<br/><br/>
	*
	* Example: <code>getJDBCPool(Datase.DB_01)</code> will return the JDBC pool for the DB_01 database.<br/><br/>
	*
	* @param targetDB a reference to the database (see static class field definitions)
	* @return the instance of the <code>JDBCConnectionPool</code> for the input database
	* @see #usesMockedPool()
	* @see JDBCConnectionPool
	* @see MockConnectionPool
	*/
	public JDBCConnectionPool getJDBCPool(int targetDB) 			{		return pool[targetDB] ;			}
	
// **********************************************************************************************************
// *****************************			  MUTATORS				*****************************************
// **********************************************************************************************************
	
	/**
	* Set the value for the <code>mock</code> flag, enabling or disabling the "Mocked Connections" feature.
	* 
	* @param flag the boolean value to be set in the <code>mock</code> class field.  
	*/
	public void setMockedPool(boolean flag) {
		
		this.mock = flag ;
		this.initPool() ;
	}
	
// **********************************************************************************************************
// *****************************			 POOL INTERFACE				*************************************
// **********************************************************************************************************

	/**
	* Return a new <code>MockConnection</code> wrapper containing the SQL connection, by retrieving
	* it from the proper pool (pure or mocked).
	*
	* @param targetDB a reference to the database (see static class field definitions)
	* @return the instance of the <code>JDBCConnectionPool</code> for the input database
	*/
	public MockConnection getMockConnection(int targetDB) throws SQLException {
		
		if (mock) {
			MockConnectionPool pool = (MockConnectionPool) getJDBCPool(targetDB) ;
			return pool.getMockConnection() ;
		} else {
			JDBCConnectionPool 	pool 		= getJDBCPool(targetDB) ;
			Connection 			connection 	= pool.getConnection() ;
			return new MockConnection(boudoir, connection) ;
		}
	}

	/**
	* Release an already opened SQL connection, wrapped by a <code>MockConnection</code> object,
	* returning it in the pool that originally produced such connection.
	*
	* @param targetDB a reference to the database (see static class field definitions)
	* @param wrapper the <code>MockConnection</code> wrapper to be released
	*/
	public void returnMockConnection(int targetDB, MockConnection wrapper) {

		if (wrapper == null) return ;

		if (mock) {
			//System.out.println(leadingName + "Returning mock: " + wrapper) ;
			MockConnectionPool pool = (MockConnectionPool) getJDBCPool(targetDB) ;
			pool.returnMockConnection(wrapper) ;
		} else {
			//System.out.println(leadingName + "Returning jdbc: " + wrapper) ;
			JDBCConnectionPool 	pool 		= getJDBCPool(targetDB) ;
			Connection 			connection 	= wrapper.getConnection() ;
			pool.returnConnection(connection) ;
		}
		
		// Keep for testing or debug
		//System.out.println(leadingName + "Returning connection from DB instance: " + this) ;
	}

// **********************************************************************************************************
// *****************************			DATABASE METADATA			*************************************
// **********************************************************************************************************

	/**
	* Print in the LOG file and on the STDOUT some meta-data information about ALL the databases used
	* by this project.
	*/
	public void logMetaData() {

		for (int i = 0 ; i < DB_NUMBER ; i++) logMetaData(i) ;
	}

	/**
	* Print in the LOG file and on the STDOUT some meta-data information about the database selected by the
	* input parameter, according to the static fields reference defined by this class.<br/><br/>
	*
	* Example: <code>logMetaData(Datase.DB_01)</code> will show meta-data about the DB_01 database.<br/>
	*
	* @param target a reference to the database (see static class field definitions)
	*/
	public void logMetaData(int target) {

		StringBuffer message = showMetaData(target) ;
		if (message != null) logger.logInfo( message.toString() ) ;
	}

	/**
	* Produces a formatted string containing some meta-data information about the database selected by the
	* input <code>target</code> parameter, according to the static fields reference defined by this class.<br/><br/>
	*
	* Example: <code>showMetaData(Datase.DB_01)</code> will produce meta-data about the DB_01 database.<br/>
	*
	* @param target a reference to the database (see static class field definitions)
	*/
	public StringBuffer showMetaData(int target) {

		// Initializing
		StringBuffer 		result 		= null ;
		Connection 			connection 	= null ;
		JDBCConnectionPool 	pool 		= getJDBCPool(target) ;

		// Retrieving a connection from the target database (via the MockConnection wrapper)
		try {
			connection = getMockConnection(target).getConnection() ;
		}
		catch(Exception e) {
			System.err.println(leadingName + "Error getting the connection") ;
			e.printStackTrace() ;
		}

		// Getting meta-data from the connection and the JDBC pool
		if (connection != null) try {
			result = getMetaData(connection) ;
			result.append("JDBC pool mocked                   : " + mock 					+ "\n"	) ;
			result.append("JDBC pool using TNS names          : " + pool.supportTnsNames() 	+ "\n"	) ;
			result.append("Sleep time for the cleaning thread : " + timeout + " s"			+ "\n"	) ;
			result.append("TNS names file                     : " + tnsnames 				+ "\n"	) ;
			result.append("JDBC pool object                   : " + pool 					+ "\n"	) ;
			result.append("**********************************************************************\n") ;
		} catch(SQLException ex) {
			System.err.println(leadingName + "Error retrieving metadata") ;
			ex.printStackTrace() ;
		}

		return result ;
	}

	/**
	* Produce the meta-data information about the input database's connection object and format this data
	* into a <code>StringBuffer</code> object.
	*
	* @return a printable string containing meta-data information about the database
	*/
	public StringBuffer getMetaData(Connection connection) throws SQLException {

		Driver driver = null ;
		String driverPath = null ;
		DatabaseMetaData metadata = connection.getMetaData() ;

		// Detecting the path of the JDBC driver
		try {
			driver 				= (Driver) Class.forName(this.driver).newInstance() ;
			java.net.URL url 	= driver.getClass().getProtectionDomain().getCodeSource().getLocation() ;
			driverPath 			= url.getPath() ;
		} catch(Exception e) 	{ System.err.println(leadingName + "Meta Data error") ; }

		// Producing the string message
		StringBuffer message = new StringBuffer() ;
		message.append("\n********************* DATABASE INFORMATION ***************************\n") ;
		message.append("Connected to (tsn):       " + metadata.getURL() 						+ "\n") ;
		message.append("Driver:                   " + metadata.getDriverName() 					+ "\n") ;
		message.append("Driver class:             " + driver.getClass() 						+ "\n") ;
		message.append("Driver archive:           " + driverPath 								+ "\n") ;
		message.append("Driver version:           " + metadata.getDriverVersion() 				+ "\n") ;
		message.append("Support Batch Updates?    " + metadata.supportsBatchUpdates() 			+ "\n") ;
		message.append("Support Stored Functions? " + metadata.supportsStoredProcedures() 		+ "\n") ;
		message.append("Auto commit mode?         " + connection.getAutoCommit() 				+ "\n") ;
		message.append("\n") ;
		message.append("Current isolation lvl:    " + connection.getTransactionIsolation() 		+ "\n") ;
		message.append("Default isolation lvl:    " + metadata.getDefaultTransactionIsolation() + "\n") ;
		message.append("Where:\n") ;
		message.append("   Level " + java.sql.Connection.TRANSACTION_READ_UNCOMMITTED 	+ " = TRANSACTION_READ_UNCOMMITTED\n"	) ;
		message.append("   Level " + java.sql.Connection.TRANSACTION_READ_COMMITTED 	+ " = TRANSACTION_READ_COMMITTED\n"		) ;
		message.append("   Level " + java.sql.Connection.TRANSACTION_REPEATABLE_READ 	+ " = TRANSACTION_REPEATABLE_READ\n"	) ;
		message.append("   Level " + java.sql.Connection.TRANSACTION_SERIALIZABLE 		+ " = TRANSACTION_SERIALIZABLE\n\n"		) ;

		return message ;
	}

	/**
	* Logs some information about the status of the JDBC pools used by this database class.<br/>
	* In current implementation this method is executed when the web service receive the command 00000011.
	*
	* @see GeneralApplication#manageCommand(int key)
    */
    public void logPoolsData() {

		StringBuffer message = new StringBuffer() ;
		message.append("\n") ;
		for (int i = 0 ; i < DB_NUMBER ; i++) {
			message.append("JDBCPool (ID = " + i + ") : " + pool[i].getPoolData() + "\n") ;
		}
		if (message != null) logger.logInfo( message.toString() ) ;
	}

} // End of class