/**
* Title: 		 GeneralDAO
* Description:	 The parent class for all Data Access Object classes (DAO)
* Copyright:     Copyright (c) 2013-2014
* Company:       Morpheus Quality Technologies
* Creation date: 06-12-2013
* Last modified: 09-01-2014
* @author 		 Stefano Adriani
* @version 		 1.0
*/

package adriani.boudoir.db ;

import adriani.boudoir.logging.SystemLogger ;
import adriani.boudoir.logging.WSLogger ;
import adriani.boudoir.util.* ;
import adriani.boudoir.Boudoir ;
import adriani.boudoir.Settings ;

import java.sql.* ;
import java.util.* ;
import java.text.SimpleDateFormat ;

/**
* GeneralDAO - general class used to provide some common methods to all database accessors (DAO classes).<br/>
* This class defines some general purposes objects: class fields, double quoting methods, introspection methods,
* standard accessors and benchmark analysis. The core utilities of this class is a set of methods wrapping
* the more common SQL statements: retrieving one single record, retrieving a result set, calling a PL/SQL
* function or procedure, calling a pipelined PL/SQL function etc.
* All connections issues SHOULD NOT be implemented at this level, but should be left to the children classes,
* in order to reflect the detailed requirements of the business logic.
*/
public class GeneralDAO {
	
	/** Reference to the top-level <code>Boudoir</code> instance.											*/
	public 		 			Boudoir 			boudoir ;
	
  	/** 
  	* The array of mock connections for the necessary database(s) (accessed by this DAO).
  	* Each element in the array represent a single connection to the indexed-element, for instance:<br/><br/>
  	* 
  	* <code>mock[DB_01]</code> : connection to 1st database<br/>
  	* <code>mock[DB_02]</code> : connection to 2nd database<br/>
  	* etc...
  	*/
	protected 				MockConnection[]	mock ;

  	/** Status of the DAO after execution of the last SQL statement. 										*/
	protected 				int 				status ;

	/** Long timestamp used to evaluate performance time when the ANALYSIS mode is used.					*/
	protected				long 				chronStart ;

	/** Static final <code>String</code> value stored as class fields to improve performances.				*/
	protected static final	String				EMPTY			= "" ;
	/** Static final <code>String</code> value stored as class fields to improve performances.				*/
	public static final 	String				QUOTE 			= "'" ;
	/** Static final <code>String</code> value stored as class fields to improve performances.				*/
	public static final 	String				PERCENT			= "%" ;
	/** Static final <code>String</code> value stored as class fields to improve performances.				*/
	public static final 	String				DOUBLE_QUOTES 	= "''" ;

	/** The message to print in the case of a general Java <code>Exception</code> .							*/
	protected static final	String				ERROR	 		= "Java error:\n\n"  ;
  	/** The message to print in the case of a <code>SQLException</code> .									*/
	protected static final	String				QUERY_NOK 		= "query NOK:\n\n"  ;
  	/** The message to print in the case of a rollback failure.												*/
	protected static final	String				ROLLBACK_NOK 	= "rollback-failed: "  ;
	/** The message to print if preparation of the prepared statement failed.								*/
	protected static final 	String 				PSTMT_NOK 		= "failed preparation of the prepared statement " ;
	/** The message to print if the query returned no data at all.											*/
	protected static final 	String 				NO_DATA_MSG		= "no data found" ;

	/** The integer storing the Oracle error code ORA-01403: no data found.									*/
	protected static final	int					NO_DATA_FOUND	= 1403  ;
	
	/** Reference to the "auto-type" feature allowing to inspect column's datatypes (default value)			*/
	public static final		int					VARCHAR			= 101 ;
	/** Reference to the "auto-type" feature allowing to inspect column's datatypes.						*/
	public static final		int					AUTO			= 102 ;

	/** DateFormat static reference for conversion from the format YYYY-MM-DD HH-MM-SS.M.					*/
	protected static 		SimpleDateFormat	db_timestamp	= new SimpleDateFormat("yyyy-MM-dd HH:mm:ss.S") ;

	/** Retrieving the logger instance to be used for this class.											*/
	protected 		 		WSLogger			logger ;

	/** The leading name of each message logged by this class.												*/
	protected 				String 				leadingName 	= "General DAO - " ;
	
	/**
	* Stateless constructor.
    */
    public GeneralDAO() {
    	
    }

	/**
	* Standard constructor, bounding this class to a single <code>Boudoir</code> instance.
	* 
	* @param boudoir the <code>Boudoir</code> instance "owning" this object.
    */
    public GeneralDAO(Boudoir boudoir) {

    	this.boudoir 	= boudoir ;
		this.mock 		= new MockConnection[ boudoir.getDatabase().DB_NUMBER ] ;    	
    	this.logger 	= boudoir.getLogger().getLogger(GeneralDAO.class) ;    	
    	
		if (boudoir.getConfiguration().VERBOSE) System.out.println(leadingName + "Created parent: " + this + "\n") ;
	}

 // **********************************************************************************************************
 // *****************************		CONNECTION MANAGEMENT		*****************************************
 // **********************************************************************************************************

 	/**
 	* Open the connection to the target database. The connection to be opened is selected by the input
 	* <code>targetDB</code> parameter, for example:<br/><br/>
 	*
 	* targetDB = DB_01 --> open a connection to the DB_01 database.<br/>
 	* targetDB = DB_02 --> open a connection to the DB_02 database.<br/>
 	* etc...<br/><br/>
 	*
 	* Connections are retrieved from the <code>Database</code> class using the <code>getMockConnection</code>
 	* mechanism, which allows to retrieve both simple connections or wrapped connections.<br/>
 	* Current implementation immediately QUIT execution if we get an error opening the connection,
 	* logging a SEVERE error message and returning <code>false</code>. Usually this means that
 	* the library will not produce a valid response: the application could receive a response
 	* containing a message like "database not found".<br/>
 	*
 	* @param targetDB the reference code of the database where to open the connection
 	* @return <code>true</code> if the connection has been opened successfully
 	* @see Database#getMockConnection(int targetDB)
 	*/
    public boolean openConnection(int targetDB) {

 		// 1. Checking if general Database environment is properly set
 		if (mock == null) {
 			logger.logSevere(leadingName + "MockConnection array not initialized") ;
 			return false ;
 		} else if (targetDB >= mock.length) {
 			logger.logSevere(leadingName + "Database reference number " + targetDB + " can not be handled") ;
 			return false ;
 		}
 		
 		// 2. Trying to retrieve a valid connection wrapper
 		try {
 			this.mock[targetDB] = boudoir.getDatabase().getMockConnection(targetDB) ;
 			return (mock[targetDB] != null) ;
 		}
 		catch(SQLException e) {
 			logger.logSevere(leadingName + "Connection to database " + targetDB + " not found") ;
 	  		e.printStackTrace() ;
 			return false ;
 		}
 		catch(Exception e) {
 			logger.logSevere(leadingName + "Java error opening connections to database " + targetDB) ;
 	  		e.printStackTrace() ;
 			return false ;
 		}
 	}

 	/**
 	* Close the connection specified by the input target parameter, for example:<br/><br/>
 	*
 	* targetDB = DB_01 --> close the connection to the DB_01 database.<br/>
 	* targetDB = DB_02 --> close the connection to the DB_02 database.<br/>
 	* etc...<br/><br/>
 	*
 	* Current implementation DOES NOT quit execution if we get an error returning the connection.
 	* This means that the library could provide a valid output, but connections to the database
 	* will be left "hanged up" and could cause a database crash. For this reason such a situation is logged
 	* as a SEVERE error by the log system.<br/>
 	*
 	* @param targetDB the number of the database where to return the connection
    */
    public void closeConnection(int targetDB) {

 		if ( (mock != null) && (targetDB < mock.length) ) {
 			boudoir.getDatabase().returnMockConnection(targetDB, this.mock[targetDB]) ;
 			this.mock[targetDB] = null ;
 		} else {
 			StringBuffer message = new StringBuffer(leadingName + "Wrong database enviroment: ") ;
 			message.append("Impossible to return connection for database " + targetDB) ;
 			logger.logSevere(message.toString()) ;
 		}
 	}

 	/**
 	* Return the connection currently opened through the target database, for example:<br/><br/>
 	*
 	* targetDB = DB_01 --> close the connection to the DB_01 database.<br/>
 	* targetDB = DB_02 --> close the connection to the DB_02 database.<br/>
 	* etc...<br/><br/>
 	*
 	* @param targetDB the number of the database offering the connection
 	* @return the currently opened connection to the target database
    */
    public Connection getConnection(int targetDB) {

 		Connection connection = null ;

 		if ( (mock != null) && (targetDB < mock.length) ) connection = mock[targetDB].getConnection() ;
 		else {
 			StringBuffer message = new StringBuffer(leadingName + "Wrong database enviroment: ") ;
 			message.append("Impossible to use connection for database " + targetDB) ;
 			logger.logSevere(message.toString()) ;
 		}

 		return connection ;
 	}
     
  	/**
  	*
  	*
    */
    public void resetPool() {
    	
    	String old 			= this.mock.toString() ;
    	String old_label 	= old.substring(old.length() - 8) ;
		this.mock = new MockConnection[ boudoir.getDatabase().DB_NUMBER ] ;
    	String now 			= this.mock.toString() ;
    	String now_label 	= now.substring(now.length() - 8) ;
		logger.logWarning(leadingName + "Resetting DAO connection pool: " + old_label + " -> " + now_label) ;
    }

// **********************************************************************************************************
// ************************************		BENCHMARK ANALYSIS			*************************************
// **********************************************************************************************************

	/**
	* Log performances data on the <em>Performances Analyser</em> and flush the relative Handler.
	* This method automatically adds a detailed timestamp to the logged message, and store the
	* starting timestamp in the <code>chronStart</code> class field.
	*
	* @param msg the string text name of the query that is going to be executed
	* @see Settings#BENCHMARK
	*/
	public void startQuery(String msg) {

		this.chronStart = System.currentTimeMillis() ;
		java.util.logging.Logger logger = java.util.logging.Logger.getLogger(boudoir.getConfiguration().BENCHMARK) ;
		logger.info("(" + chronStart + ") START QUERY: " + msg) ;
		flushLogger(logger) ;
	}

	/**
	* Log performances data on the <em>Performances Analyser</em> and flush the relative Handler.
	* This mehod compares the query end time with the query starting time and log the query
	* duration time in the performances logger.
	*
	* @param msg the string text name of the query that has just been terminated
	* @see Settings#BENCHMARK
	*/
	public void endQuery(String msg) {

		long endTime = System.currentTimeMillis() ;
		java.util.logging.Logger logger = java.util.logging.Logger.getLogger(boudoir.getConfiguration().BENCHMARK) ;
		logger.info("(" + endTime + ") END QUERY: " + msg + "   Time= " + (endTime - chronStart) +  " ms") ;
		flushLogger(logger) ;
	}

	/**
	* Flush the handler bound to the given <code>Logger</code>. Notice that the <strong>benchmark</strong>
	* feature relies always on the <code>SystemLogger</code> class, never on the Log4j mechanism.
	*
	* @param logger the key string reference to the logger object to be flushed
	* @see SystemLogger
	*/
	public void flushLogger(java.util.logging.Logger logger) {

		java.util.logging.Handler[] handlers = logger.getHandlers() ;
		if ( (handlers.length > 0) && (handlers[0] != null) ) handlers[0].flush() ;
	}

// **********************************************************************************************************
// ************************************		PERFORMANCE ANALYSIS		*************************************
// **********************************************************************************************************

	/**
	* We override the parent (empty) method in order to register the execution status of a general SQL
	* statement in the <code>SimpleAnalyzer</code> class, by calling:<br/><br/>
	*
	* <code>simpleMacro.getServiceAnalyzer().collect(status) ;</code><br/><br/>
	*
	* @param statusValue the integer status as defined by the <code>ErrorMessage</code> class fields
	* @see ErrorMessage
	*/
	public void collectStatus(int statusValue) {

		if (boudoir != null) 		boudoir.getAnalyzer().collect(status) ;
		else 						logger.logWarning(leadingName + "Service Analyser closed because SimpleMacro is null") ; 
	}

// ************************************************************************************************************
// ********************************			GENERAL SQL QUERIES			***************************************
// ************************************************************************************************************

	/**
	* Executes a SELECT query to the target database, returning a <strong>single record</strong> described
	* by a <code>StringBuffer</code> array. The length of the array is defined by the <code>cols</code>
	* input parameter, representing the number of columns to be retrieved from the database.
	* This method just forward action to the full-signature <code>getRecord()</code> method, assuming 
	* as default return types the <code>VARCHAR</code> type (for all columns).
	*
	* @param targetDB the number of the database where to execute the query
	* @param label the logical name of the statement, only for logging purposes
	* @param query the SQL statement to be executed, already filled with parameters
	* @param cols the number of columns to be retrieved from the database
	* @return a <code>StringBuffer</code> array containing the selected columns (from the same record)
	* @see #getRecord(int targetDB, String label, String query, int cols, int type)
    */
    public StringBuffer[] getRecord(int targetDB, String label, String query, int cols) {
    	
    	return this.getRecord(targetDB, label, query, cols, VARCHAR) ;
    }
    
	/**
	* Executes a SELECT query to the target database, returning a <strong>set of records</strong> described
	* by an array whose elements are others <code>StringBuffer</code> arrays:<br/><br/>
	*
	* <code>StringBuffer[n][i]</code><br/><br/>
	*
	* contains the i-th column of the n-th record.<br/><br/>
	* 
	* This method just forward action to the full-signature <code>getRecordSet()</code> method, assuming 
	* as default return types the <code>VARCHAR</code> type (for all columns).
	*
	* @param targetDB the number of the database where to execute the query
	* @param label the logical name of the statement, only for logging purposes
	* @param query the SQL statement to be executed, already filled with parameters
	* @param cols the number of columns to be retrieved from the database

	* @return a <code>StringBuffer</code> 2D array containing all the matching records
    */
    public StringBuffer[][] getRecordSet(int targetDB, String label, String query, int cols) {
    	
    	return this.getRecordSet(targetDB, label, query, cols, VARCHAR) ;
    }
	
	/**
	* Executes a SELECT query to the target database, returning a <strong>single record</strong> described
	* by a <code>StringBuffer</code> array. The length of the array is defined by the <code>cols</code>
	* input parameter, representing the number of columns to be retrieved from the database.
	*
	* @param targetDB the number of the database where to execute the query
	* @param label the logical name of the statement, only for logging purposes
	* @param query the SQL statement to be executed, already filled with parameters
	* @param cols the number of columns to be retrieved from the database
	* @param type a reference to the criterion used to evaluate columns types 
	* @return a <code>StringBuffer</code> array containing the selected columns (from the same record)
    */
    public StringBuffer[] getRecord(int targetDB, String label, String query, int cols, int type) {

		// Preparing variables
		status = ErrorMessage.DAO ;
		StringBuffer[] output = null ;
		if (boudoir.getConfiguration().BENCH) startQuery(label) ;
		//System.out.println(leadingName + "getRecord (" + ((type == AUTO) ? "AUTO" : "VARCHAR") + ")" ) ;

		// Choosing database's connection
		Connection connection = this.getConnection(targetDB) ;

		// Executing statement
		try {
	 		Statement stmt		= connection.createStatement() ;
	  		ResultSet result 	= stmt.executeQuery(query) ;
			//System.out.println("Executed query: " + query) ;
			//System.out.println( inspectResultSet(result) ) ;

			// Reading data only if we find a valid result set
			if ( result.isBeforeFirst() ) {
				output = new StringBuffer[cols] ;
				status = ErrorMessage.OK ;
				result.next() ;
				for (int i = 0 ; i < cols ; i++) {
					//System.out.println( inspectColumn(result, i + 1) ) ;					
					String data = null ;
					if (type == AUTO) 	data 		= fetchColumnData(result,i + 1) ;
					else 				data 		= result.getString(i + 1) ;				// Default: VARCHAR type
					if (data != null) 	output[i] 	= new StringBuffer(data) ;
				}
				if (boudoir.getConfiguration().VERBOSE) {
					System.out.print(leadingName + label) ;
					for (int i = 0 ; i < cols ; i++) if (output[i] != null) System.out.print(output[i].toString() + " ") ;
					System.out.println() ;
				}
			}
			else status = ErrorMessage.NO_RECORD ;
			stmt.close() ;

		} catch(SQLException e) {
			status = ErrorMessage.SQL ;
			logger.logSevere(leadingName + label + QUERY_NOK + query + "\n" + e.getMessage()) ;
	  		if (boudoir.getConfiguration().DEBUG) e.printStackTrace() ;
	  		try 					{ 	if (connection != null) connection.rollback() ;			  				}
	  		catch(SQLException rbe) { 	System.err.println(ROLLBACK_NOK + rbe.getMessage()) ;					}
		} catch(Exception e) {
			status = ErrorMessage.DAO ;
			logger.logSevere(leadingName + label + ERROR + query + "\n" + e.getMessage()) ;
	  		if (boudoir.getConfiguration().DEBUG) e.printStackTrace() ;
		}

		// Optional: logging results
		if (boudoir.getConfiguration().BENCH) 	endQuery(label) ;
		if (boudoir.getConfiguration().ANALYZE) collectStatus(status) ;
		if (status == ErrorMessage.NO_RECORD) 	logger.logWarning(leadingName + label + boudoir.getMessages().getMessage(status)) ;

		return output ;
	}

	/**
	* Executes a SELECT query to the target database, returning a <strong>set of records</strong> described
	* by an array whose elements are others <code>StringBuffer</code> arrays:<br/><br/>
	*
	* <code>StringBuffer[n][i]</code><br/><br/>
	*
	* contains the i-th column of the n-th record.<br/>
	*
	* @param targetDB the number of the database where to execute the query
	* @param label the logical name of the statement, only for logging purposes
	* @param query the SQL statement to be executed, already filled with parameters
	* @param cols the number of columns to be retrieved from the database
	* @param type a reference to the criterion used to evaluate columns types
	* @return a <code>StringBuffer</code> 2D array containing all the matching records
    */
    public StringBuffer[][] getRecordSet(int targetDB, String label, String query, int cols, int type) {

		// Preparing variables
		status = ErrorMessage.DAO ;
		StringBuffer[] record = null ;
		StringBuffer[][] output = null ;
		if (boudoir.getConfiguration().BENCH) startQuery(label) ;

		// Choosing database's connection
		Connection connection = this.getConnection(targetDB) ;

		// Executing statement
		try {
	 		Statement stmt = connection.createStatement() ;
	  		ResultSet result = stmt.executeQuery(query) ;
			//System.out.println("Executed query: " + query) ;

			// Reading data only if we find a valid result set
			if ( result.isBeforeFirst() ) {
				ArrayList list = new ArrayList() ;
				status = ErrorMessage.OK ;
				while (result.next()) {
					record = new StringBuffer[cols] ;
					for (int i = 0 ; i < cols ; i++) {
						//System.out.println( inspectColumn(result, i + 1) ) ;					
						String data = null ;
						if (type == AUTO) 	data 		= fetchColumnData(result,i + 1) ;
						else 				data 		= result.getString(i + 1) ;				// Default: VARCHAR type
						if (data != null) 	record[i] 	= new StringBuffer(data) ;
						//if (i == <value>) logger.logInfo("Column:#" + record[i] + '#') ;						
						
					}
					list.add(record) ;
				}
				int size = list.size() ;
				output = (StringBuffer[][]) list.toArray( new StringBuffer[size][cols] ) ;
				if (boudoir.getConfiguration().VERBOSE) System.out.println(leadingName + label + "found " + size + " records") ;
				// Keep for testing or debug: printing the results 
				//for (int n = 0 ; n < size ; n++) { for (int i = 0 ; i < cols ; i++) if (output[n][i] != null) System.out.print(output[n][i].toString() + " ") ; System.out.println() ; }
			}
			else status = ErrorMessage.NO_RECORD ;
			stmt.close() ;

		} catch(SQLException e) {
			status = ErrorMessage.SQL ;
			logger.logSevere(leadingName + label + QUERY_NOK + query + "\n" + e.getMessage()) ;
	  		if (boudoir.getConfiguration().DEBUG) e.printStackTrace() ;
	  		try 					{ 	if (connection != null) connection.rollback() ;			  				}
	  		catch(SQLException rbe) { 	System.err.println(ROLLBACK_NOK + rbe.getMessage()) ;					}
		} catch(Exception e) {
			status = ErrorMessage.DAO ;
			logger.logSevere(leadingName + label + ERROR + query + "\n" + e.getMessage()) ;
	  		if (boudoir.getConfiguration().DEBUG) e.printStackTrace() ;
		}

		// Optional: logging results
		if (boudoir.getConfiguration().BENCH) 	endQuery(label) ;
		if (boudoir.getConfiguration().ANALYZE) collectStatus(status) ;
		if (status == ErrorMessage.NO_RECORD) 	logger.logWarning(leadingName + label + boudoir.getMessages().getMessage(status)) ;

		return output ;
	}

// ************************************************************************************************************
// ************************************			PL/SQL FUNCTIONS		***************************************
// ************************************************************************************************************
    
	/**
	* Executes a PL/SQL FUNCTION on the target database, returning a <strong>single record</strong> wrapped
	* by a <code>StringBuffer</code> array.<br/>
	* This method just forward action to the full-signature <code>getFunction()</code> method, assuming 
	* as default return type the <code>VARCHAR</code> type.
	* 
	* @param targetDB the number of the database where to execute the query
	* @param label the logical name of the statement, only for logging purposes
	* @param query the prepared call statement to be executed, not yet filled with parameters
	* @param param the <code>Object</code> array containing the input parameters
	* @return a <code>StringBuffer</code> array containing the selected columns
	* @see #getFunction(int targetDB, String label, String query, Object[] param, int type)
    */
    public StringBuffer[] getFunction(int targetDB, String label, String query, Object[] param) {
    	
    	return this.getFunction(targetDB, label, query, param, java.sql.Types.VARCHAR) ;
    }

	/**
	* Executes a PL/SQL FUNCTION on the target database, returning a <strong>single record</strong> wrapped
	* by a <code>StringBuffer</code> array.<br/>
	* Current implementation relies on the <code>{? = call FUNCTION(?, ?, ?, ....)}</code> syntax,
	* because is generally more performing than the alternative approaches:<br/><br/>
	*
	* <code>SELECT FUNCTION(x, y, z, ...) FROM DUAL</code> (ordinary plain old java).<br/>
	* <code>SELECT FUNCTION(?, ?, ?, ...) FROM DUAL</code> (JDBC prepared statement).<br/>
	*
	* @param targetDB the number of the database where to execute the query
	* @param label the logical name of the statement, only for logging purposes
	* @param query the prepared call statement to be executed, not yet filled with parameters
	* @param param the <code>Object</code> array containing the input parameters
	* @param type the <code>java.sql.Types</code> output type
	* @return a <code>StringBuffer</code> array containing the selected columns
    */
    public StringBuffer[] getFunction(int targetDB, String label, String query, Object[] param, int type) {

		// Preparing variables
		int cols = 1 ;
		int errorCode = 0 ;
		status = ErrorMessage.DAO ;
		StringBuffer[] output = null ;
		if (boudoir.getConfiguration().BENCH) startQuery(label) ;

		// Choosing database's connection
		Connection connection = this.getConnection(targetDB) ;

		// Executing statement
		try {
			CallableStatement call = connection.prepareCall(query) ;
			// Setting output parameters
			call.registerOutParameter(1, type) ;
			// Setting input parameters
			if (param != null) for (int i = 0 ; i < param.length ; i++) {
				int n = cols + 1 + i ;
				setColumnData(call, n, param[i]) ;
			}
	  		call.execute() ;
			//System.out.println("Executed call: " + query) ;

			// Reading output data
			status = ErrorMessage.OK ;
			output = new StringBuffer[cols] ;
			for (int i = 0 ; i < cols ; i++) {
				Object 				data 		= getColumnData(call, i + 1, type) ;						
				if (data != null) 	output[i] 	= new StringBuffer( data.toString() ) ;
				else 				status 		= ErrorMessage.NO_RECORD ;
			}
			if (boudoir.getConfiguration().VERBOSE) {
				System.out.print(leadingName + label) ;
				for (int i = 0 ; i < cols ; i++) if (output[i] != null) System.out.print(output[i].toString() + " ") ;
				System.out.println() ;
			}
			call.close() ;

		} catch(SQLException e) {
			status = ErrorMessage.SQL ;
			errorCode = e.getErrorCode() ;
			// Avoiding to log a critical message if we just did not find any data
			if (errorCode != NO_DATA_FOUND) {
				logger.logSevere(leadingName + label + QUERY_NOK + query + "\n" + e.getMessage()) ;
				if (boudoir.getConfiguration().DEBUG) e.printStackTrace() ;
			} else {
				logger.logSevere(leadingName + label + NO_DATA_MSG) ;
			}
			// Trying to rollback anyway
	  		try 					{ 	if (connection != null) connection.rollback() ;			  				}
	  		catch(SQLException rbe) { 	System.err.println(ROLLBACK_NOK + rbe.getMessage()) ;					}
		} catch(Exception e) {
			status = ErrorMessage.DAO ;
			logger.logSevere(leadingName + label + ERROR + query + "\n" + e.getMessage()) ;
	  		if (boudoir.getConfiguration().DEBUG) e.printStackTrace() ;
		}

		// Optional: logging results
		if (boudoir.getConfiguration().BENCH) 	endQuery(label) ;
		if (boudoir.getConfiguration().ANALYZE) collectStatus(status) ;
		if (status == ErrorMessage.NO_RECORD) 	logger.logWarning(leadingName + label + boudoir.getMessages().getMessage(status)) ;

		return output ;
	}

	/**
	* Executes a PL/SQL FUNCTION on the target database, returning a <strong>set of records</strong> described
	* by an array whose elements are others <code>StringBuffer</code> arrays:<br/><br/>
	*
	* <code>StringBuffer[n][i]</code><br/><br/>
	*
	* contains the i-th column of the n-th record.<br/><br/>
	*
	* Current implementation is just a wrapper to the <code>getRecordSet()</code> method, since the best way
	* to execute a function returning a pipe is to use an ordinary SELECT statement.
	*
	* @param targetDB the number of the database where to execute the query
	* @param label the logical name of the statement, only for logging purposes
	* @param query the SQL statement to be executed, already filled with parameters
	* @param cols the number of columns to be retrieved from the database
	* @return a <code>StringBuffer</code> 2D array containing all the matching records
    */
    public StringBuffer[][] getFunctionPipe(int targetDB, String label, String query, int cols) {

		return getRecordSet(targetDB, label, query, cols, VARCHAR) ;
	}

// ************************************************************************************************************
// ************************************			PL/SQL PROCEDURE		***************************************
// ************************************************************************************************************

	/**
	* Executes a PL/SQL PROCEDURE call to the target database, returning a <strong>single record</strong>
	* described by a <code>StringBuffer</code> array. The length of the output array is defined by the
	* <code>cols</code> input parameter (number of columns to be retrieved from the record).
	* <br/><br/>
	* If the procedure does not require any IN parameter the input <code>param</code> array must be null.
	* If the procedure requires some IN OUT parameters, the index of these parameters MUST BE specified in
	* the <code>inout</code> input integer array.
	* If there are no IN OUT parameters, the <code>inout</code> array MUST BE <code>null</code>.<br/><br/>
	*
	* Example: to call the PL/SQL procedure <code>FOO(a IN, b IN OUT, c IN, d IN OUT, e OUT)</code>
	* this method must be invoked using:<br/> 
	* <code>param = {1,2,3,4} ; cols = 3 ; inout = {2,4}</code>.<br/><br/>
	*
	* @param targetDB the number of the database where to execute the query
	* @param label the logical name of the statement, only for logging purposes
	* @param query the SQL statement to be executed, not yet filled with parameters
	* @param param the <code>Object</code> array containing the input parameters
	* @param cols the number of output columns to be retrieved from the database
	* @param inout an int array specifying the index of the IN OUT parameters (null if there are no IN OUT params)
	* @return a <code>StringBuffer</code> array containing the selected columns
    */
    public StringBuffer[] getProcedure(int targetDB, String label, String query, Object[] param, int cols, int[] inout) {

		// Preparing variables
		status = ErrorMessage.DAO ;
		StringBuffer[] output = null ;
		if (boudoir.getConfiguration().BENCH) startQuery(label) ;

		// Choosing database's connection
		Connection connection = this.getConnection(targetDB) ;

		// Executing statement
		try {
			int bidir = 0 ;
			CallableStatement call = connection.prepareCall(query) ;
			// Setting IN parameters
			if (param != null) for (int i = 0 ; i < param.length ; i++) {
				int n = i + 1 ;
				setColumnData(call, n, param[i]) ;
			}
			// Setting IN OUT parameters (if any)
			if (inout != null) for ( ; bidir < inout.length ; bidir++) {
				call.registerOutParameter(inout[bidir], java.sql.Types.VARCHAR) ;
				//System.out.println("Registered input-output " + inout[bidir]) ;
			}
			// Setting "pure" OUT parameters (if any)
			for (int i = 0 ; i < (cols - bidir) ; i++) {
				int n = param.length + 1 + i ;
				call.registerOutParameter(n, java.sql.Types.VARCHAR) ;
				//System.out.println("Registered output " + n) ;
			}
	  		call.execute() ;
			//System.out.println("Executed call: " + query) ;

			// Reading output data
			status = ErrorMessage.OK ;
			output = new StringBuffer[cols] ;
			// Retrieving IN OUT parameters (if any)
			boolean found = false ;
			if (inout != null) for (int i = 0 ; i < inout.length ; i++) {
				String data = call.getString(inout[i]) ;
				if (data != null) {
					found = true ;
					output[i] = new StringBuffer(data) ;
				}
				if (!found) status = ErrorMessage.NO_RECORD ;
			}
			// Retrieving "pure" OUT parameters (if any)
			found = false ;
			for (int i = 0 ; i < (cols - bidir); i++) {
				String data = call.getString(param.length + i + 1) ;
				if (data != null) {
					found = true ;
					output[i] = new StringBuffer(data) ;
				}
				if (!found) status = ErrorMessage.NO_RECORD ;
			}
			// Eventually logging result on STDOUT
			if (boudoir.getConfiguration().VERBOSE) {
				System.out.print(leadingName + label) ;
				for (int i = 0 ; i < cols ; i++) if (output[i] != null) System.out.print(output[i].toString() + " ") ;
				System.out.println() ;
			}
			call.close() ;

		} catch(SQLException e) {
			status = ErrorMessage.SQL ;
			logger.logSevere(leadingName + label + QUERY_NOK + query + "\n" + e.getMessage()) ;
	  		if (boudoir.getConfiguration().DEBUG) e.printStackTrace() ;
	  		try 					{ 	if (connection != null) connection.rollback() ;			  			}
	  		catch(SQLException rbe) { 	System.err.println(ROLLBACK_NOK + rbe.getMessage()) ;				}
		} catch(Exception e) {
			status = ErrorMessage.DAO ;
			logger.logSevere(leadingName + label + ERROR + query + "\n" + e.getMessage()) ;
	  		if (boudoir.getConfiguration().DEBUG) e.printStackTrace() ;
		}

		// Optional: logging results
		if (boudoir.getConfiguration().BENCH) 	endQuery(label) ;
		if (boudoir.getConfiguration().ANALYZE) collectStatus(status) ;
		if (status == ErrorMessage.NO_RECORD) 	logger.logWarning(leadingName + label + boudoir.getMessages().getMessage(status)) ;

		return output ;
	}

// ************************************************************************************************************
// ************************************			PREPARED STATEMENTS		***************************************
// ************************************************************************************************************

	/**
	* Executes a <code>PreparedStatement</code> to the target database, returning a <strong>single record</strong>
	* described by a <code>StringBuffer</code> array. The length of the array is defined by the <code>cols</code>
	* input parameter, i.e. the number of columns to be retrieved from the database.<br/><br/>
	*
	* <strong>Remark</strong>: if the database rely on the mock connection pooling, this method will
	* not close the prepared statement, in order to leave to the <code>Database</code> class the choice
	* to eventually close or cache the prepared statement (after usage).<br/>
	*
	* @param pstmt the <code>PreparedStatement</code> to be executed, already filled with parameters
	* @param label the logical name of the statement, only for logging purposes
	* @param cols the number of columns to be retrieved from the database
	* @return a <code>StringBuffer</code> array containing the selected columns
    */
    public StringBuffer[] getPreparedRecord(PreparedStatement pstmt, String label, int cols) {

		// Preparing variables
		status = ErrorMessage.DAO ;
		StringBuffer[] output = null ;
		if (boudoir.getConfiguration().BENCH) startQuery(label) ;

		// Executing statement
		try {
	  		ResultSet result = pstmt.executeQuery() ;
			//System.out.println("PreparedStatement: " + pstmt) ;

			// Reading data only if we find a valid result set
			if ( result.isBeforeFirst() ) {
				output = new StringBuffer[cols] ;
				status = ErrorMessage.OK ;
				result.next() ;
				for (int i = 0 ; i < cols ; i++) {
					String data = result.getString(i + 1) ;
					if (data != null) output[i] = new StringBuffer(data) ;
				}
				if (boudoir.getConfiguration().VERBOSE) {
					System.out.print(leadingName + label) ;
					for (int i = 0 ; i < cols ; i++) if (output[i] != null) System.out.print(output[i].toString() + " ") ;
					System.out.println() ;
				}
			}
			else status = ErrorMessage.NO_RECORD ;
			// If we don't use the mocked pool we must close the statement to avoid the Oracle
			// error "maximum opened cursors exceeded"
			if ( !boudoir.getDatabase().usesMockedPool() ) pstmt.close() ;
			else result.close() ;

		} catch(SQLException e) {
			status = ErrorMessage.SQL ;
			logger.logSevere(leadingName + label + QUERY_NOK + "\n" + e.getMessage()) ;
	  		if (boudoir.getConfiguration().DEBUG) e.printStackTrace() ;
		}

		// Optional: logging results
		if (boudoir.getConfiguration().BENCH) 	endQuery(label) ;
		if (boudoir.getConfiguration().ANALYZE) collectStatus(status) ;
		if (status == ErrorMessage.NO_RECORD) 	logger.logWarning(leadingName + label + boudoir.getMessages().getMessage(status)) ;

		return output ;
	}

	/**
	* Executes a <code>PreparedStatement</code> to the target database, returning a <strong>set of records</strong>
	* described by an array which elements are others <code>StringBuffer</code> arrays, where:<br/><br/>
	*
	* <code>StringBuffer[n][i]</code><br/><br/>
	*
	* contains the i-th column of the n-th record.<br/><br/>
	*
	* <strong>Remark</strong>: if the database relies on the mock connection pooling, this method will
	* not close the prepared statement, in order to leave to the <code>Database</code> class the choice
	* to eventually close or cache the prepared statement (after usage).<br/>
	*
	* @param pstmt the <code>PreparedStatement</code> to be executed, already filled with parameters
	* @param label the logical name of the statement, only for logging purposes
	* @param cols the number of columns to be retrieved from the database
	* @return a <code>StringBuffer</code> double array containing all the matching records
    */
    public StringBuffer[][] getPreparedRecordSet(PreparedStatement pstmt, String label, int cols) {

		// Preparing variables
		status = ErrorMessage.DAO ;
		StringBuffer[] record = null ;
		StringBuffer[][] output = null ;
		if (boudoir.getConfiguration().BENCH) startQuery(label) ;

		// Executing statement
		try {
	  		ResultSet result = pstmt.executeQuery() ;
			//System.out.println("PreparedStatement: " + pstmt) ;

			// Reading data only if we find a valid result set
			if ( result.isBeforeFirst() ) {
				status = ErrorMessage.OK ;
				ArrayList list = new ArrayList() ;
				while (result.next()) {
					record = new StringBuffer[cols] ;
					for (int i = 0 ; i < cols ; i++) {
						String data = result.getString(i + 1) ;
						if (data != null) record[i] = new StringBuffer(data) ;
						//if (i == 5) logInfo("Column:#" + record[i] + '#') ;
					}
					list.add(record) ;
				}
				int size = list.size() ;
				output = (StringBuffer[][]) list.toArray( new StringBuffer[size][cols] ) ;
				if (boudoir.getConfiguration().VERBOSE) System.out.println(leadingName + label + "found " + size + " records") ;
				// Eventually printing the results: KEEP FOR TESTING OR DEBUG
				//for (int n = 0 ; n < size ; n++) { for (int i = 0 ; i < cols ; i++) if (output[n][i] != null) System.out.print(output[n][i].toString() + " ") ; System.out.println() ; }
			}
			else status = ErrorMessage.NO_RECORD ;
			// If we don't use the mocked pool we must close the statement to avoid the Oracle
			// error "maximum opened cursors exceeded"
			if ( !boudoir.getDatabase().usesMockedPool() ) pstmt.close() ;
			else result.close() ;

		} catch(SQLException e) {
			status = ErrorMessage.SQL ;
			logger.logSevere(leadingName + label + QUERY_NOK + "\n" + e.getMessage()) ;
	  		if (boudoir.getConfiguration().DEBUG) e.printStackTrace() ;
		}

		// Optional: logging results
		if (boudoir.getConfiguration().BENCH) 	endQuery(label) ;
		if (boudoir.getConfiguration().ANALYZE) collectStatus(status) ;
		if (status == ErrorMessage.NO_RECORD) 	logger.logWarning(leadingName + label + boudoir.getMessages().getMessage(status)) ;

		return output ;
	}

// **********************************************************************************************************
// **************************************		UPDATE QUERIES			*************************************
// **********************************************************************************************************

	/**
	* Execute the general UPDATE query passed in input. This method can be used any time the update is
	* performed using normal statements.<br/>
	*
	* @param targetDB the number of the database where to execute the query
	* @param label the logical name of the statement, only for logging purposes
	* @param query the query that must be executed
	* @return the number of rows inserted in the database
	*/
	public int executeUpdate(int targetDB, String label, String query) {

		int 		rows 		= 0 ;
		Statement 	stmt		= null ;
		Connection	connection	= this.getConnection(targetDB) ;

		try {
	 		stmt = connection.createStatement() ;
	  		rows = stmt.executeUpdate(query) ;
	  		if (rows == 0) 	status = ErrorMessage.NO_RECORD ;
			else 			status = ErrorMessage.OK ;
			stmt.close() ;
			connection.commit() ;
		} catch(SQLException e) {
			status = ErrorMessage.SQL ;
			logger.logSevere(leadingName + label + QUERY_NOK + query + "\n" + e.getMessage()) ;
	  		if (boudoir.getConfiguration().DEBUG) e.printStackTrace() ;
	  		try 					{ 	if (connection != null) connection.rollback() ;			  				}
	  		catch(SQLException rbe) { 	System.err.println(ROLLBACK_NOK + rbe.getMessage()) ;					}
		} catch(Exception e) {
			status = ErrorMessage.DAO ;
			logger.logSevere(leadingName + label + ERROR + query + "\n" + e.getMessage()) ;
	  		if (boudoir.getConfiguration().DEBUG) e.printStackTrace() ;
		}

		// Optional: logging results
		if (boudoir.getConfiguration().BENCH) 	endQuery(label) ;
		if (boudoir.getConfiguration().ANALYZE) collectStatus(status) ;
		if (status == ErrorMessage.NO_RECORD) 	logger.logWarning(leadingName + label + boudoir.getMessages().getMessage(status)) ;

	  	return rows ;
	}

	/**
	* Executes a <code>PreparedStatement</code> to the target database, returning an integer number describing
	* the number of affected rows.
	*
	* @param pstmt the <code>PreparedStatement</code> to be executed, already filled with parameters
	* @param label the logical name of the statement, only for logging purposes
	* @return the number of rows inserted in the database
    */
    public int executeUpdate(PreparedStatement pstmt, String label) {

		// Preparing variables
		int rows = 0 ;
		status = ErrorMessage.DAO ;
		if (boudoir.getConfiguration().BENCH) startQuery(label) ;

		// Executing statement
		try {
			//System.out.println(leadingName + "PreparedStatement: " + pstmt) ;
	  		rows = pstmt.executeUpdate()  ;
	  		if (rows == 0) 	status = ErrorMessage.NO_RECORD ;
			else 			status = ErrorMessage.OK ;
			pstmt.close() ;
			if (boudoir.getConfiguration().VERBOSE) System.out.println(leadingName + label + "update " + rows + " rows") ;
		} catch(SQLException e) {
			status = ErrorMessage.SQL ;
			logger.logSevere(leadingName + label + QUERY_NOK + "\n" + e.getMessage()) ;
	  		if (boudoir.getConfiguration().DEBUG) e.printStackTrace() ;
		}

		// Optional: logging results
		if (boudoir.getConfiguration().BENCH) 	endQuery(label) ;
		if (boudoir.getConfiguration().ANALYZE) collectStatus(status) ;
		if (status == ErrorMessage.NO_RECORD) 	logger.logWarning(leadingName + label + boudoir.getMessages().getMessage(status)) ;

		return rows ;
	}

// **********************************************************************************************************
// **************************************		ACCESSORS			*****************************************
// **********************************************************************************************************

	/**
	* Return an int code describing the status of the DAO after execution of the last SQL-oriented label.
	*
	* @return an integer code describing the status of the DAO after execution of the last SQL statement
	*/
	public int getStatus() 				{		return this.status ;										}
	

// **********************************************************************************************************
// **************************************		AUTO-READ-DATA		*****************************************
// **********************************************************************************************************
	
	/**
	* Fetch the column data automatically. This is achieved by inspecting the <code>ResultSetMetaData</code> 
	* for the input indexed column, and returning a formatted <code>StringBuffer</code> object containing 
	* the column's data. This is just a shortcut to invocation of following methods:<br/><br/>
	* 
	* <code><pre>
	* int colType = getColumnType(result, n) ;<br/>
	* return getColumnData(result, n, colType) ;<br/>
	* </pre></code>
	*
	* @param rs the <code>ResultSet</code> as returned by the database
	* @param col the index of the column to be inspected
	* @return a <code>String</code> representation of the column's content
	* @see #getColumnType(ResultSet rs, int col)
	* @see #getColumnData(ResultSet rs, int col, int type) throws SQLException
	*/
	public String fetchColumnData(ResultSet rs, int col) throws SQLException  {
	
		int colType = getColumnType(rs, col) ;
		return getColumnData(rs, col, colType) ;
	}
	
	/**
	* Inspect the <code>ResultSetMetaData</code> for the input indexed column, and return a formatted
	* <code>StringBuffer</code> object containing the column's metadata.
	*
	* @param rs the <code>ResultSet</code> as returned by the database
	* @param col the index of the column to be inspected
	* @return the integer reference to the <code>java.sql.Types</code> column type
	*/
	public int getColumnType(ResultSet rs, int col) {
		
    	try {
    		ResultSetMetaData metadata = rs.getMetaData() ;
    		int type = metadata.getColumnType(col) ;
    		return type ;
		} catch(Throwable e) {
	  		System.err.println("getColumnType - Throwable: " + e.getMessage()) ;
			e.printStackTrace() ;
			return java.sql.Types.VARCHAR ;
		}
	}
	
	/**
	* Return an <code>String</code> containing data of the input indexed column,
	* read using the proper method for the <code>java.sql.Types</code> column type.  
	*
	* @param rs the <code>ResultSet</code> as returned by the database
	* @param col the index of the column to be read
	* @param type the <code>java.sql.Types</code> describing the column
	* @return a <code>String</code> representation of the column's content
	* @throws SQLException 
	*/
	public String getColumnData(ResultSet rs, int col, int type) throws SQLException  {
		
		Object data = null ;
		switch(type) {
			case java.sql.Types.BOOLEAN		:	data = rs.getBoolean	(col) ;		break ;
			case java.sql.Types.DATE		:	data = rs.getDate		(col) ;		break ;			
			case java.sql.Types.DOUBLE		:	data = rs.getDouble		(col) ;		break ;			
			case java.sql.Types.FLOAT		:	data = rs.getFloat		(col) ;		break ;			
			case java.sql.Types.INTEGER		:	data = rs.getInt		(col) ;		break ;
			case java.sql.Types.TIME		:	data = rs.getTime		(col) ;		break ;			
			case java.sql.Types.TIMESTAMP	:	data = rs.getTimestamp	(col) ;		break ;			
			case java.sql.Types.VARCHAR		:	data = rs.getString		(col) ;		break ;
			default							:	data = rs.getString		(col) ;
		}
		//System.out.println(leadingName + "getColumnData for java.sql.Types = " + type) ;
		if (data != null) 	return data.toString() ;
		else				return null ;
	}
	
	/**
	* Return an <code>String</code> containing data of the input indexed column,
	* read using the proper method for the <code>java.sql.Types</code> column type.  
	*
	* @param call the <code>CallableStatement</code> executed to the database
	* @param col the index of the column to be read
	* @param type the <code>java.sql.Types</code> describing the column
	* @return a <code>String</code> representation of the column's content
	* @throws SQLException 
	*/
	public String getColumnData(CallableStatement call, int col, int type) throws SQLException  {
		
		Object data = null ;
		switch(type) {
			case java.sql.Types.BOOLEAN		:	data = call.getBoolean		(col) ;		break ;
			case java.sql.Types.DATE		:	data = call.getDate			(col) ;		break ;
			case java.sql.Types.DOUBLE		:	data = call.getDouble		(col) ;		break ;			
			case java.sql.Types.FLOAT		:	data = call.getFloat		(col) ;		break ;			
			case java.sql.Types.INTEGER		:	data = call.getInt			(col) ;		break ;
			case java.sql.Types.TIME		:	data = call.getTime			(col) ;		break ;			
			case java.sql.Types.TIMESTAMP	:	data = call.getTimestamp	(col) ;		break ;			
			case java.sql.Types.VARCHAR		:	data = call.getString		(col) ;		break ;
			default							:	data = call.getString		(col) ;
		}
		//System.out.println(leadingName + "getColumnData for java.sql.Types = " + type) ;
		if (data != null) 	return data.toString() ;
		else				return null ;
	}
	
// **********************************************************************************************************
// **************************************		AUTO-SET-DATA		*****************************************
// **********************************************************************************************************
	
	/**
	* Set an object parameter into the input <code>CallableStatement</code> object, buy introspecting 
	* the parameter type and invoking the proper <code>setXXX()</code> from the <code>java.sql</code> API.  
	*
	* @param call the <code>CallableStatement</code> to be prepared
	* @param n the index of the column to be set
	* @param an <code>Object</code> representation of the parameter value
	* @throws SQLException 
	*/
	public void setColumnData(CallableStatement call, int n, Object param) throws SQLException  {

			if 	(param instanceof Double				)	call.setDouble		(n, (Double) param					) ;		
		else if (param instanceof Float					)	call.setFloat		(n, (Float) param					) ;
		else if	(param instanceof Integer				)	call.setInt			(n, (Integer) param					) ;		
		else if	(param instanceof Long					)	call.setLong		(n, (Long) param					) ;		
		else if (param instanceof String				)	call.setString		(n, (String) param					) ;
		else if (param instanceof StringBuffer			)	call.setString		(n, param.toString()				) ;
		else if (param instanceof java.sql.Date			)	call.setDate		(n, (java.sql.Date) param			) ;
		else if (param instanceof java.sql.Time			)	call.setTime		(n, (java.sql.Time) param			) ;
		else if (param instanceof java.sql.Timestamp	)	call.setTimestamp	(n, (java.sql.Timestamp) param		) ;
		else if (param == null							)	call.setString		(n, null							) ;
	}

// **********************************************************************************************************
// **************************************		UTILITIES			*****************************************
// **********************************************************************************************************

	/**
	* Adds double quotation marks to the input String to avoid SQL error. In example, if the input String is
	* like <em>Video's</em>, the returned String will be: <em>Video''s</em>.<br/>
	*
	* @param input the string that can contain single quote characters
	* @return the same string with all single quotes converted to double quotes
	*/
	public String doubleQuote(String input) {

		if ( (input == null) || (input.length() <=0) ) return input ;
		String result = input.replaceAll(QUOTE, DOUBLE_QUOTES) ;
		return result ;
	}

	/**
	* Execute a <code>doubleQuote()</code> method on the input string, then surround the result with
	* a pair of single quotes.<br/>
	* Example: a string like <em>Video's</em> will become <em>'Video''s'</em>.
	*
	* @param input the string that can contain single quote characters
	* @return the same string with all single quotes converted to double quotes, enclosed within quotes
	*/
	public String quote(String input) {

		return "'" + doubleQuote(input) + "'" ;
	}

	/**
	* Parse the string value of the input <code>String</code> into an integer value.
	* The <code>no_value</code> parameter represent the default integer value to be returned if
	* the parsing is not successful.
	*
	* @param string the input string object
	* @param no_value the integer value to be returned if conversion fails
	* @return the integer value of the input string (or the <code>no_value</code> number)
	*/
	protected int valueOfString(String string, int no_value) {

		int result = no_value ;
		try 				{	result = Integer.parseInt( string.trim() ) ;								}
		catch(Exception e) 	{ 	result = no_value ;															}
		return result ;
	}

	/**
	* Parse the string value of the input <code>String</code> into a timestamp value.
	*
	* @param string the input string object
	* @return the timestamp value of the input string
	*/
	protected java.sql.Timestamp valueOfTimestamp(String string) {

		java.sql.Timestamp result = null ;
		try {
			java.util.Date date = db_timestamp.parse( string.trim() ) ;
			long time = date.getTime() ;
			result = new java.sql.Timestamp(time) ;
		} catch(Exception e) {
			result = null ;
		}
		return result ;
	}

	/**
	* Helper method, used to compare StringBuffer contents.
	*
	* @param first a <code>StringBuffer</code> object
	* @param second another <code>StringBuffer</code> object
	* @return <code>true</code> if the two string object contains the same text
	*/
	public boolean compareSB(StringBuffer first, StringBuffer second) {

		String string = null ;
		boolean result = false ;
		if (first != null) string = first.toString() ;
		if (string != null) result = string.equalsIgnoreCase( second.toString() ) ;
		return result ;
	}
	
// **********************************************************************************************************
// **************************************		INTROSPECTION		*****************************************
// **********************************************************************************************************

	/**
	* This method can be invoked immediately after the <code>stmt.executeQuery(query)</code> call
	* to inspect the retrieved <code>ResultSet</code>. The obtained meta data are returned as a
	* <code>StringBuffer</code> object.
	*
	* @param rs the <code>ResultSet</code> as returned by the database
	* @return a <code>StringBuffer</code> object containing meta data information about the result set
	*/
    public StringBuffer inspectResultSet(ResultSet rs) {

		StringBuffer message = new StringBuffer() ;

		try {
			message.append("\nRESULT SET METADATA:\n"								) ;
			message.append("ResultSet object       : " + rs					+ "\n"	) ;
			message.append("ResultSet current row  : " + rs.getRow()		+ "\n"	) ;
			message.append("Is after last          : " + rs.isAfterLast()	+ "\n"	) ;
			message.append("Is before first        : " + rs.isBeforeFirst()	+ "\n"	) ;
			message.append("Is first               : " + rs.isFirst()		+ "\n"	) ;
			// Examining Fetch Direction
			int fd = rs.getFetchDirection() ;
			switch (fd) {
				case ResultSet.FETCH_UNKNOWN	: 	message.append("Fetch direction        : UNKNOWN") ; 	break ;
				case ResultSet.FETCH_FORWARD	: 	message.append("Fetch direction        : FORWARD") ; 	break ;
				case ResultSet.FETCH_REVERSE	: 	message.append("Fetch direction        : REVERSE") ; 	break ;
			}
			message.append("\n") ;
			// Examining Holdability
			int h = rs.getConcurrency() ;
			switch (h) {
				case ResultSet.CONCUR_READ_ONLY	:	message.append("Concurrency            : READ_ONLY") ; 	break ;
				case ResultSet.CONCUR_UPDATABLE	:	message.append("Concurrency            : UPDATABLE") ; 	break ;
			}
			message.append("\n") ;
			// Listing metadata
			ResultSetMetaData metadata = rs.getMetaData() ;
			message.append("Metadata - Columns     : " + metadata.getColumnCount()) ;
			message.append("\n") ;

		} catch(Throwable e) {
	  		System.err.println("Test query - Throwable: " + e.getMessage()) ;
			e.printStackTrace() ;
		}

		return message ;
	}

    /**
     * Determines the number of rows in a <code>ResultSet</code>. Upon exit, if the cursor was not
     * currently on a row, it is just before the first row in the result set (a call to
     * {@link ResultSet#next()} will go to the first row).
     *
     * @param set The <code>ResultSet</code> to check (must be scrollable).
     * @return The number of rows.
     * @throws SQLException If the <code>ResultSet</code> is not scrollable.
     */
    public static int getRowCount(ResultSet set) throws SQLException  {

       int rowCount;
       int currentRow = set.getRow();            // Get current row
       rowCount = set.last() ? set.getRow() : 0; // Determine number of rows
       if (currentRow == 0)                      // If there was no current row
          set.beforeFirst();                     // We want next() to go to first row
       else                                      // If there WAS a current row
          set.absolute(currentRow);              // Restore it
       return rowCount;
    }
    
	/**
	* Inspect the <code>ResultSetMetaData</code> for the input indexed column, and return a formatted
	* <code>StringBuffer</code> object containing the column's metadata.
	*
	* @param rs the <code>ResultSet</code> as returned by the database
	* @param col the index of the column to be inspected
	* @return a <code>StringBuffer</code> object containing meta data information about the column
	*/
    public StringBuffer inspectColumn(ResultSet rs, int col) {
    	
    	StringBuffer message = new StringBuffer() ;
    	
    	try {
    		ResultSetMetaData metadata = rs.getMetaData() ;
    		message.append("ColumnLabel          : " + metadata.getColumnLabel		(col)	) ;
    		message.append("\n") ;
    		message.append("ColumnName           : " + metadata.getColumnName		(col)	) ;
    		message.append("\n") ;
    		message.append("ColumnClassName      : " + metadata.getColumnClassName	(col)	) ;
    		message.append("\n") ;
    		message.append("ColumnDisplaySize    : " + metadata.getColumnDisplaySize(col)	) ;
    		message.append("\n") ;
    		message.append("ColumnType           : " + metadata.getColumnType		(col)	) ;
    		message.append("\n") ;
    		message.append("ColumnTypeName       : " + metadata.getColumnTypeName	(col)	) ;
    		message.append("\n") ;
    		message.append("Precision            : " + metadata.getPrecision		(col)	) ;
    		message.append("\n") ;
    		message.append("Scale                : " + metadata.getScale			(col)	) ;
    		message.append("\n") ;
		} catch(Throwable e) {
	  		System.err.println("inspectColumn - Throwable: " + e.getMessage()) ;
			e.printStackTrace() ;
		}
    	
    	return message ;
    }
    
	/**
	* Method useful for debug and testing, used to verify the scope of the children classes.
    */
    public void finalize() {

		if (boudoir.getConfiguration().VERBOSE) System.out.println(leadingName + "Finalized object: " + this) ;
	}

} // End of class