/**
* Title: 		 ObjectPool
* Description:	 A general purposes manager for pooling of Java objects
* Copyright:     Copyright (c) 2013
* Company:       Morpheus Quality Technologies
* Creation date: 13-12-2013
* Last modified: 13-12-2013
* @author 		 Stefano Adriani
* @version 		 1.0
*/

package adriani.boudoir.db ;

import adriani.boudoir.*;
import adriani.boudoir.logging.WSLogger;
import adriani.boudoir.util.*;

import java.util.Hashtable ;
import java.util.Enumeration ;

/**
* Abstract class managing a generic pool of objects (usually threads or database connections).<br/>
* The class offers two hashtables: the <em>locked</em> hashtable and the <em>unlocked</em> hashtable.
* The <em>locked</em> hashtable stores a reference to the "busy" objects: if the objects are DB connections,
* then a "busy" object could be a connection already in use. The <em>unlocked</em> hashtable stores all
* the "free" objects (available).<br/>
* The normal life cycle of the objects should be:<br/><br/>
*
* 1. The object is instantiated by the <code>create()</code> abstract method.<br/>
* 2. The object is retrieved by calling the <code>checkOut()</code> method.<br/>
* 3. The object is released by calling the <code>checkIn()</code> method.<br/><br/>
*
* Implementation of the <code>create()</code> abstract method is left to the child class and should
* not be called directly: it's up to the <code>checkOut()</code> method to decide if it's necessary
* to create a new object (usually because the unlocked hashtable is empty) or if it's possible to
* return an already existing object.<br/>
*/
public abstract class ObjectPool {
	
	/** Reference to the top-level <code>Boudoir</code> instance.											*/
	public 		 			Boudoir 		boudoir ;

	/** The time of the last <code>checkOut()</code> call (or the creation time).							*/
	private 				long 			lastCheckOut ;

	/** Sleeping time for the cleanup thread (in seconds).													*/
	public 					int 			expirationTime ;

	/** Maximum allowed number of objects in the pool.														*/
	public 					int 			maximum ;

	/** Contains the objects of the pool already in use.													*/
	private 				Hashtable 		locked ;

	/** Contains the objects of the pool still available.													*/
	private 				Hashtable 		unlocked ;

	/** Helper class calling the <code>cleanUp()</code> method.												*/
	private 				CleanUpThread 	cleaner ;

	/** Boolean flag used only by this class to easy testing and debug of critical thread issues.			*/
	private 				boolean			verbose 		= false ;

	/** Retrieving the logger instance to be used for this class.											*/
	protected 		 		WSLogger		logger ;

	/** Beginning of each logging message produced by this class.											*/
	protected static		String 			leadingName 	= "ObjectPool - " ;
	
	/**
	* Standard constructor, bounding this class to a single <code>Boudoir</code> instance.
	* 
	* @param boudoir the <code>Boudoir</code> instance "owning" this object.
    */
    public ObjectPool(Boudoir boudoir) {

    	this.boudoir 	= boudoir ;
    	this.logger 	= boudoir.getLogger().getLogger(ObjectPool.class) ;
    }	

	/**
	* Constructor.
	*
	* @param expirationTime the sleep time for the <code>CleanUpThread</code>, in seconds
	*/
	public ObjectPool(Boudoir boudoir, int expirationTime) {
		
		this(boudoir) ;

		// Creating class attributes
		this.locked 		= new Hashtable() ;
		this.unlocked 		= new Hashtable() ;
		this.expirationTime = expirationTime ;
		this.lastCheckOut 	= System.currentTimeMillis() ;

		// Creating the cleaner thread
		this.cleaner = new CleanUpThread(this, expirationTime) ;
		this.cleaner.start() ;
	}
	
	/**
	* Called by the user when he finishes to use an object previously retrieved using <code>checkOut()</code>.
	* The object is removed from the <em>locked</em> table and put again in the <em>unlocked</em> table.
	*
	* @param object a reference to the object to check in
	*/
	@SuppressWarnings("unchecked")		// Avoid warning about: locked.put(object, new Long(now))
	synchronized void checkIn(Object object) {

		if (object != null) {
			locked.remove(object) ;
			if (verbose) System.out.println(leadingName + "Unlocked object: " + printObject(object)) ;
			unlocked.put(object, new Long( System.currentTimeMillis())) ;
			if (boudoir.getConfiguration().VERBOSE) System.out.println(leadingName + "Checked in object: " + printObject(object)) ;
		}
	}

	/**
	* Return a ready-to-use object to the user.<br/><br/>
	*
	* This method can be considered the inner core of the pool mechanism, which is:
	* first we look in the unlocked hash table, if we find a valid object we store such object
	* in the locked table and then we return the object. On the opposite, if there isn't any valid
	* object in the unlocked hash table, we check the maximum pool value: if we are below this threshold
	* we create a brand new object, using the <code>create()</code> method. If we have already reached the
	* maximum pool capacity we return <code>null</code> and log a severe message.
	*
	* @return object a reference to object just checked out
	*/
	@SuppressWarnings("unchecked")		// Avoid warning about: locked.put(object, new Long(now))
	synchronized Object checkOut() throws Exception {

		long now = System.currentTimeMillis() ;
		this.lastCheckOut = now ;

		Object object ;
		if (unlocked.size() > 0) {

			Enumeration e = unlocked.keys() ;

			while (e.hasMoreElements()) {

				object = e.nextElement() ;

				if (validate(object)) {
					// We lock the first valid object (in the JDBC case valid == connection not closed)
					unlocked.remove(object) ;
					locked.put(object, new Long(now)) ;
					if (boudoir.getConfiguration().VERBOSE) System.out.println(leadingName + "Checked out object: " + printObject(object)) ;
					return object ;
				} else {
					// If we find an invalid object, we discard it.
					unlocked.remove(object) ;
					expire(object) ;
					object = null ;
				}

			} // End of while

		} // End if

		// If we did not found a valid object in the unlocked list, we create a new one
		int howMany = locked.size() ;
		if (howMany < maximum) {
			object = create() ;
			locked.put(object, new Long(now)) ;
			if (boudoir.getConfiguration().VERBOSE) System.out.println(leadingName + "Created object: " + printObject(object)) ;
		} else {
			object = null ;
			logger.logSevere(leadingName + "Reached maximum pool capacity: " + howMany) ;
		}
		return object ;
	}

	/**
	* Removes from the <em>unlocked</em> list all the expired objects.<br/>
	* This method is invoked automatically by the cleaner thread (see <code>expirationTime</code>).<br/><br/>
	*
	* <strong>Note:</strong> after cleanup of the <em>unlocked</em> list, the method calls <code>System.gc()</code>
	* to delete all the memory footsteps.
	*
	* @see CleanUpThread
	*/
	synchronized void cleanUp() {

		// Defining variables
		Object object ;
		long lastPut 	= 0 ;		// Time of the last checkin() or checkout() operation
		int exp_lock 	= 0 ;		// Number of expired objects from the LOCKED table
		int exp_unlock 	= 0 ;		// Number of expired objects from the UNLOCKED table

		// Preparing time parameters
		long delta 	= expirationTime*1000 ;
		long now 	= System.currentTimeMillis() ;

		// Retrieving UNLOCKED objects
		int present = 0 ;
		Enumeration free = unlocked.keys() ;

		// Looking for the old objects in the UNLOCKED table
		while ( free.hasMoreElements() ) {

			present++ ;
			object = free.nextElement() ;
			lastPut = ((Long) unlocked.get(object)).longValue() ;
			if (verbose) System.out.println(leadingName + "Counting unlocked elements: " + printObject(object) ) ;

			// Expiring objects
			if ( (now - lastPut) > delta ) {
				unlocked.remove(object) ;
				expire(object) ;
				exp_unlock++ ;
				java.util.Date usage = new java.util.Date(lastPut) ;
				logger.logInfo(leadingName + "Expired unlocked " + object + " last usage: " + usage) ;
				object = null ;
			}

		} // End while

		// Counting the objects in the LOCKED table
		int busy = locked.size() ;
		Enumeration used = locked.keys() ;
		if (verbose) while ( used.hasMoreElements() ) {
			object = used.nextElement() ;
			System.out.println(leadingName + "Counting locked elements: " + printObject(object) ) ;
		} // End while

		// Logging pool status
		StringBuffer message = new StringBuffer(leadingName) ;
		message.append("Expired " + exp_unlock + " unlocked objects [left: " + (present - exp_unlock) + "] - ") ;
		message.append("Currently locked: " + busy + " objects") ;
		logger.logInfo(message.toString()) ;

		// Cleaning up the memory footsteps
		System.gc() ;
		if (verbose) System.out.println(leadingName + "cleaned garbage") ;
	}

	/**
	* Terminates the cleaner thread.
	*
	* @see CleanUpThread
	*/
	synchronized void shutdown() {

		try {
			if (cleaner.isAlive()) cleaner.running = false ;
			cleaner.join(500) ;
		} catch (InterruptedException ix) {
			logger.logSevere(leadingName + "Cleaning thread terminated !") ;
		}
	}

// **********************************************************************************************************
// ************************					MUTATORS			*********************************************
// **********************************************************************************************************

	/**
	* Mutator method.
	*/
	public void setPoolSize(int max) 			{		this.maximum = max ;								}

// **********************************************************************************************************
// ************************					ACCESSORS			*********************************************
// **********************************************************************************************************

	/**
	* Accessor method.
	*/
	public int getPoolSize(int max) 			{		return this.maximum  ;								}

// **********************************************************************************************************
// ************************				ABSTRACT METHODS		*********************************************
// **********************************************************************************************************

	/**
	* Implemented by the child class (Example: see the <code>JDBCConnectionPool</code> class).
	*
	* @return object a reference to object just checked out
	*/
	abstract Object create() throws Exception ;

	/**
	* Implemented by the child class (Example: see the <code>JDBCConnectionPool</code> class).
	*
	* @param object a reference to object to validate
	*/
	abstract boolean validate(Object object) ;

	/**
	* Implemented by the child class (Example: see the <code>JDBCConnectionPool</code> class).
	*
	* @param object a reference to object to destroy
	*/
	abstract void expire(Object object) ;

// **********************************************************************************************************
// ************************					UTILITIES				*****************************************
// **********************************************************************************************************

	/**
	* Print the serialize reference number of the input object.

	* @param object a generic Java object, e.g. MyObject@1bd4722
	* @return the serial number of the input object, e.g. 1bd4722
	*/
	public String printObject(Object object) {

		String fullName = object.toString() ;
		String[] result = fullName.split("@") ;
		if (result == null) return "unknown" ;
		else return result[1] ;
	}

	/**
	* Return a text message describing the status of this pool: number of locked objects, number of unlocked
	* objects and the thread status.
	*
	* @return a <code>StringBuffer</code> containing the printable message
    */
    public StringBuffer getPoolData() {

		// Getting pool status
		int used = locked.size() ;
		int free = unlocked.size() ;
		java.util.Date last = new java.util.Date(lastCheckOut) ;

		// Preparing text message
		StringBuffer message = new StringBuffer("\n") ;
		message.append("Locked               : " + used 				+ " objects\n"	) ;
		message.append("Unlocked             : " + free 				+ " objects\n"	) ;
		message.append("Sleep Time           : " + expirationTime 		+ " s\n"		) ;
		message.append("Pool Capacity        : " + maximum 				+ " objects\n"	) ;
		message.append("Cleaner thread alive ? " + cleaner.isAlive() 	+ "\n"			) ;
		message.append("Last Check Out       : " + last 				+ "\n"			) ;
		return message ;
	}

	/**
	* Method useful for debug and testing, the test the scope of the pool objects.
	* This method should be called only when the application server is shut down, since this class
	* is stored as a static object in a static container class (see for example the <code>Database</code> class).
    */
    public void finalize() {

		logger.logInfo(leadingName + "Finalized ObjectPool instance: " + this) ;
	}

} // End of class

// **********************************************************************************************************
// ************************					HELPER CLASS			*****************************************
// **********************************************************************************************************

/**
* Helper thread taking care to call the <code>cleanUp()</code> method of the main class (<code>ObjectPool</code>).<br/>
*
* @see ObjectPool
*/
class CleanUpThread extends Thread {

	/** The boolean flag used to check condition of the endless loop in the <code>run()</code> method.			*/
	protected 					boolean 				running ;

	/** The parent class containing this helper class.															*/
	private 					ObjectPool 				pool ;

	/** Sleeping time in seconds for this thread																*/
	private 					long 					sleepTime ;

  	/** Beginning of each logging message produced by this class												*/
	protected static 			String 					leadingName ;

	/**
	* Constructor.
	*
	* @param pool the <code>ObjectPool</code> target of the <code>cleanUp()</code> method
	* @param sleepTime the sleep time for this thread, in seconds
	*/
	CleanUpThread(ObjectPool pool, int sleepTime) {

		this.running 	= true ;
		this.pool 		= pool ;
		this.sleepTime 	= sleepTime*1000 ;

		// We have one thread of this type for each object pool !
		String name = "CleanUpThread:" + this.getId() ;
		this.setName(name) ;
		this.leadingName = name + " - ";
	}

	/**
	* This methods runs forever, calling the <code>cleanUp()</code> method every <em>sleepTime</em> seconds.<br/>
	* To terminate this thread the user should invoke the <code>shutdown()</code> method.
	*/
	public void run() {

		// Printing verbose details
		if (pool.boudoir.getConfiguration().VERBOSE) {
			int minutes = (int) (sleepTime/60000) ;
			System.out.println(leadingName + "Started with timeout: " + minutes + " minutes") ;
		}

		// Getting logger from the ObjectPool class
		WSLogger my_log = pool.logger ;

		// Main cycle
		while (this.running) {

			// Sleeping for all the sleepTime cycle
			try 							{	sleep(sleepTime) ; 												}
			catch (InterruptedException e)	{ 	my_log.logSevere(leadingName + "Interrupted while sleeping") ; 	}
			// Cleaning the pool
			try 							{	pool.cleanUp() ; 												}
			catch (Exception e)				{	my_log.logSevere(leadingName + "cleanUp() error: " + e) ;		}
			if (pool.boudoir.getConfiguration().VERBOSE) System.out.println(leadingName + "Checked objects pool") ;

		} // End of while
	}

} // End of helper class