/**
* Title: 		 ExternalLogger
* Description:	 The initializer class for the Log4j library
* Copyright:     Copyright (c) 2013
* Company:       Morpheus Quality Technologies
* Creation date: 06-12-2013
* Last modified: 06-12-2013
* @author 		 Stefano Adriani
* @version 		 1.0
*/

package adriani.boudoir.logging ;

import adriani.boudoir.*;

import java.net.URL ;
import java.util.* ;
import org.apache.log4j.* ;
import org.apache.log4j.xml.DOMConfigurator ;

/**
* The class managing the logging mechanism when using the Log4j library. Notice that in the Log4j paradigm
* this class is not necessary, since any class should contain a class attribute like the following:<br/><br/>
*
* <code>protected static Logger logger = Logger.getLogger(MyApp.class) ;</code><br/><br/>
*
* This project needs this class only to ensure compatibility with the logging system offered by the
* <code>SystemLogger</code> class, in this way it's possible to switch between the two logging system
* simply changing the <code>logger.type</code> property in the service configuration file.
*
* @see SystemLogger
*/
public class ExternalLogger implements LoggerManager {
	
	/** Reference to the top-level <code>Boudoir</code> instance.											*/
	public 		 			Boudoir 			boudoir ;

	/** The absolute log path retrieved from the configuration file.										*/
	protected 				String 				loggerPath ;

  	/** The rotation date pattern set for this logger.														*/
	protected 				String 				datePattern ;

  	/** Beginning of each logging message printed on the STDOUT.											*/
	protected static 		String 				leadingName 		= "ExternalLogger - " ;
	
	/**
	* Standard constructor, bounding this class to a single <code>Boudoir</code> instance.
    */
    public ExternalLogger(Boudoir boudoir) {

    	this.boudoir = boudoir ;
    	
    	if (boudoir == null) {
    		this.boudoir = new Boudoir() ;
    		System.err.println(leadingName + "ERROR: using default configuration") ;
    	}
	}

	/**
	* This method forces the Logger to load configuration from the <code>log4j.properties</code> file.<br/>
	* Such file is searched first in the classpath (e.g. in the Axis2/WEB-INF directory), if not found is 
	* searched within the AAR file (embedded configuration). If the properties file is still not found, 
	* this method set the basic configuration (Log4j default) for all the Log4j logger objects.
	*
	* @param path not used (required only to implement the <code>LoggerManager</code> interface)
	* @param logLevel not used (required only to implement the <code>LoggerManager</code> interface)
	* @param rotate not used (required only to implement the <code>LoggerManager</code> interface)
	* @param mode not used (required only to implement the <code>LoggerManager</code> interface)
	*/
	public void init(String path, String logLevel, int rotate, boolean mode) {
		
		try {
			// 1st: looking for EXTERNAL Log4j configuration (e.g. in the Axis2/WEB-INF directory)			
			this.getExternalConfiguration() ;
		} catch(MissingResourceException mre) {	
			// 2nd: if not found we search the INTERNAL configuration (legacy file embedded in the AAR package)
			try {
				String 			legacyPath 	= boudoir.getConfiguration().LOG_PREFIX + boudoir.getConfiguration().LOG_CONFIG ;			
				ResourceBundle 	bundle 		= ResourceBundle.getBundle(legacyPath) ;
				Properties 		properties 	= getProperties(bundle) ;
				PropertyConfigurator.configure(properties) ;
				if (boudoir.getConfiguration().DEBUG) System.out.println(leadingName + "Using Log4j configuration: " + legacyPath + ".properties (internal)") ;
			} catch(Exception e) {
				// 3rd: last chance: using Log4j basic configuration
				BasicConfigurator.configure() ;
				if (boudoir.getConfiguration().DEBUG) {
					System.out.println(leadingName + "Log4j configuration: " + boudoir.getConfiguration().LOG_CONFIG + ".properties NOT FOUND!") ;
					System.out.println(leadingName + "Using Log4j default configuration") ;
				}
			}
		}
	}
	
	/**
	* This method tries to load the Log4j configuration from an "external" file, meaning that the Log4j 
	* configuration file is NOT embedded into the AAR file, but it's located in another place, usually
	* belonging to the classloader's classpath (e.g. the Axis2/WEB-INF directory).<br/>
	* Such "external" configuration file is searched first with the <code>properties</code> extension:
	* if not found the <code>XML</code> extension is searched. If none of these files is found,
	* a <code>MissingResourceException</code> is thrown.
	*
	* @throws MissingResourceException if no Log4j configuration file (external) is not found
	*/
	public void getExternalConfiguration() throws MissingResourceException {

		try {
			// 1. Looking for external PROPERTIES Log4j configuration (e.g. in the Axis2/WEB-INF directory) 
			ResourceBundle 	bundle 		= ResourceBundle.getBundle(boudoir.getConfiguration().LOG_CONFIG) ;
			Properties 		properties 	= getProperties(bundle) ;
			PropertyConfigurator.configure(properties) ;
			if (boudoir.getConfiguration().DEBUG) System.out.println(leadingName + "Using Log4j configuration: " + boudoir.getConfiguration().LOG_CONFIG + ".properties (external)") ;
		} catch(MissingResourceException mre) {
			// 2. Looking for external XML Log4j configuration (e.g. in the Axis2/WEB-INF directory)			
			if (boudoir.getConfiguration().DEBUG) System.out.println(leadingName + "External " + boudoir.getConfiguration().LOG_CONFIG + ".properties not found") ;
			String 	xmlName = boudoir.getConfiguration().LOG_CONFIG + ".xml" ;
			URL 	url 	= Thread.currentThread().getContextClassLoader().getResource(xmlName) ;
			if (url != null) {
				String path = url.getPath() ;
				DOMConfigurator.configure(path);
			} else {
				if (boudoir.getConfiguration().DEBUG) System.out.println(leadingName + "External " + xmlName + " not found") ;				
				throw new MissingResourceException(mre.getMessage(), boudoir.getConfiguration().LOG_CONFIG, "none") ;
			}
			if (boudoir.getConfiguration().DEBUG) System.out.println(leadingName + "Using Log4j configuration: " + xmlName + " (external)") ;			
		}
	}
	
	/**
	* NOT USED AT THE MOMENT (place holder).
	*
	* @param path the path reference for the log file
	* @param level the logging minimal level as a Logger.Level value (INFO, WARNING or SEVERE)
	* @param mode if <code>true</code> creates the standard log file, if <code>false</code> creates the
	* benchmark analysis file
	* @see SystemLogger#getLogFile(String optionalPath, boolean mode)
	*/
	public void createLogger(String path, java.util.logging.Level level, boolean mode) {

		System.err.println(leadingName + "Method not implemented: createLogger()") ;
	}

	/**
	* NOT USED AT THE MOMENT (place holder).
	*/
	public void rotate() {

		System.err.println(leadingName + "Method not implemented: rotate()") ;
	}

	/**
	* NOT USED AT THE MOMENT (place holder).
	*/
	public boolean close() {

		System.err.println(leadingName + "Method not implemented: close()") ;
		return false ;
	}

// **********************************************************************************************************
// **********************************			ACCESSORS		*********************************************
// **********************************************************************************************************

	/**
	* Return the current <code>Boudoir</code> instance owning this logger. 
	*
	* @return the current <code>Boudoir</code> instance owning this logger
	*/
	public Boudoir getBoudoir() 				{		return boudoir ;									}
	
	/**
	* Return the <code>parentDir</code> path used as base directory to locate the log file.<br/>
	* Usually this parameter is retrieved from the <em>RootLogger</em> reference, because the method
	* <code>root.getAllAppenders()</code> returns only the EXPLICIT appenders, and not the appenders
	* obtained from a parent logger. In this case we do not use the <em>RootLogger</em>, because we
	* assume the Log4j configuration file is something like:<br/>
	*
	* <pre><code>
	* log4j.additivity.boudoir          = false
	* log4j.additivity.boudoir.db       = false
	* log4j.additivity.boudoir.service  = false
	* log4j.logger.boudoir              = INFO, stdout, myfile
	* log4j.logger.boudoir.db           = INFO, stdout, myfile
	* log4j.logger.boudoir.service      = INFO, stdout, myfile
	* </code></pre><br/><br/>
	*
	* Hence each logger object has his own explicit reference to the output file (<em>myfile</em>), for this
	* reason we can use directly <code>smartLogger.getAllAppenders()</code>.
	*
	* @param logger a valid org.apache.log4j.Logger instance
	* @return the current parent directory for logging on the file system
	*/
	public String getParentDir(Object logger) {

		if (loggerPath == null) {

			Enumeration e = null ;
			Logger root = (Logger) logger ;
			// Logger root = logger.getRootLogger() ;			// See Javadoc documentation above
			if (root != null) e = root.getAllAppenders() ;
			if (boudoir.getConfiguration().VERBOSE) System.out.println(leadingName + "Getting parent directory for : " + root) ;

			// Looking for the first instance of a FileAppender object
			if (e != null) while (e.hasMoreElements()) {
				Appender appender = (Appender) e.nextElement() ;
				if (appender instanceof FileAppender) {
					FileAppender fileAppender = (FileAppender) appender ;
					loggerPath = fileAppender.getFile() ;
					return loggerPath ;
				}
			}
		}

		return loggerPath ;
	}

	/**
	* Not used in current implementation, since roll over and rotation of the log file is performed
	* automatically by the Log4j external module.
	*
	* @return a <code>null</code> reference, since logger rotation is performed by Log4j
	*/
	public RotateThread getRotateThread() {

		System.err.println(leadingName + "Method not implemented: getRotateThread()") ;
		return null ;
	}

	/**
	* Return the rotation time set for this logger, already formatted and ready to print.<br/>
	* Usually this parameter is retrieved from the <em>RootLogger</em> reference, because the method
	* <code>root.getAllAppenders()</code> returns only the EXPLICIT appenders, and not the appenders
	* obtained from a parent logger. In this case we do not use the <em>RootLogger</em>, because we
	* assume the Log4j configuration file is something like:<br/>
	*
	* <pre><code>
	* log4j.additivity.template         = false
	* log4j.additivity.template.dba     = false
	* log4j.additivity.template.logic   = false
	* log4j.logger.template             = INFO, stdout, myfile
	* log4j.logger.template.dba         = INFO, stdout, myfile
	* log4j.logger.template.logic       = INFO, stdout, myfile
	* </code></pre><br/>
	*
	* Hence each logger object has his own explicit reference to the output file (<em>myfile</em>), for this
	* reason we can use directly: <code>smartLogger.getAllAppenders()</code>.
	*
	* @param logger a valid org.apache.log4j.Logger instance
	* @return a string describing the rotation time set for this logger
	*/
	public String getRotationTime(Object logger) {

		if (datePattern == null) {

			Enumeration e = null ;
			Logger root = (Logger) logger ;
			// Logger root = smartLogger.getRootLogger() ;	// See Javadoc documentation above
			if (root != null) e = root.getAllAppenders() ;
			if (boudoir.getConfiguration().VERBOSE) System.out.println(leadingName + "Getting rotation time for : " + root) ;

			// Assuming the file is the first instance of a DailyRollingFileAppender object
			if (e != null) while (e.hasMoreElements()) {
				Appender appender = (Appender) e.nextElement() ;
				if (appender instanceof DailyRollingFileAppender) {
					DailyRollingFileAppender daily = (DailyRollingFileAppender) appender ;
					datePattern = daily.getDatePattern() ;
					return datePattern ;
				}
			}
		}

		return datePattern ;
	}

// ******************************************************************************************************
// **********************************		UTILITIES		*********************************************
// ******************************************************************************************************

	/**
	* Convert ResourceBundle into a Properties object.<br/>
	* This method is necessary because Axis2 uses the Log4j mechanism to log events, as defined in the
	* <code>log4j.properties</code> file in the <code>axis2\WEB-INF\classes</code> directory. This means that
	* all the Logger objects instantiated by the Axis2 classloader will use the configuration specified by
	* the <code>log4j.properties</code> mentioned above. Hence, to avoid setting the logging configuration in
	* that file, we load our Log4j settings as a <code>ResourceBundle</code> object and then we convert it
	* into a <code>Properties</code> object.
	*
	* @param resource the resource bundle to convert
	* @return the properties version of the resource bundle
	*/
	protected Properties getProperties(ResourceBundle resource) {

		Enumeration keys = resource.getKeys() ;
		Properties properties = new Properties() ;

		while ( keys.hasMoreElements() ) {
			String key = (String) keys.nextElement() ;
			properties.put(key, resource.getString(key)) ;
			//System.out.println("Set " + key + " = " + resource.getString(key)) ;
		}

		return properties ;
	}

} // End of class