/**
* Title: 		 LoggerManager
* Description:	 General interface class for all the available logger managers
* Copyright:     Copyright (c) 2013
* Company:       Morpheus Quality Technologies
* Creation date: 06-12-2013
* Last modified: 06-12-2013
* @author 		 Stefano Adriani
* @version 		 1.0
*/

package adriani.boudoir.logging ;

import adriani.boudoir.Boudoir ;

/**
* General interface for defining, setting and handling the log mechanism.
*/
public interface LoggerManager {

	/**
	* The main method to be called to initialize the log mechanism.<br/>
	* Depending by subclass implementation this method can create a stand alone file system logger object,
	* (writing directly on the file system) or configure an external logging module (like apache.log4j).
	* In the first case the input parameters could be required or optional, depending by the subclass
	* implementation. In the second case the input parameters can be <code>null</code>, since the log4j mechanism
	* relies on an external configuration file.
	*
	* @param path an ALREADY existing path from which logging folders must be created
	* @param logLevel the logging minimal level as a String (INFO, WARNING or SEVERE)
	* @param rotate the timeout (in minutes) for rotation of the log file
	* @param mode if <code>true</code> initializes the standard log file, if <code>false</code> initializes the
	* benchmark analysis file
	*/
	public void init(String path, String logLevel, int rotate, boolean mode) ;

	/**
	* This method is a placeholder for subclasses relying on a stand alone logging mechanism, like the
	* <code>SystemLogger</code> class. If the logging feature is performed by an external module, for
	* example the apache.log4j package, implementation of this method is not required.
	*
	* @param path the path reference for the log file
	* @param level the logging minimal level as a Logger.Level value (INFO, WARNING or SEVERE)
	* @param mode if <code>true</code> creates the standard log file, if <code>false</code> creates the
	* benchmark analysis file
	* @see SystemLogger#getLogFile(String optionalPath, boolean mode)
	*/
	public void createLogger(String path, java.util.logging.Level level, boolean mode) ;

	/**
	* This method is a placeholder for subclasses relying on a stand alone logging mechanism, like the
	* <code>SystemLogger</code> class. If the logging feature is performed by an external module, for
	* example the apache.log4j package, implementation of this method is not required.
	*
	* @see RotateThread
	*/
	public void rotate() ;

	/**
	* Close the logger by closing all related handlers. The rotation thread will end just another cycle,
	* then will exit. This method should be required only by the <code>SystemLogger</code> class.
	*
	*/
	public boolean close() ;

// **********************************************************************************************************
// **********************************			ACCESSORS		*********************************************
// **********************************************************************************************************
	
	/**
	* Return the current <code>Boudoir</code> instance owning this logger. 
	*
	* @return the current <code>Boudoir</code> instance owning this logger
	*/
	public Boudoir getBoudoir() ;

	/**
	* Return the <code>parentDir</code> path used as base directory to locate the log file.
	*
	* @param logger a valid general Logger instance
	* @return the current parent directory for logging on the file system
	*/
	public String getParentDir(Object logger) ;

	/**
	* Return the <code>RotateThread</code> object (if any) used to rotate the log file.
	*
	* @return a reference to the <code>RotateThread</code> object (if used)
	*/
	public RotateThread getRotateThread() ;

	/**
	* Return the rotation time set for this logger, already formatted and ready to print.
	*
	* @param logger a valid general Logger instance
	* @return a string describing the rotation time set for this logger
	*/
	public String getRotationTime(Object logger) ;

} // End of interface