/**
* Title: 		 WSLogger
* Description:	 The "wrapper factory" implementing a log mechanism independent by the actual logging system
* Copyright:     Copyright (c) 2013
* Company:       Morpheus Quality Technologies
* Creation date: 31-10-2013
* Last modified: 06-12-2013
* @author 		 Stefano Adriani
* @version 		 1.0
*/

package adriani.boudoir.logging ;

import adriani.boudoir.* ;
import adriani.boudoir.util.Utilities ;
import adriani.boudoir.app.GeneralApplication ;

import java.io.* ;
import java.util.* ;

/**
* This class has two different purposes: <em>Factory</em> and single <em>logger instance</em>.<br/><br/>
*
* All the <code>static</code> attributes and methods written here are used to implement a
* <strong>factory</strong> which creates <code>WSLogger</code> single objects, used by all the other classes
* to write message on the logging system. This section contains:<br/><br/>
*
*<ul>
* <li>Static class attributes: <code>loggerManager</code> object, <code>benchLogger</code> object,
* <code>table</code> hashtable, <code>logPath</code>, rotation time,  initialized flag.</li>
* <li>Static methods: The <code>initialize()</code> method (to be run once) and the <code>getLogger()</code>
* method, the last representing the actual Factory for creating loggers.</li>
* <li>Static accessors: allowing to access the static fields listed above.</li>
* </ul>
*
* Using the Factory, any other class can retrieve a <code>WSLogger</code> instance, which is a wrapper
* for the actual logger being used (JAVA or LOG4J). The logger instance offers:<br/><br/>
*
* <em>Class attributes</em>: one of the available logging objects (Java logging or Log4j).<br/>
* <em>Methods</em>: the standard logging interface (logInfo, logWarning, logSevere, isInfoEnabled).<br/>
*/
public class WSLogger {
	
	/** Reference to the top-level <code>Boudoir</code> instance.													*/
	public 		 			Boudoir 						boudoir ;

	/**
	* Reference to the <code>LoggerManager</code> object created at the first logger request.
	* Depending by the service configuration, the <code>loggerManager</code> can be a 
	* <code>SystemLogger</code> (pure Java logger) or an <code>ExternalLogger</code> (Log4j logger).
	*/
	protected 		 		LoggerManager 					loggerManager ;

	/**
	* Reference to the <code>LoggerManager</code> implementing the SQL benchmark feature.
	* The <code>benchLogger</code> logs on file detailed information about the single SQL queries, this
	* is different from the <code>serviceAnalyzer</code> object, which collects global and statistical data.<br/><br/>
	* 
	* <strong>Remark</strong>: this logger is always an instance of the <code>SystemLogger</code> class.
	*/
	protected 		 		LoggerManager 					benchLogger ;

  	/** Logger path, loaded from the configuration file (used only by the <code>SystemLogger</code> class).			*/
	protected 		 		String 							logPath 		= "." ;		// Default value
	
  	/** Logger rotation time, loaded from the configuration file (used only by the <code>SystemLogger</code> class).*/
	protected 		 		int 							rotate 			= 720 ;		// Default value

	/** Boolean flag depicting if this object has been already initialized.											*/
	protected 		 		boolean							ready 			= false ;	// Default value

	/**
	* Boolean flag selecting the level on details message on STDOUT.
	* We use this flag, and not the <code>Settings.VERBOSE</code> flag because we want to keep separated
	* the verbosity of the library (general purposes) from the verbosity of this class (logging management).
	*/
	protected 		 		boolean							verbose			= false ;	// Default value

	/** The hash table tracking all the <code>WSLogger</code> objects produced by the factory.				*/
	private 	 			Hashtable<String, WSLogger>		table ;

  	/** Reference to the logger object relying on the Log4j mechanism (<code>ExternalLogger</code>) 		*/
	protected 				org.apache.log4j.Logger 		smartLogger ;

  	/** Reference to the logger written using the POJO approach (<code>SystemLogger</code>).				*/
	protected 				java.util.logging.Logger 		javaLogger ;

  	/** Beginning of each logging message produced by this class.											*/
	protected static 		String 							leadingName 	= "WSLogger - " ;
	
	/** The logger instance to be used for this class (static).												*/
	protected static 		WSLogger						myLogger 		= null ;
	
	/**
	* Standard constructor, bounding this class to a single <code>Boudoir</code> instance.
	* 
	* @param boudoir the <code>Boudoir</code> instance "owning" this object
	*/
	public WSLogger(Boudoir boudoir) {
		
		this.boudoir = boudoir ;
	}

	/**
	* The input <code>name</code> parameter is used only if the logging mechanism relies on the LOG4J
	* library. Otherwise, if the JAVA logging system is used, all the logger instances will share
	* the same logic name (defined by the <code>Settings.LOGGER</code> field).
	*
	* @param boudoir the <code>Boudoir</code> instance "owning" this object
	* @param name the logic name to be used for the LOG4J logger object
	* @see Settings
	*/
	public WSLogger(Boudoir boudoir, String name) {
		
		this(boudoir) ;
		if (boudoir.getConfiguration().JAVA) 	this.javaLogger		= java.util.logging.Logger.getLogger(boudoir.getConfiguration().LOGGER) ;
		if (boudoir.getConfiguration().LOG4J) 	this.smartLogger	= org.apache.log4j.Logger.getLogger(name) ;
		//System.out.println(leadingName + "Created using : " + (Settings.LOG4J ? "LOG4J" : "JAVA")) ;
	}
	
	/**
	* Method useful for debug and testing, the verify the scope of the various logger instances.
    */
    public void finalize() {

		if (verbose) System.out.println(leadingName + "Finalized object: " + this) ;
	}

// **********************************************************************************************************
// ************************************			FACTORY				*****************************************
// **********************************************************************************************************

	/**
	* Loads the logging parameters from the <code>configuration.properties</code> file, then creates the
	* required logger object, i.e. one subclass of the <code>LoggerManager</code> interface:<br/><br/>
	*
	* If <code>logger.type = LOG4J</code> the logger will be a <code>ExternalLogger</code> instance.<br/>
	* If <code>logger.type = JAVA</code> the logger will be a <code>SystemLogger</code> instance.<br/>
	*/
	private synchronized void initialize() {

		// 1. Storing initialize status (to avoid multiple initializing calls)
		ready 	= true ;
		table 	= new Hashtable<String, WSLogger>() ;

		// 2. Setting default values
		String level 			= "INFO" ;
		String logdir 			= "logs" ;
		String linuxPath 		= "/home/boudoir";
		String windowsPath 		= "C:\\Home\\boudoir" ;

		// 3. Loading data from the bd_configuration.properties file
		ResourceBundle rb	= Utilities.loadBundle( this.boudoir.getConfiguration().CONFIG ) ;
		if (rb != null) {
			// Retrieving string properties
			logdir				= rb.getString("logger.dir"		) ;
			linuxPath			= rb.getString("path.linux"		) ;
			windowsPath			= rb.getString("path.windows"	) ;
			level				= rb.getString("logger.level"	) ;

			// Parsing boolean properties
			String type							= rb.getString("logger.type"		) ;
			String bench						= rb.getString("library.benchmark"	) ;
			boudoir.getConfiguration().LOG4J 	= type.equalsIgnoreCase("LOG4J"	) ;
			boudoir.getConfiguration().JAVA		= type.equalsIgnoreCase("JAVA"	) ;
			boudoir.getConfiguration().BENCH	= bench.equalsIgnoreCase("YES"	) ;

			// Parsing integer properties
			String value		= rb.getString("logger.rotate") ;
			try 				{ rotate = Integer.parseInt(value.trim()) ;		}
			catch(Exception e) 	{ rotate = 720 ;								}	// Default value (12 hours)

			// Storing configuration in the Settings class
			boudoir.getConfiguration().LOGDIR		= logdir ;
			boudoir.getConfiguration().ROTATE		= rotate ;
			if (boudoir.getConfiguration().VERBOSE) System.out.println(WSLogger.leadingName + "Loaded configuration file: " + boudoir.getConfiguration().CONFIG + ".properties") ;
		} else {
			System.err.println(WSLogger.leadingName + "Using default configuration properties") ;
		}

		// 4. Retrieving the log path as defined in the Settings.CONFIG properties file
		if 		(File.separatorChar == '\\') 	logPath = windowsPath ;
		else if (File.separatorChar == '/') 	logPath = linuxPath ;

		// 5. Creating the logging mechanism
		if (boudoir.getConfiguration().JAVA) {
			loggerManager = new SystemLogger(boudoir) ;
			loggerManager.init(logPath, level, rotate, true) ;
		} else if (boudoir.getConfiguration().LOG4J) {
			loggerManager = new ExternalLogger(boudoir) ;
			loggerManager.init(logPath, level, rotate, true) ;
		}

		// 6. Eventually preparing the benchmark logger
		if (boudoir.getConfiguration().BENCH) initBenchmark() ;
		
		// 7. Creating a logging object for the "utility" package
		myLogger = this.getLogger(WSLogger.class) ;
		if (myLogger.isInfoEnabled()) myLogger.logInfo(leadingName + "Initialized logger system: " + loggerManager) ; 
	}

	/**
	* The main FACTORY to be used by all other classes to ask and obtain a <code>WSLogger</code>
	* instance. Remember that a <code>WSLogger</code> instance MUST NOT be used to invoke the
	* <code>initialize()</code> and <code>getLogger()</code> methods.<br/>
	* Instead, a single <code>WSLogger</code> instance MUST BE used only for actual logging, calling one
	* of the standard interface methods: <code>logInfo()</code>, <code>logWarning()</code>,
	* <code>logSevere()</code> and <code>isInfoEnabled()</code>.
	*
	* @param caller the class reference of the caller, e.g. <code>Boudoir.class</code>
	* @return an instance of this class, to be used as a wrapper for the log mechanism
	*/
	public synchronized WSLogger getLogger(Class<?> caller) {
		
		// Checking if we need to initialize this class (to be run only once)
		if ( !isReady() ) initialize() ;
		//else System.out.println(WSLogger.leadingName + "WSLogger already initialized") ;

		// Checking if we already have a logger for this package
		String 		name 	= caller.getPackage().getName() ;
		WSLogger 	logger 	= (WSLogger) table.get(name) ;

		// If there isn't any logger, we create a brand new one
		if (logger == null) {
			logger = new WSLogger(this.boudoir, name) ;
			table.put(name, logger) ;
			if (verbose) System.out.println(leadingName + "Created WSLogger@" + printObject(logger) + " for " + caller ) ;
		}
		// Else we return the existing one
		else {
			if (verbose) System.out.println(leadingName + "Reusing WSLogger@" + printObject(logger) + " for " + caller ) ;
		}
		
		// Returning logger instance
		return logger ;
	}

// **********************************************************************************************************
// **********************************		FACTORY	ACCESSORS		*****************************************
// **********************************************************************************************************

	/**
	* Factory accessor method.
    */
	public boolean isReady() 							{			return ready ; 							}

	/**
	* Return current implementation of the <code>LoggerManager</code> object, which is responsible for
	* creating and initializing the log mechanism. This object can be a <code>ExternalLogger</code> 
	* (is Log4j is used) or a <code>SystemLogger</code> (if POJO logging is used).
	*
	* @return a reference to the general <code>LoggerManager</code> object
    */
	public LoggerManager getLoggerManager() 			{			return loggerManager ; 					}

	/**
	* Return current implementation of the <code>LoggerManager</code> object, which is responsible for
	* creating and initializing the benchmark optional feature. This object is always an instance
	* of the <code>SystemLogger</code> class.
	*
	* @return a reference to the benchmark <code>LoggerManager</code> object
    */
	public LoggerManager getBenchmarkLogger() {
		
		if (benchLogger == null) initBenchmark() ;
		return benchLogger ; 					
	}

// **********************************************************************************************************
// **********************************		INSTANCE ACCESSORS		*****************************************
// **********************************************************************************************************

	/**
	* Return the <code>parentDir</code> path used as base directory to locate the log file.
	*
	* @return the current parent directory for logging on the file system
	*/
	public String getParentDir() {

		String path = "unset" ;
		if (loggerManager instanceof SystemLogger) {
			path = loggerManager.getParentDir(null) ;
		} else if (loggerManager instanceof ExternalLogger) {
			path = loggerManager.getParentDir(smartLogger) ;
		}
		return path ;
	}

	/**
	* Return the rotation time set for this logger, already formatted and ready to print.
	*
	* @return a string describing the rotation time set for this logger
	*/
	public String getRotationTime() {

		String time = "unset" ;
		if (loggerManager instanceof SystemLogger) {
			time = loggerManager.getRotationTime(null) ;
		} else if (loggerManager instanceof ExternalLogger) {
			time = loggerManager.getRotationTime(smartLogger) ;
		}
		return time ;
	}
	
// **********************************************************************************************************
// ************************************			BENCHMARK			*****************************************
// **********************************************************************************************************
	
	/**
	* Initializes the Benchmark logger, allowing to measure execution time of each single SQL statement.
	* In previous versions this code-block was embedded within the <code>initialize()</code>: we moved it 
	* out from that method in order to allow the possibility to activate the benchmark logger at run-time.
	*/
	public void initBenchmark() {

		if (boudoir.getConfiguration().VERBOSE) System.out.println(leadingName + "Initializing new benchmark on path: " + logPath) ;		
		benchLogger = new SystemLogger(boudoir) ;
		benchLogger.init(logPath, "INFO", boudoir.getConfiguration().ROTATE, false) ;
	}
	
	/**
	* Closes the Benchmark logger.
	*/
	public void closeBenchmark() {
		
		if (boudoir.getConfiguration().VERBOSE) System.out.println(leadingName + "Closing old benchmark object") ;		
		// 1. Closing the logger manager (manages handlers)
		benchLogger.close() ;
		// 2. Removing the reference to old object (in order to easy future new benchmark instances)
		benchLogger = null ;
	}
	
// **********************************************************************************************************
// ************************************			LOGGING				*****************************************
// **********************************************************************************************************

	/**
	* Log an INFO message on the current <code>Logger</code> object and flush the relative Handler.
	*
	* @param msg the text message to be logged on the current logger
	*/
	public void logInfo(String msg) {

		if (boudoir.getConfiguration().JAVA) {
			javaLogger.info(msg) ;
			flushLogger(javaLogger) ;
		}
		else if (boudoir.getConfiguration().LOG4J) {
			smartLogger.info(msg) ;
		}
	}

	/**
	* Log a WARNING message on the current <code>Logger</code> and flush the relative Handler.
	*
	* @param msg the text message to be logged on the current logger
	*/
	public void logWarning(String msg) {

		if (boudoir.getConfiguration().JAVA) {
			javaLogger.warning(msg) ;
			flushLogger(javaLogger) ;
		}
		else if (boudoir.getConfiguration().LOG4J) {
			smartLogger.warn(msg) ;
		}
	}

	/**
	* Log a SEVERE(i.e. ERROR) message on the current <code>Logger</code> and flush the relative Handler.
	*
	* @param msg the text message to be logged on the current logger
	*/
	public void logSevere(String msg) {

		if (boudoir.getConfiguration().JAVA) {
			javaLogger.severe(msg) ;
			flushLogger(javaLogger) ;
		}
		else if (boudoir.getConfiguration().LOG4J) {
			smartLogger.error(msg) ;
		}
	}

	/**
	* Flush the Handler bound to the input <code>Logger</code>. Used only when the <code>SystemLogger</code>
	* class is used (instead of the Log4j mechanism).
	*
	* @param logger a reference to the logger object to be flushed
	*/
	public void flushLogger(java.util.logging.Logger logger) {

		java.util.logging.Handler[] handlers = logger.getHandlers() ;
		if ( (handlers.length > 0) && (handlers[0] != null) ) handlers[0].flush() ;
		//System.out.println(leadingName + "Flushing logger: " + logger + " ("+ handlers.length + ")") ;
	}

	/**
	* Improves performances by avoiding to evaluate the concatenated string message when the log call will not
	* be executed (because not enabled for logger current level).
	* This method return always <code>false</code> if the logger object is not found.
	*
	* @return <code>true</code> if the INFO log level is enabled on the current Logger
	*/
	public boolean isInfoEnabled() {
		
		if (boudoir.getConfiguration().JAVA) {
			return (javaLogger.getLevel() == java.util.logging.Level.INFO) ;
		}
		else if (boudoir.getConfiguration().LOG4J) {
			return smartLogger.isInfoEnabled() ;
		}
		return false ;
	}

// **********************************************************************************************************
// ************************					UTILITIES				*****************************************
// **********************************************************************************************************

	/**
	* Print the instance reference number of the input object.
	*
	* @param object a generic Java object, e.g. MyObject@1bd4722
	* @return the serial number of the input object, e.g. 1bd4722
	*/
	public static String printObject(Object object) {

		String fullName = object.toString() ;
		String[] result = fullName.split("@") ;
		if (result == null) return "unknown" ;
		else return result[1] ;
	}

} // End of class