/**
* Title: 		 KeywordsManager (legacy name: ServiceData)
* Description:	 The ValueObject class wrapping the main business logic parameters
* Copyright:     Copyright (c) 2013
* Company:       Morpheus Quality Technologies
* Creation date: 06-12-2013
* Last modified: 12-12-2013
* @author 		 Stefano Adriani
* @version 		 1.0
*/

package adriani.boudoir.util ;

import java.io.* ;
import java.util.* ;
import java.util.regex.Pattern ;

import adriani.boudoir.Boudoir ;
import adriani.boudoir.app.GeneralAnalyzer;

/**
* The container for all settings used by the business service logic (string values etc.).<br/>
* Instantiation of this class should not be necessary since these methods are called only once, within a static
* code block of the <code>ServiceMacro</code> class, in order to retrieve all the static settings used by the
* business logic. All these settings are stored into static attributes of this class, hence these parameters
* are available without the need to instantiate this class.
*/
public class KeywordsManager {
	
	/** Reference to the top-level <code>Boudoir</code> instance.											*/
	public 		 			Boudoir 					boudoir ;

	/** List of the keywords used by this service (loaded from the property file).							*/
	protected 				Hashtable<String, String> 	hash ;

	/** List of the keywords used by this service (loaded from the property file).							*/
	protected 		 		StringBuffer[]				data_keys ;

   	/** Beginning of each logging message for this class.													*/
	protected static 		String 						leadingName 	= "KeywordsManager - " ;
	
	/**
	* Standard constructor, bounding this class to a single <code>Boudoir</code> instance.
	* 
	* @param boudoir the <code>Boudoir</code> instance "owning" this object.
    */
    public KeywordsManager(Boudoir boudoir) {

    	this.boudoir 	= boudoir ;
    }	
	
 // *************************************************************************************************************
 // ***************************				READING KEYWORDS 				 ************************************
 // *************************************************************************************************************
	
	/**
	* Load all the keywords from the <code>keywords.properties</code> file and store them into the
	* <code>data_keys</code> class attribute. The actual name of the configuration file is defined by the
	* <code>Settings.KEYWORDS</code> parameter.
	*
	* @param source the name of the properties file (usually is "keywords.properties")
	* @return <code>true</code> if the keywords have been successfully loaded from the properties file
    */
    public boolean loadKeywords(String source) {

		boolean fromFile ;
		String 	value 	= null ;
				hash 	= new Hashtable<String, String>() ;
		// Deprecated
		//Vector v = new Vector(100) ;

		// 1. Loading all values from the properties file
		try {
			ResourceBundle 	rb 	= ResourceBundle.getBundle(source) ;
			Enumeration 	e 	= rb.getKeys() ;
			while ( e.hasMoreElements() ) {
				String key = (String) e.nextElement() ;
				value = rb.getString(key) ;
				hash.put(key, value) ;
				// Deprecated
				//v.add( new StringBuffer(value) ) ;
			}
			fromFile = true ;
			//System.out.println(leadingName + "Loaded " + hash.size() + " values from: " + source + ".properties") ;

		} catch(MissingResourceException mre) {
			fromFile = false ;
			System.err.println(leadingName + "Service data file not found") ;
			mre.printStackTrace() ;
		}

		// 2. If successful loads the keywords into this class's array
		if (fromFile) {
			Collection<String> c = hash.values() ;
			data_keys = new StringBuffer[hash.size()] ;
			String[] array = (String[]) c.toArray( new String[c.size()] ) ;
			for (int i = 0 ; i < array.length; i++) data_keys[i] = new StringBuffer(array[i]) ;
			if (boudoir.getConfiguration().VERBOSE) System.out.println(leadingName + "Loaded " + data_keys.length + " keywords") ;
		}	else {
			// 2.a Otherwise we use the hard-coded values (see below)
			data_keys = getHardCodedValues() ;
			if (boudoir.getConfiguration().VERBOSE) System.out.println(leadingName + "Using " + data_keys.length + " hardcoded keywords") ;
		}

		// Keep for testing or debug
		//printHashtable(hash) ;		System.exit(0) ;

		// 3. Returning result
		return fromFile ;
	}
    
 // *************************************************************************************************************
 // ***************************				KEYWORDS GENERATION				************************************
 // *************************************************************************************************************
    
	/**
	* Create a <code>keywords.java</code> data file containing the Java source definition of all the keywords.
	* This method should be used only during development, when the properties file should be always available,
	* to produce the hard-coded list that will be retrieved by the <code>getHardCodedValues()</code> method.
	*
	* @return true if the <code>keywords.java</code> has been produced successfully
	* @see #getHardCodedValues()
    */
    public boolean produceKeywords() {

		// 1. Initializing
		String sink = "keywords.java" ;
		BufferedWriter writer = null ;

		if (data_keys == null) {
			System.err.println(leadingName + "Keywords not loaded") ;
			return false ;
		}

		try {
			int howMany = data_keys.length ;
            FileWriter fileWriter = new FileWriter(sink) ;
			writer = new BufferedWriter(fileWriter) ;
			for (int i = 0; i < howMany; i++) {
				writer.write("buffer[" + i + "] = new StringBuffer(\"" + data_keys[i] + "\") ;\n") ;
			}
			writer.close() ;
        } catch (IOException exception) {
            System.err.println(leadingName + "Error reading file") ;
			return false ;
        }

		//System.out.println(leadingName + "Produced Java file: " + sink) ;
		return true ;
	}

 	/**
 	* Return an hard-coded list of all the necessary keywords.
 	*
 	* @return the <code>StringBuffer</code> array containing all the required keywords
 	* @see #produceKeywords()
     */
     public static StringBuffer[] getHardCodedValues() {

 		StringBuffer[] buffer = new StringBuffer[11] ;

 		buffer[0]  = new StringBuffer("Axis2"		) ;
 		buffer[1]  = new StringBuffer("Service"		) ;
 		buffer[2]  = new StringBuffer("Web"			) ;
 		buffer[3]  = new StringBuffer("Bug fixing"	) ;
 		buffer[4]  = new StringBuffer("Java"		) ;
 		buffer[5]  = new StringBuffer("Run"			) ;
 		buffer[6]  = new StringBuffer("Deploy"		) ;
 		buffer[7]  = new StringBuffer("Develop"		) ;
 		buffer[8]  = new StringBuffer("Tomcat"		) ;
 		buffer[9]  = new StringBuffer("Publish"		) ;
 		buffer[10] = new StringBuffer("Test"		) ;

 		return buffer ;
 	}

// *************************************************************************************************************
// ***************************					ACCESSORS				****************************************
// *************************************************************************************************************

	/**
	* Accessor method.
	*/
	public Hashtable getHashtable()						{		return hash ;									}
	
// *************************************************************************************************************
// ***************************					CLASS UTILITIES				************************************
// *************************************************************************************************************

	/**
	* Create a <code>keywords.properties</code> data file containing the list of all the necessary keywords.
	* The input source file can be specified running the <code>SimpleClient</code> from the command line,
	* using the option <code>-c <em>file name</em></code>. If no path it's specified, the source file
	* will be searched in the current directory. Eventually it's possible to specify an absolute path,
	* for example: <code>C:/Develop/name_list.txt</code>.
	*
	* @param source the source file name
	* @return true if the <code>keywords.properties</code> has been produced successfully
    */
    public boolean createKeywords(String source) {

		// 1. Initializing
		BufferedReader reader = null ;
		BufferedWriter writer = null ;
		String sink = "keywords.properties" ;
		System.out.println() ;

		// 2. Opening input stream using a BufferedReader
		try {
			FileReader fileReader = new FileReader(source) ;
			reader = new BufferedReader(fileReader) ;
			System.out.println(leadingName + "Opened file: " + source) ;
        } catch (FileNotFoundException filenotfound) {
            System.err.println(leadingName + "Could not open file: " + source) ;
			return false ;
        }

		// 3. Reading and writing data into a BufferedWriter
		int i = 0 ;
		String line = null ;
		try {
            FileWriter fileWriter = new FileWriter(sink) ;
			writer = new BufferedWriter(fileWriter) ;
			// Main cycle running the input file
			while ( (line = reader.readLine()) != null ) {
				line = i++ + "=" + line ;
				writer.write(line, 0, line.length()) ;
				writer.newLine() ;
			}
			reader.close() ;
			writer.close() ;
        } catch (IOException exception) {
            System.err.println(leadingName + "Error reading file") ;
			return false ;
        }

		// 4. Logging result and returning
		//System.out.println(leadingName + "Created data file: " + sink) ;
		return true ;
	}

	/**
	* Tokenize the input string into an array of boolean, using the separator passed as input.
	* The size of the boolean array will match the maximum value contained in the input string. 
	* 
	* @param string the string containing all the tokens (e.g. "1, 2, 3")
	* @param sep the tokens separator (e.g ",")
	* @return the boolean array containing all parsed tokens
    */
    public static boolean[] getBooleanTokens(String string, String sep) {
    	
    	int			max		= 0 ;
    	int[] 		array 	= getIntegerTokens(string, sep) ;
    	for (int i : array) if (i > max) max = i ;
    	
    	boolean[] 	result	= new boolean[max + 1] ;
    	for (int i : array) result[i] = true ;
    	//for (int i = 0 ; i <= max; i++) System.out.println(leadingName + i + "->" + result[i]) ;
    	
    	return result ;
    }
    
	/**
	* Tokenize the input string into an array of integer, using the separator passed as input.
	* 
	* @param string the string containing all the tokens (e.g. "1, 2, 3")
	* @param sep the tokens separator (e.g ",")
	* @return the int array containing all parsed tokens
    */
    public static int[] getIntegerTokens(String string, String sep) {
    	
		Pattern 	pattern = Pattern.compile(sep) ;
		String[] 	tokens	= pattern.split(string) ;
    	int[]		result  = new int[tokens.length] ;
		for (int i = 0 ; i < tokens.length; i++) {
			try 				{	result[i] = Integer.parseInt( tokens[i].trim() ) ;			}
			catch(Exception e) 	{ 	result[i] = 0 ;												}
			//System.out.println(leadingName + "getParamToken: " + result[i]) ;
		}
    	return result ;
    }
    
 // *************************************************************************************************************
 // ***************************				TEST OR DEBUG					************************************
 // *************************************************************************************************************
    
	/**
	* Prints on the STDOUT the list of all the keywords loaded by this class.
	* This method should be used only during development, for testing or debug.
	*
	* @return true if the keywords list has been printed successfully
    */
    public boolean printKeywords() {

		if (data_keys == null) return false ;
		int howMany = data_keys.length ;

		System.out.println(leadingName + "KEYWORD LIST:\n") ;
		for (int i = 0; i < howMany; i++) System.out.println("buffer[" + i + "] = \"" + data_keys[i] + "\"") ;
		System.out.println() ;

		return true ;
	}

 	/**
 	* Prints on the STDOUT the list of all the keywords loaded by this class.
 	* This method should be used only during development, for testing or debug.
 	*
 	* @param hash an hash table, usually one of those produced by the <code>loadKeywords()</code> method
     */
     public static void printHashtable(Hashtable<String, String> hash) {

 		int counter = 0 ;
 		Enumeration e = hash.keys() ;

 		System.out.println() ;
 		while ( e.hasMoreElements() ) {
 			counter++ ;
 			String key 		= (String) e.nextElement() ;
 			String value 	= hash.get(key) ;
 			System.out.println(leadingName + counter + " : " + key + " = " + value) ;
 		}
 	}

} // End of class