/**
* Title: 		 Utilities
* Description:	 General purposes class for handling bundles, time-date conversion etc.
* Copyright:     Copyright (c) 2013
* Company:       Morpheus Quality Technologies
* Creation date: 06-12-2013
* Last modified: 06-12-2013
* @author 		 Stefano Adriani
* @version 		 1.0
*/

package adriani.boudoir.util ;

import java.util.* ;
import java.text.SimpleDateFormat ;

/**
* General utilities class for managing resource bundle or conversion of different formats, 
* typically date and time formats.<br/><br/>
* 
* Some notes about the standard Java date and time lexicon:<br/><br/>
*
* <strong>Calendar.ZONE_OFFSET</strong>: this offset represent the offset of the date specified
* in the data string description, for example "+01:00" or "+02:00".<br/>
* Example: the date "2010-02-26T15:40:00+00:00" will be converted into a calendar with offset equals 0,
* the date "2010-02-26T15:40:00+01:00" will be have an offset set to 3600000, while the date
* "2010-02-26T15:40:00+02:00" will be have an offset set to 7200000.<br/><br/>
*
* <strong>TimeZone</strong>: describes the time zone bound to a <code>Calendar</code> object.<br/>
* If the user changes the machine setting (e.g. on Windows: setting daylight saving in the control panel)
* the default <code>TimeZone</code> object will immediately reflect this change, as visible for example
* by calling the <code>useDaylightTime()</code> method. Instead, the raw offset of the time zone will
* not be affected by such a change, since it describes the difference with respect of the UTC time.
* In other words, the raw offset for the Italian time zone is always 3600000, and will never change.<br/>
* Notice that usually a calendar relies on the local <strong>default</strong> time zone instead of using
* the custom time zone depicting where the calendar was created (fortunately).<br/><br/>
*
*/
public abstract class Utilities {

	/** The data pattern used by Log4j for the timestamp, in this case: yyyy-MM-dd HH:mm:ss,SSS				*/
	protected static 		SimpleDateFormat 	sdf_log 			= new SimpleDateFormat("yyyy-MM-dd HH:mm:ss,SSS") ;
	/** The data pattern used by Log4j for the timestamp, in this case: yyyy-MM-dd HH:mm:ss,SSS				*/
	protected static 		SimpleDateFormat 	sdf_ita 			= new SimpleDateFormat("dd-MMM-yy HH:mm:ss,SSS", Locale.ITALY) ;
	/** The data pattern used as output for the time, in this case: HH:mm:ss,SSS							*/
	protected static 		SimpleDateFormat 	sdf_time 			= new SimpleDateFormat("HH:mm:ss") ;
	/** The data pattern used as output for the date, in this case: yyyy-MM-dd HH							*/
	protected static 		SimpleDateFormat 	sdf_date 			= new SimpleDateFormat("dd-MMM-yy") ;
	/** The data pattern used by Oracle to describe dates													*/
	protected static 		SimpleDateFormat 	sdf_oracle 			= new SimpleDateFormat("yyyy-MM-dd HH:mm:ss") ;
	/** Java utility object using same pattern of the FIRSTCHECK web service								*/
	public static final		SimpleDateFormat 	sdf_oracle_fs		= new SimpleDateFormat("yyyy-MM-dd") ;	

	/**
	* SimpleDateFormat static reference for conversion from the format yyyy-MM-dd'T'HH:mm:ss'+00:00',
	* referred to the standard UTC time.
	*/
	protected static 		SimpleDateFormat 	sdf_utc				= new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss'+00:00'") ;
	/**
	* SimpleDateFormat static reference for conversion from the format yyyy-MM-dd'T'HH:mm:ss'+01:00',
	* used in Italy when NOT in the saving light season.
	*/
	protected static 		SimpleDateFormat 	sdf_winter 			= new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss'+01:00'") ;
	/**
	* SimpleDateFormat static reference for conversion from the format yyyy-MM-dd'T'HH:mm:ss'+02:00',
	* used in Italy DURING the saving light season.
	*/
	protected static 		SimpleDateFormat 	sdf_summer 			= new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss'+02:00'") ;


  	/** Beginning of each logging message produced by this class.											*/
	protected static 		String 				leadingName 		= "Utilities - " ;

	/**
	* Due the Axis2 design the <code>Calendar</code> obtained from a SOAP REQUEST does not consider the
	* time zone, but it's created using the local time zone, which for Italy is GMT+01:00
	* (Central Europe Time) in winter and GMT+02:00 (Central Europe Summer Time) in summer.
	* For this reason, the time zone of the input datetime parameter will be specified in the
	* <code>Calendar.ZONE_OFFSET</code> field.<br>
	* Example: if the SOAP input date is like "2010-02-26T15:40:00+02:00", then the calendar
	* object will have an offset of 7200000 ms.<br/><br/>
	*
	* This method allows to emulate the Axis2 behavior when taking datetime values from another source,
	* like an hard-coded string array or a properties file.<br/>
	*
	* @param value a <code>string</code> object describing a <code>Date</code>
	* @return the same date converted into an equivalent <code>Calendar</code> object
	*/
	public static Calendar getCalendar(String value) throws java.text.ParseException {

		Date 		time		= null ;
		char 		type		= value.charAt(21) ;
		Calendar	calendar	= Calendar.getInstance() ;
		System.out.println(leadingName + "Parsing date: " + value + " [" + type + "]") ;

		// Emulating behavior of the SOAP input date, which set the time zone offset in the calendar
		// and does not set the daylight saving offset.
		switch(type) {
			case	'1'	:	time = sdf_winter.parse(value) ;
							calendar.setTime(time) ;
							calendar.set(Calendar.DST_OFFSET, 0) ;
							calendar.set(Calendar.ZONE_OFFSET, 3600000) ;
							break ;
			case	'2'	:	time = sdf_summer.parse(value) ;
							calendar.setTime(time) ;
							calendar.set(Calendar.DST_OFFSET, 0) ;
							calendar.set(Calendar.ZONE_OFFSET, 7200000) ;
							break ;
			default		:	time = sdf_utc.parse(value) ;
							calendar.setTime(time) ;
							calendar.set(Calendar.DST_OFFSET, 0) ;
							calendar.set(Calendar.ZONE_OFFSET, 0) ;
		}

		// This date will be the straight conversion, without offset, always in CET format (defaul)
		//System.out.println(leadingName + "Parsed date: " + time) ;
		// This calendar will use the local machine time zone
		//System.out.println(leadingName + "Inspecting calendar: " + Utilities.introspectElement(calendar)) ;

		return calendar ;
	}

 // *********************************************************************************************************
 // **********************************			SHORT DATES				*************************************
 // *********************************************************************************************************

    /**
	* Convert the input <code>Calendar</code> to the string format, using the format
	* <code>dd-MMM-yy</code>.
	*
	* @param calendar the <code>Calendar</code> to be converted
	* @return the formatted version of the input calendar
	* @see #sdf_date
    */
    public static String getShortDate(Calendar calendar) {

    	if (calendar == null)	return null ;
    	else					return getShortDate(calendar.getTimeInMillis()) ;
    }
    
    /**
	* Convert the input <code>java.util.Date</code> to the string format, using the format
	* <code>dd-MMM-yy</code>.
	*
	* @param date the <code>java.util.Date</code> to be converted
	* @return the formatted version of the input calendar
	* @see #sdf_date
    */
    public static String getShortDate(java.util.Date date) {

    	if (date == null)		return null ;
    	else					return getShortDate(date.getTime()) ;
    }

    /**
	* Convert the input <code>timestamp</code> to the string format, using the format
	* <code>dd-MMM-yy</code>.
	*
	* @param timestamp the timestamp to be converted
	* @return the formatted version of the input timestamp
	* @see #sdf_date
    */
    public static String getShortDate(long timestamp) {

		try {
			Date date = new Date(timestamp) ;
			return sdf_date.format(date) ;
		} catch (Exception e) {
			e.printStackTrace() ;
			return null ;
		}
	}
    
// *********************************************************************************************************
// **********************************			TIMESTAMPS				************************************
// *********************************************************************************************************
    
	/**
	* Parse the string value of the input <code>StringBuffer</code> into a <code>java.util.date value</code>
	* using a pattern like 'yyyy-MM-dd HH:mm:ss'. 
	*
	* @param buffer the input <code>StringBuffer</code> object
	* @return the <code>java.util.date</code> value of the input string
	* @see #sdf_oracle 
	*/
	public static java.util.Date valueOfDate(StringBuffer buffer) {
		
		if (buffer != null)	return valueOfDate(buffer.toString()) ;
		else				return null ;
	}
	
	/**
	* Parse the string value of the input <code>StringBuffer</code> into a <code>java.sql.date value</code>,
	* using a pattern like 'yyyy-MM-dd HH:mm:ss'.
	*
	* @param buffer the input StringBuffer object
	* @return the java.util.date value of the input string
	* @see #sdf_oracle 
	*/
	public static java.sql.Date valueOfSQLDate(StringBuffer buffer) {
		
		if (buffer != null)	return valueOfSQLDate(buffer.toString()) ;
		else				return null ;
	}
    
	/**
	* Parse the string value of the input <code>String</code> into a <code>java.sql.date value</code>,
	* using a pattern like 'yyyy-MM-dd HH:mm:ss'. 
	*
	* @param string the input string object
	* @return the <code>java.sql.date</code> value of the input string
	* @see #sdf_oracle 
	*/
	public static java.sql.Date valueOfSQLDate(String string) {

		long time = 0 ;
		java.util.Date date = valueOfDate(string) ;
		if (date != null) time = date.getTime() ;
		return new java.sql.Date(time) ;
	}
	
	/**
	* Parse the string value of the input <code>String</code> into a <code>java.util.date value</code>,
	* using a pattern like 'yyyy-MM-dd HH:mm:ss'. 
	*
	* @param string the input string object
	* @return the <code>java.util.Date</code> value of the input string
	* @see #sdf_oracle
	*/
	public static java.util.Date valueOfDate(String string) {
		
		java.sql.Date result = null ;
		try {
			java.util.Date date = sdf_oracle.parse( string.trim() ) ;
			return date ;
		} catch(Exception e) {
			e.printStackTrace() ;
			result = null ;
		}
		return result ;
	}

// *********************************************************************************************************
// **********************************		VERY SHORT DATES			************************************
// *********************************************************************************************************
	
	/**
	* Parse the string value of the input <code>String</code> into a <code>java.sql.date value</code>,
	* using a pattern like 'yyyy-MM-dd'. 
	*
	* @param string the input string object
	* @return the <code>java.sql.date</code> value of the input string
	* @see #sdf_oracle_fs
	*/
	public static java.sql.Date shortValueOfSQLDate(String string) {
		
		java.sql.Date result = null ;
		try {
			java.util.Date date = sdf_oracle_fs.parse( string.trim() ) ;
			long time = date.getTime() ;
			result = new java.sql.Date(time) ;
		} catch(Exception e) {
			result = null ;
		}
		return result ;
	}
	
	/**
	* Parse the string value of the input <code>StringBuffer</code> into a <code>java.sql.date value</code>,
	* using a pattern like 'yyyy-MM-dd'. 
	*
	* @param buffer the input string object
	* @return the <code>java.sql.date</code> value of the input string
	* @see #sdf_oracle_fs 
	*/
	public static java.sql.Date shortValueOfSQLDate(StringBuffer buffer) {
		
		if (buffer != null)	return shortValueOfSQLDate(buffer.toString()) ;
		else				return null ;
	}	
	
// *********************************************************************************************************
// **********************************		CALENDARS					************************************
// *********************************************************************************************************
	
	/**
	* Convert the input <code>java.sql.date</code> into a <code>Calendar</code> instance.
	*
	* @param date the input <code>java.sql.date</code> object
	* @return the <code>Calendar</code> value of the input date
	*/
	public static Calendar dateToCalendar(java.sql.Date date){
		
		  Calendar calendar = Calendar.getInstance() ;
		  calendar.setTime(date) ;
		  return calendar ;
	}
	
// *********************************************************************************************************
// **********************************		GENERAL CONVERSIONS			************************************
// *********************************************************************************************************
	
	/**
	* Parse the string value of the input <code>String</code> date into a <code>java.util.date value</code>,
	* using the input pattern. 
	*
	* @param date the input date in string format
	* @param pattern the string pattern to be used in the conversion (e.g. "dd-MMM-yy") 
	* @return the <code>java.util.Date</code> value of the input string
	* @see #sdf_oracle
	*/
	public static java.util.Date parseDate(String date, String pattern) {
		
		java.sql.Date 		result 	= null ;
		SimpleDateFormat 	sdf 	= new SimpleDateFormat(pattern) ;
		try {
			java.util.Date parsed = sdf.parse( date.trim() ) ;
			return parsed ;
		} catch(Exception e) {
			result = null ;
		}
		return result ;
	}
	
    /**
	* Convert the input <code>java.util.Date</code> to the string format, using the format
	* <code>dd-MMM-yy</code>.
	*
	* @param date the <code>java.util.Date</code> to be converted
	* @param pattern the string pattern to be used in the conversion (e.g. "dd-MMM-yy") 
	* @return the formatted version of the input date
	* @see #sdf_date
    */
    public static String formatDate(java.util.Date date, String pattern) {

    	if (date == null)		return null ;
    	else					return formatDate(date.getTime(), pattern) ;
    }
	
    /**
	* Convert the input <code>timestamp</code> to the string format, using the input
	* pattern to define the conversion format.
	*
	* @param timestamp the timestamp to be converted
	* @param pattern the string pattern to be used in the conversion (e.g. "dd-MMM-yy") 
	* @return the formatted version of the input timestamp
    */
    public static String formatDate(long timestamp, String pattern) {

    	SimpleDateFormat sdf = new SimpleDateFormat(pattern) ;
		try {
			Date date = new Date(timestamp) ;
			return sdf.format(date) ;
		} catch (Exception e) {
			e.printStackTrace() ;
			return null ;
		}
	}
	
 // *********************************************************************************************************
 // **********************************			INSPECTS DATES			*************************************
 // *********************************************************************************************************

	/**
	* Utility for printing introspection information about the ADBBeans elements.
	* For the time being this method support introspection only for datetime objects, in order to
	* investigate behavior of an Axis2 strange behavior:<br/><br/>
	*
	* <a href="https://issues.apache.org/jira/browse/AXIS-2673?page=com.atlassian.jira.plugin.system.issuetabpanels:all-tabpanel">
	* Axis2 issue</a>
	* <br/><br/>
	*
	* @param object the XML node value retrieved from a <code>ADBBean</code> object
	* @return a formatted string ready to be printed on the output stream
	*/
	public static String introspectElement(Object object) {

		Date			today	= new Date() ;
		String			RET		= new String("\n") ;
		StringBuffer 	text 	= new StringBuffer(RET) ;

		// Inspecting calendar objects
		if (object instanceof Calendar) {
			Calendar calendar 	= (Calendar) object ;
			Date 	date		= calendar.getTime() ;
			TimeZone timeZone 	= calendar.getTimeZone() ;
			//text.append(calendar.toString() + RET) ;
			text.append(	"TIMEZONE        : "							) ;
			text.append(	timeZone.toString()						+	RET	) ;
			text.append(	"CALENDAR DAY    : "							) ;
			text.append(	calendar.get(Calendar.DAY_OF_MONTH) + "-" + calendar.get(Calendar.MONTH) 	+ RET ) ;
			text.append(	"CALENDAR HOUR   : "							) ;
			text.append(	calendar.get(Calendar.HOUR_OF_DAY) 	+ ":" + calendar.get(Calendar.MINUTE)	+ RET ) ;
			text.append(	"CALENDAR OFFSET : "							) ;
			text.append(	calendar.get(Calendar.ZONE_OFFSET)		+	RET	) ;
			text.append(	"CALENDAR DST    : "							) ;
			text.append(	calendar.get(Calendar.DST_OFFSET)		+	RET	) ;
			text.append(	"TIME ZONE ID    : "							) ;
			text.append(	timeZone.getID()						+	RET	) ;
			text.append(	"NAME            : "							) ;
			text.append(	timeZone.getDisplayName()				+	RET	) ;
			text.append(	"USE DAYLIGHT    : "							) ;
			text.append(	timeZone.useDaylightTime()				+	RET	) ;
			text.append(	"IN DAYLIGHT     : "							) ;
			text.append(	timeZone.inDaylightTime(date)			+	RET	) ;
			text.append(	"RAW OFFSET      : "							) ;
			text.append(	timeZone.getRawOffset()					+	RET	) ;
			text.append(	"SYSTEM DATE     : "							) ;
			text.append(	today									+	RET	) ;
			text.append(	"CALENDAR DATE   : "							) ;
			text.append(	date									+	RET	) ;
		}
		else return "Unsopported" ;

		return text.toString() ;
	}
	
 // *********************************************************************************************************
 // *******************************			RESOURCES BUNDLE			*************************************
 // *********************************************************************************************************
	
	/**
	* Load the resource bundle identified by the input parameter.
	* If an error occurs we can not log the error on the log system, since we failed to acknowledge the logger 
	* configuration (which is depicted in the configuration file we failed to load).<br/><br/>
	* 
	* @param bundle name of the resources bundle to be loaded 
	* @return the instance of the <code>ResourceBundle</code> matching the <code>Settings.CONFIG</code> static field 
	* (if found), null otherwise
	*/
	public static ResourceBundle loadBundle(String bundle) {

		try {
			return ResourceBundle.getBundle(bundle) ;
		} catch(MissingResourceException e) {
			System.err.println(leadingName + "Log configuration properties file not found: " + bundle) ;
			e.printStackTrace() ;
			return null ;
		}
	}

} // End of class