/**
* Title: 		 SimpleClient
* Description:	 Simple client allowing to test the library's features from the command line
* Copyright:     Copyright (c) 2013-2014
* Company:       Morpheus Quality Technologies
* Creation date: 29-10-2013
* Last modified: 10-02-2014
* @author 		 Stefano Adriani
* @version 		 1.0
*/

package adriani.boudoir.test ;

import adriani.boudoir.* ;
import adriani.boudoir.db.* ;
import adriani.boudoir.util.* ;
import adriani.boudoir.logging.WSLogger ;
import adriani.boudoir.app.GeneralApplication ;

import adriani.boudoir.test.util.ClientThread ;
import adriani.boudoir.test.wrappers.Wrapper ;
import adriani.boudoir.test.wrappers.CustomerWrapper ;
import adriani.boudoir.test.wrappers.EntityTimeWrapper ;

import java.util.Calendar ;
import java.text.ParseException ;

/**
* A very simple command line class offering a <code>main()</code> method to allow invocation of the
* main library API directly on the command line.<br/>
* If all the necessary environment variables are correctly defined it should be possible to run
* this client writing:<br/><br/>
*
* <code>java -cp %CLASSPATH% adriani.boudoir.test.SimpleClient</code><br/><br/>
*
* For details run <code>SimpleClient -h</code>.
*/
public class SimpleClient {
	
	/** 
	* Top level instance of the Boudoir library.<br/>	
	* This object should be a singleton within the scope on the project accessing the library.
	* In this scenario, the <code>SimpleClient</code> emulates such "external" project. 														
	*/
	public 		 			Boudoir 		boudoir ;
	
	/** Static final value stored as class field to improve performances.									*/
	public static final		String			EMPTY 				= "" ;
	/** Static final value stored as class field to improve performances.									*/
	public static final		String			BLANK 				= " " ;
	/** Static final value stored as class field to improve performances.									*/
	public static final	 	String			ZERO 				= "0" ;
	/** Static final value stored as class field to improve performances.									*/
	public static final 	StringBuffer	NULL 				= new StringBuffer("NULL") ;
	
	/** Retrieving the logger instance to be used for this class.											*/
	protected 				WSLogger		logger 				= null ;

  	/** Beginning of each logging message produced by this class.											*/
	protected static 		String			leadingName 		= "SimpleClient - " ;

    /**
	* Main method used to call the web service from the command line, forwarding action to the
	* <code>run(String[] args)</code> method.<br/><br/>
	*
	* Default behavior (no option at all) will launch the simple client in the actual mode, using
	* hard-coded data and calling the first method, i.e. the getDonator() method.
	*
	* @param args the standard input string array, usually read from STDIN.
	* @see #run(String[] args, int method)
    */
    public static void main(String[] args) {

		// Running the client
		new SimpleClient().run(args) ;
		// Comment next line if necessary (usually we have pool threads running)
		System.exit(0) ;
	}
    
	/**
	* Standard constructor, bounding this class to a single <code>Boudoir</code> instance.
	* In this case the <code>Boudoir</code> instance is build using the default constructor, 
	* which creates a <code>Boudoir</code> bound to the default logical name (i.e. "bd").
	* 
	* @param boudoir the <code>Boudoir</code> instance "owning" this object
	* @see Boudoir#DEFAULT_NAME
    */
    public SimpleClient() {
    	
    	this.boudoir = new Boudoir() ;
    }

    /**
	* If the client is run in the emulate mode the method forwards action to one the <code>testXXX()</code>
	* methods, which invokes the service logic using the hard-coded input parameters.<br/>
	* Otherwise, if the client is run in the actual mode, action is forwarded to one the actual
	* methods (named as the WSDL methods), which calls the service skeleton passing the input parameters 
	* (as above).
	*
	* @param args the standard input string array, usually read from STDIN
    */
    public void run(String[] args) {
    	
    	// 1. loading the proper logger for this class
		this.logger = boudoir.getLogger().getLogger(SimpleClient.class) ;

    	// 2. Logging run environment
    	if (boudoir.getConfiguration().VERBOSE) {
    		String runInfo = (args != null) ? (args.length + " args") : "no args" ;
    		System.out.println(leadingName + "Running with " + runInfo) ;
    	}

		// 3. Managing options (if any)
		if ( (args != null) && (args.length > 0) && (args[0].startsWith("-")) ) {
			if (args[0].length() > 1) manageOptions(args) ;
			else System.out.println(leadingName + "Please specify a valid option, or use -h to get some help") ;
		} 
		
		// 4. Handling remote command manager
		else if ( (args != null) && (args.length > 0) ) {
			new GeneralApplication(this.boudoir).manageCommand(args[0]) ;
		}

		// 4. Ending
		if (boudoir.getConfiguration().VERBOSE) System.out.println(leadingName + "ending run() method ...") ;
	}

// *********************************************************************************************************
// **********************************		MANAGING OPTIONS		****************************************
// *********************************************************************************************************

    /**
	* Manage all the possible options available from the command line.
	* The available test offered by the <code>SimpleClient</code> are executed by calling:<br/><br/>
	*
	* <code>java -cp %CLASSPATH% adriani.boudoir.client.SimpleClient -t n</code><br/><br/>
	*
	* Where the <code>n</code> parameter select one of the following test:<br/><br/>
	*
	* 0. Test the logger system offered by the <code>WSLogger</code> class.<br/>
	* 1. Forces the class loader to load the <code>GeneralApplication</code> class in order to verify that
	* the configuration and initialization sequence is executed correctly.<br/>
	* 2. Verify database environment (driver and connection pool) and all active threads.<br/>
	* 3. Print the list of all active threads.<br/>
	* 4. Test a generic code-block (unit-testing during development).<br/>
	*
	* @param args the input string array
    */
    public void manageOptions(String[] args) {

		// Adjusting input
		String option = args[0].substring(0,2) ;
		String combo  = "" ;
		if ( args[0].length() > 2) combo = args[0].substring(0,3) ; 
		StringBuffer[] buffer = new StringBuffer[args.length] ;
		for (int i = 0 ; i < args.length ; i++) buffer[i] = new StringBuffer(args[i]) ;

		// Printing help screen
		if (option.equalsIgnoreCase("-h")) {
			printHelp() ;
		}
		// Printing remote manager commands list
		else if (option.equalsIgnoreCase("-l")) {
			StringBuffer text = new GeneralApplication(this.boudoir).getCommandsList() ;
			System.out.println(text.toString()) ;
		}
		// Creating the keywords.properties file from a given txt file
		else if (option.equalsIgnoreCase("-c")) {
			String source = null ;
			if ( (args.length > 1) && (args[1] != null) ) source = args[1] ;
			else source = "name_list.txt" ;
			KeywordsManager keywordsManager = boudoir.getKeywordsManager() ;
			keywordsManager.createKeywords(source) ;
		}
		// Dumping the content of the keywords.properties file into the keywords.java file (source code-block)
		else if (option.equalsIgnoreCase("-d")) {
			System.out.println() ;
			KeywordsManager keywordsManager = boudoir.getKeywordsManager() ; 
			keywordsManager.loadKeywords( boudoir.getConfiguration().KEYWORDS ) ;
			keywordsManager.produceKeywords() ;
		}
		// Executing a basic test
		else if (option.equalsIgnoreCase("-t")) {
			int code = getIntValue(buffer, 1) ;
			switch(code) {
				case 	0	: 	testLogger() ; 					break ;
				case 	1	: 	testConfiguration() ;			break ;
				case 	2	: 	testDatabase() ;				// No break, we want cascade behavior
				case 	3	: 	new ThreadManager(this.boudoir).printList() ; 	break ;
				case 	4	: 	keepAlive(args) ;				break ;
				case 	5	: 	testBenchmark() ;				break ;
				case 	6	: 	testLoadPool(6 ,1) ;			break ;
				case 	7	: 	testMockedPool(buffer) ;		break ;
				case 	8	: 	testCodeBlock(args) ;			break ;
				default		: 	System.out.println(leadingName + "WRONG TEST OPTION: please specify a number from 0 to 7") ;
			}
		}
		// Executing a query or a query-delayed combo
		else if ( option.equalsIgnoreCase("-q") || combo.equalsIgnoreCase("-qt") ) {
			testQuery(buffer, combo) ;
		}
		// Emulating a complete library usage
		else if (option.equalsIgnoreCase("-m")) {
			int method =  getIntValue(args, 1) ;
			switch(method) {
				case	1	:	testMethod(1, buffer) ;										break ;
				case	2	:	testMethod(2, buffer) ;										break ;
				default		:	System.out.println("Please specify the method's number") ;
			}
			System.exit(0) ;
		}
		// Emulating a complete library usage
		else if (option.equalsIgnoreCase("-a")) {
			testAll(buffer) ;
		}
		// Invalid option
		else {
			System.out.println("Invalid option \n") ;
		}
	}

// *********************************************************************************************************
// **********************************			UNIT TEST 			****************************************
// *********************************************************************************************************

    /**
	* Test the logging system.
    */
    public void testLogger() {

		// Initializing test
		System.out.println() ;
		System.out.println(leadingName + "Testing logger: " + this.logger) ;

		// Logging test
		logger.logInfo		( "testing information level"	) ;
		logger.logWarning	( "testing warning level"		) ;
		logger.logSevere	( "testing severe level"		) ;
		if (logger.isInfoEnabled()) System.out.println(leadingName + "Check 'if isInfoEnabled()' is true") ;
		else System.out.println(leadingName + "Check 'if isInfoEnabled()' is false") ;
	}

    /**
	* Test method used to verify that boudoir-oriented configuration of the Simple Application  
	* (see the <code>SimpleLogic</code> class) is loaded correctly from the configuration files. 
    */
    public void testConfiguration() {
    	
		// Initializing test
		System.out.println() ;
		System.out.println(leadingName + "Testing configuration") ;
		
    	// Initializing a new custom Boudoir instance		
    	Boudoir myBoudoir = new Boudoir("test") ;
    	myBoudoir.setAnalyzer( new SimpleAnalyzer(myBoudoir) ) ;

		// Initializing the Simple Application
		SimpleLogic 	simpleLogic 	= new SimpleLogic(myBoudoir) ;
		simpleLogic.showConfiguration() ;
	}

    /**
	* Test method used to verify that the static code block in the <code>Database</code> class is found
	* and correctly executed. Moreover, when this method is invoked, some meta-data about the database's
	* connections and related pools are printed on the STDOUT.
    */
    public void testDatabase() {
    	
		// 1. Initializing test
		System.out.println() ;
		System.out.println(leadingName + "Testing database") ;

		// 2. Forcing the VM to load the Database CLASS (not the instance)
		Class 		dbClass		= null ;
		Database 	database	= boudoir.getDatabase()  ;
		try									{ dbClass = Class.forName("adriani.boudoir.db.Database") ;		}
		catch(ClassNotFoundException e)		{ e.printStackTrace() ;											}
		System.out.println(leadingName + "Loaded class: " 			+ dbClass								) ;
		System.out.println(leadingName + "DB_01 connections pool: " + database.getJDBCPool(Database.DB_01)	) ;
		//System.out.println(leadingName + "DB_02 connections pool: " + Database.getJDBCPool(Database.DB_02)) ;

		// 3. Getting meta-data from the database
		StringBuffer msg_01 = database.showMetaData( Database.DB_01 ) ;
		//StringBuffer msg_02 = Database.showMetaData(Database.DB_02) ;
		
		// 4. Showing meta-data about the database		
		System.out.println(msg_01.toString()) ;
		//System.out.println(msg_02.toString()) ;
	}

    /**
	* Stop execution of current thread for the number of seconds specified by the 3rd element of the
	* input array (i.e. the <code>args[2]</code> string value). This method should be used only for 
	* testing or debug. For instance, it could help testing the rotation time of the logging system,
	* by keeping alive execution the time necessary to switch the system date.<br/><br/>
	* 
	* To test this method run the <code>SimpleClient</code> using the <code>-t 4</code> option.
	*
	* @param args the input <code>args</code> string array
	* @see keepAlive(int delay)
    */
    public void keepAlive(String[] args) {

    	int delay = getIntValue(args, 2) ;
    	keepAlive(delay) ;
	}
    
    /**
	* Stop execution of current thread for the number of seconds specified by the input parameter.<br/>
	* This method should be used only for testing or debug. 
	* For instance, it could help testing the rotation time of the logging system,
	* by keeping alive execution for the time necessary to switch the system date.<br/><br/>
	* 
	* @param delay the number of seconds current thread will be stopped
	* @see keepAlive(String[] args) 
    */
    public void keepAlive(int delay) {
    	
		System.out.println(leadingName + "Keep alive parameter = " + delay + " seconds") ;
		if (delay > 0) {
			if (logger.isInfoEnabled()) logger.logInfo(leadingName + "Stopping current thread for " + delay + " s") ;
			new GeneralApplication(this.boudoir).delay(delay*1000) ; 
		} else {
			if (logger.isInfoEnabled()) logger.logInfo(leadingName + "Skipped delay (found 0 value)s") ;
		}
		if (logger.isInfoEnabled()) logger.logInfo(leadingName + "Restarting thread after " + delay + " s") ;
    }
    
    /**
	* Verify that the benchmark feature can be activated at runtime, performing following steps:<br/>:<br/>
	* 
	* <ol>
	* <li>Crates, initializes and invokes the Simple Application (see <code>SimpleLogic</code>)</li> 
	* <li>Changes the benchmark status invoking the Remote Command Manager feature</li> 
	* <li>Invokes again the Simple Application (see <code>SimpleLogic</code>)</li> 
	* </ol>
	* 
	* Expected result: the benchmark log file (e.g. "/home/temp/my-logs/1/analysis_dd_hh_mm.txt") should contain
	* only ONE benchmark entry (and not two, even if we invoked the application twice). 
    */
    public void testBenchmark() {

    	// Initializing a new custom Boudoir instance
    	Boudoir myBoudoir = new Boudoir("test") ;
    	myBoudoir.setAnalyzer( new SimpleAnalyzer(myBoudoir) ) ;

		// Initializing the Simple Application
		SimpleLogic 	simpleLogic 	= new SimpleLogic(myBoudoir) ;

		// Invoking the 1st method
		StringBuffer[]	input			= getInput(null, 1) ;		
		Wrapper 		customerWrapper = simpleLogic.getCustomerDummy(input) ;
		logger.logInfo( getCustomerMessage(input, customerWrapper)	) ;
		
		// Changing the Benchmark status via the Remote Command Manager
		String			prefix			= boudoir.getConfiguration().COMMAND ;
		StringBuffer[]	command			= new StringBuffer[] { new StringBuffer(prefix + "10") } ; 
						customerWrapper = simpleLogic.getCustomerDummy(command) ;
		logger.logInfo( getCustomerMessage(input, customerWrapper)	) ;	
		
		// Note: we could easily run the command above using (but we wanted to emulate external command invocation)
		//simpleLogic.manageCommand(10) ;
		
		// Invoking the 1st method
						customerWrapper = simpleLogic.getCustomerDummy(input) ;
		logger.logInfo( getCustomerMessage(input, customerWrapper)	) ;
	}
    
    /**
	* Run a complete loading test on the JDBC connection pool, by executing <code>num</code> concurrent calls
	* to the Simple Application (see <code>SimpleLogic</code>), and stopping each thread for <code>sec</code>
	* seconds (in order to avoid using always the same connection).<br/><br/>
	* 
	* ATTENTION: this test require 10-20 seconds to be completed, depending by the cleanup time set in the
	* configuration file (suggested value: 5 seconds).<br/><br/>
	* 
	* EXPECTED RESULT: results of this test strongly depend by the Boudoir test configuration file. 
	* To verify test results checks the log file during execution. We should have:<br/><br/>
	* <ul>
	* <li>First logging pause: all pool objects should be LOCKED</li>
	* <li>Second logging pause: all pool objects should be UNLOCKED</li>
	* <li>Third logging pause: pool should be EMPTY</li>
	* <li>if <code>num</code> param exceeds pool capacity, we should have errors due to pool saturation</li>
	* </ul>
	* 
	* <strong>Multithreading approach</strong>: all the layers behind the <code>SimpleLogic</code> ARE NOT thread safe. 
	* For example, the <code>SimpleLogic</code>, <code>SimpleMacro</code> and <code>SimpleDAO</code> classes 
	* store call status and the DB connection reference as class fields, then they CAN NOT be used in multithreading
	* environment, because each thread could have his own call status and DB connection.<br/>
	* Only the <code>Database</code> object can be shared among different threads, and for it's obtained using
	* <code>boudoir.getDatabase()</code>: this mechanism allows to use the JDBC pool even in a multithreading
	* environment, if the client takes care to instantiate a single <code>SimpleLogic</code> object for 
	* each single thread (see implementation of this method).<br/><br/>
	* 
	* @param num number of concurrent threads accessing the Simple Application
	* @param sec number of seconds (for each single call) to be stopped (keeping the DB connection locked)
	*/
    public void testLoadPool(int num, int sec) {

    	// Initializing a new custom Boudoir instance
    	Boudoir myBoudoir = new Boudoir("test") ;
    	myBoudoir.setAnalyzer( new SimpleAnalyzer(myBoudoir) ) ;

		// Setting delay time for each call (allows to lock different connections in the JDBC pool)
		SimpleLogic.DELAY 				= 3000*sec ;

		// TIME 0: filling the pool by repeating invocation of the application method 
		StringBuffer[]	input			= getInput(null, 1) ;
		ClientThread[]	threads			= new ClientThread[num] ;
		for (int i = 0 ; i < num ; i++) {
			threads[i] = new ClientThread(myBoudoir) ;
			// REMARK: we MUST use a different SimpleLogic instance for each thread (see method description),
			// but if all these objects must share the same JDBC pool, they MUST rely on the same Boudoir object. 
			threads[i].setTarget(new SimpleLogic(myBoudoir), input) ;
			threads[i].start();
		}

		// TIME 1: checking pool status - now the pool should be full 
		// (waiting 1 second to make sure all threads opened a connection)
		// Expected result: all pool objects should be LOCKED 
		keepAlive(1) ;
		SimpleLogic simpleLogic = new SimpleLogic(myBoudoir) ;		
		simpleLogic.manageCommand(11) ;
		
		// TIME sec+2 : waiting few seconds to return all connections to the JDBC connections pool
		// Expected result: all pool objects should be UNLOCKED
		keepAlive(sec+2) ;
		simpleLogic.manageCommand(11) ;
		
		// TIME sec+11: waiting 10 seconds to clean the JDBC pool (depends by Boudoir test configuration file)
		// Expected result: pool should be EMPTY
		keepAlive(10) ;
		simpleLogic.manageCommand(11) ;

		// Expected result: 5 calls succesfull, 1 call failed (depends by Boudoir test configuration file)
		simpleLogic.manageCommand(5) ;
	}

    /**
	* Test the "Mocked Connections" features, i.e. the possibility to wrap the JDBC connections 
	* returned by the driver using the <code>MockConnection</code> object, and, in the same way,
	* to wrap the standard JDBC pool using the <code>MockConnectionPool</code> class.<br/>
	* Actually, the <code>MockConnectionPool</code> class is ALWAYS used in the Boudoir paradigma,
	* leaving the proper "use_mocked" flag to false. Hence by enabling the Mocked Connection
	* feature we just mean enabling the proper flag within the  <code>MockConnectionPool</code> class.<br/><br/>
	* 
	* @param args the input <code>args</code> string array
    */
    public void testMockedPool(StringBuffer[] params) {
    	
    	// Creating a new custom Boudoir instance, different from that used by the SimpleClient
    	Boudoir myBoudoir = new Boudoir("test") ;
    	
    	// Setting a custom Analyzer for this application
    	SimpleAnalyzer myAnalyzer = new SimpleAnalyzer(myBoudoir) ;
    	myBoudoir.setAnalyzer(myAnalyzer) ;

		// Creating Simple Application
		SimpleLogic 	simpleLogic 	= new SimpleLogic(myBoudoir) ;
		
		// Setting hard-coded test value		
		StringBuffer[]	input			= getInput(null, 1) ;			// Hard-coded input
		StringBuffer[]	filtered		= adjustInput(params) ;			// Filters options
		
		// Keep for testing or debug: checks input
//		 for (StringBuffer s : params) 		System.out.println(leadingName + "Params:" + s) ;
//		 for (StringBuffer s : filtered)	System.out.println(leadingName + "Filtered:" + s) ;
//		 for (StringBuffer s : input) 		System.out.println(leadingName + "input:" + s) ;		
		
		// Enabling Mocked Connections automatic test
		if ( compareSB(filtered[0], new StringBuffer("y")) ) {
			myBoudoir.getDatabase().setMockedPool(true) ;
			logger.logInfo("MOCKED CONNECTIONS: enabled") ;
		} else {
			if (myBoudoir.getDatabase().usesMockedPool() ) {
				myBoudoir.getDatabase().setMockedPool(false) ;
			}
			logger.logInfo("MOCKED CONNECTIONS: disabled") ;
		}
		
		// Activating automatic test
		simpleLogic.getSimpleMacro().TEST_CASE 	= SimpleLogic.MOCKED ;		
		
		// Response object shared by all following calls
		Wrapper	customerWrapper ;
		
		// Keep for testing or debug: checks that the mocked feature has been enable
		String			prefix			= boudoir.getConfiguration().COMMAND ;
		StringBuffer[]	command			= { new StringBuffer(prefix + 4) } ;
						customerWrapper = simpleLogic.getCustomerDummy(command) ;
		logger.logInfo( getCustomerMessage(params, customerWrapper)	) ;
		
		// First call
		customerWrapper = simpleLogic.getCustomerDummy(input) ;
		logger.logInfo( getCustomerMessage(params, customerWrapper)	) ;
		
		// Second call
		customerWrapper = simpleLogic.getCustomerDummy(input) ;
		logger.logInfo( getCustomerMessage(params, customerWrapper)	) ;
    }
    
    /**
	* General test, used as a place-holder to write brand new code-block. After testing, usually this code-block
	* is moved and renamed, creating a new method in another class (or is discharged).<br/><br/>
	* 
	* To test this method run the <code>SimpleClient</code> using the <code>-t 5</code> option.
	*
	* @param args the input <code>args</code> string array
    */
    public void testCodeBlock(String[] args) {

		System.out.println(leadingName + "Doing nothing ...") ;
	}
    
// *********************************************************************************************************
// **********************************		  TEST QUERIES			****************************************
// *********************************************************************************************************
    
    /**
	* Execute a simple dummy SQL query to verify the format of the JDBC result set retrieved from the database.
	* <br/>
	* Current implementation performs a simple SELECT query in the DB_02 database. The query to be executed
	* is selected by the <code>argument[1]</code> parameter, while the input parameter used  to fill the
	* statement (if available) is specified by the <code>argument[2]</code> parameter. In other words this
	* method must be called using following syntax:<br/><br/>
	*
	* <code>SimpleClient -q n p</code><br/><br/>
	*
	* The value of the parameter <code>n</code> select of the following DAO methods:<br/><br/>
	*
	* 0	: getRecord(target, label, query, 1)<br/>
	* 1	: getRecord(target, label, query, 4)<br/>
	* 2	: getRecord(target, label, query, 4)<br/>
	* 3	: getRecordSet(target, label, query, 2)<br/>
	* 4	: getFunction(target, label, query, param_1, type)<br/>
	* 5	: getProcedure(target, label, query, param_2, 1, inout)<br/>
	* 6	: getFunctionPipe(target, label, query, 2)<br/>
	* <br/>
	*
	* @param buffer the input string array, converted into a <code>StringBuffer</code> object
	* @param combo optional 2nd option, allowing combination of options (e.g. -t and -q becomes -tq or -qt)
	* @see GeneralDAO#getRecord(int targetDB, String label, String query, int cols)
	* @see GeneralDAO#getRecordSet(int targetDB, String label, String query, int cols)
	* @see GeneralDAO#getFunction(int targetDB, String label, String query, Object[] param, int type) 
	* @see GeneralDAO#getProcedure(int targetDB, String label, String query, Object[] param, int cols, int[] inout)
	* @see GeneralDAO#getFunctionPipe(int targetDB, String label, String query, int cols)
    */
    public void testQuery(StringBuffer[] buffer, String combo) {

		// 1. Creating the DBA environment
		int			test		= 0 ;
		String 		key 		= "1" ;					// Default value
		int 		target		= Database.DB_01 ;
		SimpleDAO 	simpleDAO	= new SimpleDAO(this.boudoir) ;
		String 		label 		= "db_01.testQuery: " ;

		// 2.1 Setting input parameters
		if (buffer != null) {
			if (buffer.length > 1) test = getIntValue(buffer, 1) ;
			if (buffer.length > 2) key	= buffer[2].toString() ;
			System.out.println(leadingName + "Input testQuery() params: " + test + BLANK + key) ;
		}
		// 2.2 Setting default parameters for procedures/functions
		int[] 			inout 	= { 2 } ;
		Object[]		param_1	= new Object[] 	{ new StringBuffer("A general entity")	, new Integer(4) 		} ;
		Object[]		param_2	= new Object[] 	{ new Integer(4)						, new StringBuffer("0")	} ;
		// Optionally overwriting the above default values (only for test nr. 4 and 5)
		if ( (buffer != null) && (buffer.length > 2) ) {
			param_1 = new Object[] { null				, new Integer(key)		} ;
			param_2 = new Object[] { new Integer(key) 	, new StringBuffer("0")	} ;
		}
		// 2.3 Customer parameter (should be required only by PostegreSQL)
		int type = java.sql.Types.INTEGER ;
		
		// 3. Checking if we have COMBO OPTIONS.
		// Remark: this test it's just an easy way to test the benchmark feature. An alternate approach
		// (more complete) is offered by the Simple Application (option -a): see SimpleClient.
		if ( (combo != null) && (combo.length() > 0) ) {
			if ( combo.equalsIgnoreCase("-qt") ) {
				// For the time being we stop current thread for a fixed amount of time.
				// We chose this solution to avoid too many messy input parameters for the "-q" option.
				keepAlive(120) ;	
			}
		}

		// 4. Preparing statement
		String query = null ;
		switch(test) {
			case 0	:	query = "SELECT SYSDATE FROM DUAL" ;											break ;
			case 1	:	query = "SELECT * FROM TEST_CUSTOMERS WHERE CUSTOMER_ID = '"  + key + "'" ;		break ;
			case 2	:	query = "SELECT * FROM TEST_ENTITY_TIMES WHERE ID_ENTITY = '" + key + "'" ;		break ;
			case 3	:	query = "SELECT * FROM TEST_ENTITY_TIMES WHERE ID_ENTITY < '" + key + "'" ;		break ;
			case 4	:	query = "{? = call TEST_GET_ENTITY_TIME(?, ?)}" ;								break ;
			case 5	:	query = "{call TEST_VERIFY_ENTITY_ID(?,?)}" ;									break ;			
			case 6	:	query = "SELECT * FROM TABLE(TEST_GET_ALL_TIMES('Benchmark_query_1', 5))" ;		break ;
			default	:	query = "SELECT SYSDATE FROM DUAL" ;
		}
		System.out.println(leadingName + "Prepared query: " + query) ;

		// 5. Opening connection to the target database
		if (simpleDAO.openConnection(target)) {

			// 5.1 Executing statement
			Object result = null ;
			switch(test) {
				case 0	:	result = simpleDAO.getRecord(target, label, query, 1) ;					break ;
				case 1	:	result = simpleDAO.getRecord(target, label, query, 4) ;					break ;
				case 2	:	result = simpleDAO.getRecord(target, label, query, 4) ;					break ;
				case 3	:	result = simpleDAO.getRecordSet(target, label, query, 2) ;					break ;
				case 4	:	result = simpleDAO.getFunction(target, label, query, param_1, type) ;		break ;
				case 5	:	result = simpleDAO.getProcedure(target, label, query, param_2, 1, inout) ;	break ;
				case 6	:	result = simpleDAO.getFunctionPipe(target, label, query, 2) ;				break ;				
				default	:	result = simpleDAO.getRecord(target, label, query, 1) ;
			}

			// 5.2 Returning connection
			simpleDAO.closeConnection(target) ;

			// 5.3: 1st case (single record)
			if ((result != null) && (result instanceof StringBuffer[])) {
				System.out.print("\nFound record: ") ;
				StringBuffer[] record = (StringBuffer[]) result ;
				for (int i = 0 ; i < record.length ; i++) {
					if (record[i] != null) System.out.print(record[i].toString() + BLANK) ;
				}
			// 5.4: 2nd case (record set)
			} else if ((result != null) && (result instanceof StringBuffer[][])) {
				System.out.println("\nFound record set:") ;
				StringBuffer[][] set = (StringBuffer[][]) result ;
				for (int n = 0 ; n < set.length ; n++) {
					if (set[n] != null) for (int i = 0 ; i < set[n].length ; i++) {
						if (set[n][i] != null) System.out.print(set[n][i].toString() + BLANK) ;
					}
					System.out.println() ;
				}
			// 5.5: 3rd case (no result)
			} else if (result == null) System.out.println("\nTest query with no result") ;
			System.out.println() ;
			
		// 6. Handling no database 
		} else System.out.println(leadingName + "No connections to target database") ;
	}
    
// *********************************************************************************************************
// **********************************		  TEST APPLICATION		****************************************
// *********************************************************************************************************
    
    /**
	* Executes the main method of the dummy simple application included into the <em>test</em> package.<br/>
	* The dummy application is composed by following classes:<br/><br/>
	* 
	* <code>SimpleLogic</code>: the application interface exposing the high-level method<br/>
	* <code>SimpleMacro</code>: the middle layer managing DAO objects and DB connections<br/>
	* <code>SimpleDAO</code>: the bottom layer where SQL statement are executed<br/>
	* <code>SimpleAnalyzer</code>: the custom version of the SimpleAnalyzer class<br/> 
	* <code>CustomerWrapper</code>: a simple DTO wrapping results of the<code>SimpleLogic</code> class<br/><br/>
	* 
	* Application results are logged into the log file (and to STDOUT if the logger it's configured to do so).
	*
	* @param method the integer index of the method to be tested (in current implementation: 1 or 2)
	* @param params the input <code>args</code> string array
    */
    public void testMethod(int method, StringBuffer[] params) {
    	
    	// Creating a new custom Boudoir instance, different from that used by the SimpleClient
    	Boudoir myBoudoir = new Boudoir("test") ;
    	
    	// Setting a custom Analyzer for this application
    	SimpleAnalyzer myAnalyzer = new SimpleAnalyzer(myBoudoir) ;
    	myBoudoir.setAnalyzer(myAnalyzer) ;

		// Creating Simple Application
		SimpleLogic 	simpleLogic 	= new SimpleLogic(myBoudoir) ;
		
		// Adjusting input parameters
		StringBuffer[]	filtered		= adjustInput(params) ;
		
		// Keep for testing or debug: check input parameters
		// for (int i = 0 ; i < input.length ; i++) System.out.println(leadingName + i + ") " + input[i]) ;
		
		// Invoking Simple Application
		switch(method) {
			case	1	:		StringBuffer[]	input_1				= getInput(filtered, 1) ;
								Wrapper 		customerWrapper 	= simpleLogic.getCustomerDummy(input_1) ;
								logger.logInfo( getCustomerMessage(params, customerWrapper)	) ;
								break ;
			case	2	:		StringBuffer[]	input_2				= getInput(filtered, 10) ;
								Wrapper 		entityTimeWrapper 	= simpleLogic.getEntityTimesDummy(input_2) ;
								logger.logInfo( getTimesMessage(params, entityTimeWrapper)	) ;
								break ;
		}
	}
    
    /**
	* Run a complete functional-test, exploiting the dummy simple application included into the 
	* <em>test</em> package. After execution of a test a detailed report is logged.<br/><br/>
	*  
	* The dummy application is intended only as an example of usage of the Boudoir library, and does not offer 
	* any actual feature. Its main purpose is to allow testing those Boudoir features that makes sense
	* only when the library is used by an application. Such features are<br/>  
	* 
	* <ul>
	* <li>SimpleAnalyzer feature, creating usage statistics</li>
	* <li>Benchmark feature, logging execution times of SQL statement</li>
	* <li>Exit condition management, such as error handling and validation issues</li>
	* </ul>
	*
	* Application results are logged into the log file (and to STDOUT if the logger it's configured to do so). 
	* Expected values:<br/><br/>
	* 
	* - All counters set to 1<br/> 
	* - "No results" counter set to 2<br/> 
	*
	* @param params the input <code>args</code> string array
    */
    public void testAll(StringBuffer[] params) {
    	
    	// Creating a new custom Boudoir instance, different from that used by the SimpleClient
    	Boudoir myBoudoir = new Boudoir("test") ;
    	
    	// Setting a custom Analyzer for this application
    	SimpleAnalyzer myAnalyzer = new SimpleAnalyzer(myBoudoir) ;
    	myBoudoir.setAnalyzer(myAnalyzer) ;

		// Creating Simple Application
		SimpleLogic 	simpleLogic 	= new SimpleLogic(myBoudoir) ;
		
		// Adjusting input parameters		
		StringBuffer[]	filtered		= adjustInput(params) ;			// Filters options
		StringBuffer[]	input			= getInput(filtered, 1) ;		// Add hard-coded values
		
		// Keep for testing or debug: checks input
//		 for (StringBuffer s : params) 		System.out.println(leadingName + "Params:" + s) ;
//		 for (StringBuffer s : filtered)	System.out.println(leadingName + "Filtered:" + s) ;
//		 for (StringBuffer s : input) 		System.out.println(leadingName + "Input:" + s) ;
				
		// Emulating normal usage
		Wrapper 		customerWrapper = simpleLogic.getCustomerDummy(input) ;
		logger.logInfo( getCustomerMessage(params, customerWrapper)	) ;

		// Keep for testing or debug: quits after first execution
		// System.exit(0) ;
		
		// Emulating command manager
		String			prefix			= boudoir.getConfiguration().COMMAND ;
		StringBuffer[]	command			= { new StringBuffer(prefix + 1) } ;
		customerWrapper 				= simpleLogic.getCustomerDummy(command) ;
		logger.logInfo( getCustomerMessage(params, customerWrapper)	) ;
		
		// Emulating NO RESULT case (unexpected case, when SimpleMacro.status = null)
		simpleLogic.TEST_CASE 			= SimpleLogic.NO_RESULT ;
		customerWrapper 				= simpleLogic.getCustomerDummy(input) ;
		logger.logInfo( getCustomerMessage(params, customerWrapper)	) ;
		
		// Emulating NO RESULT case (expected case, when SimpleMacro.status = NO_DATA)
		StringBuffer[]	missing			= { new StringBuffer("1234567890") } ; 
		customerWrapper 				= simpleLogic.getCustomerDummy(missing) ;
		logger.logInfo( getCustomerMessage(params, customerWrapper)	) ;
		
		// Emulating LOGIC ERROR case
		simpleLogic.TEST_CASE 			= SimpleLogic.LOGIC_ERROR ;
		customerWrapper 				= simpleLogic.getCustomerDummy(input) ;
		logger.logInfo( getCustomerMessage(params, customerWrapper)	) ;
		
		// Emulating FATAL case
		try {
			simpleLogic.TEST_CASE 		= SimpleLogic.FATAL ;
			customerWrapper 			= simpleLogic.getCustomerDummy(input) ;
			logger.logInfo( getCustomerMessage(params, customerWrapper)	) ;
		} catch(Exception e) {
			logger.logSevere(leadingName + "SIMPLE CLIENT CRASHED\n") ;
		} finally {
			simpleLogic.TEST_CASE 		= SimpleLogic.NO_TEST ;			
		}
		
		// Emulating VALIDATION ERROR case (at Macro layer)
		simpleLogic.getSimpleMacro().TEST_CASE 	= SimpleLogic.VALIDATION_ERROR ;
		customerWrapper 						= simpleLogic.getCustomerDummy(input) ;
		logger.logInfo( getCustomerMessage(params, customerWrapper)	) ;
		
		// Emulating MACRO ERROR case (at Macro layer)
		simpleLogic.getSimpleMacro().TEST_CASE 	= SimpleLogic.MACRO_ERROR ;
		customerWrapper 						= simpleLogic.getCustomerDummy(input) ;
		logger.logInfo( getCustomerMessage(params, customerWrapper)	) ;
		
		// Resetting error conditions
		simpleLogic.getSimpleMacro().TEST_CASE 	= SimpleLogic.NO_TEST ;
		
		// Emulating DATABASE CONNECTION ERROR
		Database database						= myBoudoir.getDatabase() ;
		database.init("art", "xartdullt2") ;
		simpleLogic.getSimpleMacro().getServiceDAO().resetPool() ;
		customerWrapper							= simpleLogic.getCustomerDummy(input) ;
		logger.logInfo( getCustomerMessage(params, customerWrapper)	) ;
		
		// Recalling expected results
		StringBuffer expected = new StringBuffer() ;
		expected.append(" ********* EXPECTED RESULTS *********\n\n") ;
		expected.append("- Total calls to the library: 9\n") ;
		expected.append("- All \"Calls counters\" set to 1, except\n") ;
		expected.append("- \"No result counter\" set to 2\n") ;
		expected.append("*******************************************\n\n") ;
		logger.logInfo(expected.toString()) ;
		
		// Logging run analysis
		simpleLogic.manageCommand(5) ;
    }
    
// *********************************************************************************************************
// ***************************			FORMATTING INPUT/OUTPUT			************************************
// *********************************************************************************************************
    
    /**
    * Adjust input parameters when testing the library using the <code>SimpleLogic</code> dummy application.
    * This method allows to filter "options-like" element from the input array. For example, if the input
    * array is <code>{"-m", "1", "abc"}</code> this method should return <code>{"1", "abc"}</code>.
    * 
    * <strong>Remark</strong>: this method filter the input assuming each option is followed by an argument.
    * For instance, the "-m" option should always be followed by a number. For this reason here we remove TWO
    * elements from the array if the first element is an option.
    * 
    * @param params the <code>StringBuffer</code> array to be used as input
    * @return the input <code>StringBuffer</code> without the options-like elements
    */
    public StringBuffer[] adjustInput(StringBuffer[] params) {
    	
    	if ( (params != null) && (params.length > 0) && (params[0].charAt(0) == '-') ) {
    		int skip = 0 ;
    		if (params[0].charAt(1) == 'a') skip = 1 ;
    		if (params[0].charAt(1) == 'm') skip = 2 ;
    		StringBuffer[] filtered = new StringBuffer[params.length - skip] ;
    		for (int i = 0 ; i < filtered.length ; i++) {
    			filtered[0] = new StringBuffer(params[i + skip]) ;
    		}
    		return filtered ;
		} 
    	return params ;
    }
    
    /**
    * Produces input parameters when testing the library using the <code>SimpleLogic</code> dummy application.
    * Input parameters are read from STDIN (if available) or filled with hard-coded default values (if STDIN 
    * is empty).
    * 
    * @param params the <code>StringBuffer</code> array retrieved from STDIN
    * @param index an index identifying which hard-coded value must be used if STDIN is empty
    * @return the input <code>StringBuffer</code>, or an array filled with default values
    */
    public StringBuffer[] getInput(StringBuffer[] params, int index) {
    	
    	StringBuffer[] input = null ;
		if ( (params != null) && (params.length > 0) ) { 
			input = new StringBuffer[] { params[0] } ;
		} else {
			switch(index) {
				// Hard-coded values for the 1st method
				case 0	:	input = new StringBuffer[] { new StringBuffer("0498978842") } ;
							break ;
				case 1	:	input = new StringBuffer[] { new StringBuffer("0431999536") } ;
							break ;
				case 2	:	input = new StringBuffer[] { new StringBuffer("010202404") } ;
							break ;
				// Hard-coded values for the 2nd method
				case 10	:	input = new StringBuffer[] { new StringBuffer("1") } ;
							break ;
				case 11	:	input = new StringBuffer[] { new StringBuffer("2") } ;
							break ;
				case 12	:	input = new StringBuffer[] { new StringBuffer("3") } ;
							break ;
				// Default value (assuming the default method)
				default	:	input = new StringBuffer[] { new StringBuffer("0498978842") } ;
			}
		}
		return input ;
    }
    
    /**
    * Produces a formatted message containing a short log of wrapper content.
    * 
    * @param params the <code>StringBuffer</code> array wrapping the application input
    * @param output the <code>CustomerWrapper</code> object wrapping the application output
    * @return an already formatted message describing the application output
    */
    public String getCustomerMessage(StringBuffer[] params, Wrapper output) {
    	
    	CustomerWrapper wrapper = (CustomerWrapper) output ; 
    	
		StringBuffer message = new StringBuffer() ;
		message.append(	"\n\nSimple Application results for input: "	) ;
		for (StringBuffer par : params) {
			message.append(	par.toString()								) ;
			message.append(	" "											) ;
		}
		message.append(	"\n\n"											) ;
		message.append(	"NAME        : "								) ;
		message.append(	wrapper.getName()								) ;
		message.append(	"\n"											) ;
		message.append(	"SURNAME     : "								) ;
		message.append(	wrapper.getSurname()							) ;
		message.append(	"\n"											) ;		
    	return message.toString() ;
    }
    
    /**
    * Produces a formatted message containing a short log of wrapper content.
    * 
    * @param params the <code>StringBuffer</code> array wrapping the application input
    * @param output the <code>CustomerWrapper</code> object wrapping the application output
    * @return an already formatted message describing the application output
    */
    public String getTimesMessage(StringBuffer[] params, Wrapper output) {
    	
    	EntityTimeWrapper wrapper = (EntityTimeWrapper) output ; 
    	
		StringBuffer message = new StringBuffer() ;
		message.append(	"\n\nSimple Application results for input: "	) ;
		for (StringBuffer par : params) {
			message.append(	par.toString()								) ;
			message.append(	" "											) ;
		}
		message.append(	"\n\n"											) ;
		message.append(	"TIME        : "								) ;
		message.append(	wrapper.getTime()								) ;
		message.append(	"\n"											) ;
    	return message.toString() ;
    }
    
// *********************************************************************************************************
// **********************************		COMMAND LINE			****************************************
// *********************************************************************************************************
 	
	/**
	* Print some help about the <code>SimpleClient</code> command line options on the STDOUT.
	*
	* @return the formatted string message printed on STDOUT
	*/
	public StringBuffer printHelp() {
		
		// Getting Settings instance
		Settings settings = this.boudoir.getConfiguration() ; 

		// Preparing text message
 		StringBuffer message = new StringBuffer() ;
 		message.append("\n") ;
 		message.append("Boudoir client - Invoking the library's API from the command line, ") ;
 		message.append("version " + settings.VERSION + "\n") ;
 		message.append("\n") ;
 		message.append("Usage: java -jar boudoir.jar [options] [ARGS]\n") ;
 		message.append("or     java -cp %CLASSPATH% adriani.boudoir.test.SimpleClient [options] [ARGS]\n") ;
 		message.append("\n") ;
 		message.append("ARGS : the sequence of input parameters request by the invoked method\n") ;
 		message.append("\n\n") ;
 		message.append("OPTIONS\n") ;
 		message.append("-h          print this HELP screen\n") ;
 		message.append("-l          print the LIST of all the available \"Remote Manager Commands\"\n") ;
 		message.append("-c file     CREATE the bd_data.properties file using the specified file as source\n") ;
 		message.append("-d          DUMP the keywords properties into the bd_data.java file\n") ;
 		message.append("-t n        executes the n-th general TEST offered by the SimpleClient class\n") ;
 		message.append("-q n param  executes the n-th SQL QUERY offered by the SimpleClient class\n") ;
 		message.append("-m param    emulates a call to a simple METHOD of a dummy application (see SimpleLogic)\n") ;
 		message.append("-a param    executes ALL functional test for the boudoir library (see SimpleLogic)\n") ;
 		message.append("\n\n") ;
 		message.append("TESTING (-t | -q) : see the API documentations (JavaDocs) for details.\n") ;

 		// Producing output
 		System.out.println(message) ;
 		return message ;
 	}
	
// *********************************************************************************************************
// **********************************		ACCESSORS			********************************************
// *********************************************************************************************************

	/**
	* Getter method.
	*/
	public Boudoir getBoidoirInstance() 					{			return this.boudoir ;				}

// *********************************************************************************************************
// **********************************		UTILITIES			********************************************
// *********************************************************************************************************

	/**
	* Parse an element of the input array, returning the parsed integer value.
	*
	* @param string the input string array, converted into a <code>String</code> object
	* @param index the index of the element to be extracted
	* @return the integer number corresponding to the index-th element of the array
    */
    public int getIntValue(String[] string, int index) {

		int code = 0 ;
		if (string.length <= index) return code ;
		try 				{ code = Integer.parseInt( string[index].trim() ) ;					}
		catch(Exception e) 	{ code = 0 ;														}
		return code ;
	}

    /**
	* Parse an element of the input array, returning the parsed integer value.
	*
	* @param buffer the input string array, converted into a <code>StringBuffer</code> object
	* @param index the index of the element to be extracted
	* @return the integer number corresponding to the index-th element of the array
    */
    public int getIntValue(StringBuffer[] buffer, int index) {

		int code = 0 ;
		if (buffer.length <= index) return code ;
		try 				{ code = Integer.parseInt( buffer[index].toString().trim() ) ;		}
		catch(Exception e) 	{ code = 0 ;														}
		return code ;
	}
    
    /**
	* Parse an element of the input string, returning the parsed date value.
	*
	* @param value the input date in string format
	* @return the parsed date
    */
	public java.util.Date getDateValue(StringBuffer value)  {

		Calendar calendar = this.getCalendarValue(value) ;
		return new java.util.Date( calendar.getTimeInMillis() ) ;
	}
	
    /**
	* Parse an element of the input string, returning the parsed calendar value.
	*
	* @param value the input date in string format
	* @return the parsed calendar
    */
	public Calendar getCalendarValue(StringBuffer value)  {

		try {
			return Utilities.getCalendar( value.toString() ) ;
		} catch (ParseException e) {
			logger.logWarning(leadingName + "Parse exception: " + value) ;
			return null ;
		} catch (Exception e) {
			logger.logWarning(leadingName + "Invalid calendar: " + value) ;
			return null ;
		}
	}
	
 	/**
 	* Helper method.
 	*
 	* @param first a <code>StringBuffer</code> object
 	* @param second another <code>StringBuffer</code> object
 	* @return <code>true</code> if the two string object contains the same text
 	*/
 	public boolean compareSB(StringBuffer first, StringBuffer second) {

 		String 	string = null ;
 		boolean result = false ;
 		if (first != null) 	string = first.toString() ;
 		if (string != null) result = string.equalsIgnoreCase( second.toString() ) ;
 		return result ;
 	}
	
} // End of class