/**
* Title: 		 SimpleDAO (legacy name: ServiceDAO)
* Description:	 Example of implementation of the Data Access Object layer
* Copyright:     Copyright (c) 2013
* Company:       Morpheus Quality Technologies
* Creation date: 03-12-2013
* Last modified: 11-02-2014
* @author 		 Stefano Adriani
* @version 		 1.0
*/

package adriani.boudoir.test ;

import adriani.boudoir.* ;
import adriani.boudoir.db.* ;
import adriani.boudoir.util.* ;
import adriani.boudoir.logging.WSLogger;

import adriani.boudoir.test.model.* ;

import java.sql.* ;
import java.util.ArrayList ;

/**
* The class wrapping all the SQL queries and statement necessary to retrieve data from the database(s).
* This class stores the necessary references to the database connections. These fields are not present in the
* <code>GeneralDAO</code> parent class because usually it's better to avoid storing DB connection as class fields.
* Here we store the connections to improve performances by sharing the same connection only when invoking
* a sequence of queries from the <code>SimpleMacro</code> class, usually during execution of the same method.<br/><br/>
*
* The connections are obtained from the <code>Database</code> class using the <code>MockConnection</code>
* wrapper, which allows to perform some specific test/debug operations and could improve performances when
* using prepared statements in a connection pooling environment.<br/>
* Note that it's up to the <code>Database</code> class to manage the connections pooling (if used), in order
* to uncouple the implementation of this class from the management of connections:
* no matters which pool class is used, a DAO class should always retrieve connections just calling:<br/><br/>
*
* <code>boudoir.getDatabase().getMockConnection(targetDB)</code><br/><br/>
*
* and returning them by calling:<br/><br/>
*
* <code>boudoir.getDatabase().returnMockConnection(targetDB, this.mock[targetDB])</code>
*/
public class SimpleDAO extends GeneralDAO {

  	/** A reference to the <code>SimpleMacro</code> object which created this DA0 (optional).				*/
	protected 				SimpleMacro 		simpleMacro  ;
	
  	/** SAMPLE query in string format used to create the corresponding prepared statement.					*/
	protected static final 	String 				GET_CUSTOMER 			= "SELECT NAME, SURNAME FROM TEST_CUSTOMERS WHERE TELEPHONE = ?" ;
  	/** SAMPLE query in string format used to create the corresponding prepared statement.					*/
	protected static final 	String 				GET_ENTITY_TIME			= "SELECT END_TIME FROM TEST_ENTITY_TIMES where id_entity = ?" ;

	/** Retrieving the logger instance to be used for this class.											*/
	protected 		 		WSLogger			logger ;
	
	/**
	* Alternate constructor, allowing to create a DAO object without the need
	* to create also a <code>SimpleMacro</code> instance.
	*
	* @param boudoir the <code>Boudoir</code> instance "owning" this object
    */
    public SimpleDAO(Boudoir boudoir) {
    	
    	super( boudoir ) ;
		this.leadingName 	= "SimpleDAO - " ;
		this.logger 		= boudoir.getLogger().getLogger(SimpleDAO.class) ;
    }	

	/**
	* Standard constructor, bounding this DAO class to the master <code>Boudoir</code> via 
	* the <code>SimpleMacro</code> object.
	*
	* @param macro the <code>SimpleMacro</code> class creating and using this DAO
    */
    public SimpleDAO(SimpleMacro macro) {

    	this( macro.getBoudoir() ) ;
		this.simpleMacro 	= macro ;
	}

// ************************************************************************************************************
// *** 		First sample: example of query retrieving customers from the TEST_CUSTOMERS table				***
// ************************************************************************************************************

	/**
	* Sample basic query: remove this method when developing actual DAO method. Usually we should have one
	* method like this for each query (or group of similar queries). 
	*
	* @param number the whole telephone number (without the international prefix)
	* @param customer the customer ID reference number (optional)
	* @return the customer (name, surname) pair as found in the database
    */
    public CustomerDTO[] getCustomerData(String number, String customer) {

		PreparedStatement 	pstmt 	= null ;
		String 				method = "mydb.getCustomerName: " ;

		try {
			//System.out.println(leadingName + "Preparing:#" + number + "#") ;
			pstmt = this.getConnection(Database.DB_01).prepareStatement(GET_CUSTOMER) ;
			pstmt.setString(1, number.toString()) ;
			StringBuffer[][] result = getPreparedRecordSet(pstmt, method, 2) ;
			if ( (result != null) && (result.length > 0) ) {
				CustomerDTO[] array = new CustomerDTO[result.length] ;
				for (int i = 0 ; i < result.length ; i++) {
					array[i] = new CustomerDTO(result[i]) ;
					//System.out.println(leadingName + " *** DTO " + i + " -> " + array[i].getName() ) ;
				}
				return array ;
			}
			else return null ;
		} catch(Exception e) {
			e.printStackTrace() ;
			logger.logSevere(leadingName + method + PSTMT_NOK + e) ;
			return null ;
		} finally {
			// This will be executed even if an exception occurs in the catch block, or after
			// the method stepped over the return statement
			try 				{ 	pstmt.close() ;															}
			catch(Exception e) 	{	logger.logSevere(leadingName + "Impossible to close pstmt " + pstmt) ;	}
		}
	}
	
// ************************************************************************************************************
 // *** 		Second sample: example of query retrieving times from the TEST_ENTITY_TIMES table			***
// ************************************************************************************************************
    
	/**
	* Sample basic query: remove this method when developing actual DAO method. Usually we should have one
	* method like this for each query (or group of similar queries). 
	*
	* @param id the entity id reference
	* @return the DTO wrapping end time for the entity_time matching the input entity_id
    */
    public EntityTimeDTO getEntityTime(int id) {

		PreparedStatement 	pstmt 	= null ;
		String 				method = "mydb.getEntityTime: " ;

		try {
			//System.out.println(leadingName + "Preparing:#" + id + "#") ;
			pstmt = this.getConnection(Database.DB_01).prepareStatement(GET_ENTITY_TIME) ;
			pstmt.setInt(1, id) ;
			StringBuffer[] result = getPreparedRecord(pstmt, method, 1) ;
			if ( (result != null) && (result.length > 0) ) return new EntityTimeDTO(result) ;
			else return null ;
		} catch(Exception e) {
			e.printStackTrace() ;
			logger.logSevere(leadingName + method + PSTMT_NOK + e) ;
			return null ;
		} finally {
			// This will be executed even if an exception occurs in the catch block, or after
			// the method stepped over the return statement
			try 				{ 	pstmt.close() ;															}
			catch(Exception e) 	{	logger.logSevere(leadingName + "Impossible to close pstmt " + pstmt) ;	}
		}
	}	
    
// ************************************************************************************************************
// *************************************			TEST & DEBUG		***************************************
// ************************************************************************************************************

	/**
	* Produces a <code>StringBuffer[][]</code> object created from some hard-coded data for a general
	* SOAP method. This method should be used only during development to obtain an output from a SOAP
	* method as long as the actual database is not available.
	*
	* @param param the array containing all the input data
	* @param howMany the number of record to be retrieved
	* @param cols the number of columns for each record
	* @return a <code>StringBuffer[][]</code> object obtained from hard-coded data
    */
    public StringBuffer[][] getDummyQuery(StringBuffer[] param, int howMany, int cols) {

		// Emulating the "no data" situation
		if ((howMany == 0) || (cols == 0)) return null ;

		// Preparing variables
		String label = "fakeQuery" ;
		if (boudoir.getConfiguration().BENCH) startQuery(label) ;
		StringBuffer[][] output = new StringBuffer[howMany][cols] ;
		System.out.println(leadingName + "getDummyQuery parameters: " + param) ;

		// Eventually printing input parameters on the STDOUT
		if (boudoir.getConfiguration().VERBOSE) {
			System.out.print(leadingName + "Running fake query using input: ") ;
			for (int i = 0 ; i < param.length ; i++) System.out.print(param[i] + " ") ;
			System.out.println() ;
		}

		// Producing the whole result set
		StringBuffer[] 	record 	= null ;
		ArrayList 		list 	= new ArrayList() ;
		for (int n = 0 ; n < howMany ; n++) {
			record = new StringBuffer[cols] ;
			// Producing a single record
			for (int i = 0 ; i < cols ; i++) {
				int index = (i < param.length) ? i : (param.length - 1) ;
				if (n == 0)	{
					record[i] = new StringBuffer("dummy"					) ;
				} else {
					record[i] = new StringBuffer("Echo " + param[index]		) ;
				}
				//System.out.println(leadingName + "Produced fake output: " + i + "->" + record[i]) ;
			}
			list.add(record) ;
		}

		// Creating output array
		int size = list.size() ;
		output = (StringBuffer[][]) list.toArray( new StringBuffer[size][cols] ) ;

		// Eventually logging results
		if (boudoir.getConfiguration().BENCH) 		endQuery(label) ;
		if (boudoir.getConfiguration().ANALYZE)		boudoir.getAnalyzer().collect(ErrorMessage.OK) ;
		if (logger.isInfoEnabled()) 				logger.logInfo(leadingName + "Fake query produced: " + size + " records") ;

		return output ;
	}

	/**
	* Dummy query used to test the mocked pool mechanism if there is not any query using the
	* prepared statement approach.
	*
	* @return current sysdate from the database
    */
    public StringBuffer testPreparedStatement(int number) {
	
    	PreparedStatement 	pstmt 		= null ;
    	String				method		= "testPreparedStatement: " ;
    	String 				STATEMENT 	= "SELECT SYSDATE FROM DUAL WHERE 1 = ?" ; 
		if (boudoir.getConfiguration().VERBOSE) System.out.println("Testing prepared statement: " + STATEMENT + " with params: " + number) ;
		
		try {
			pstmt = mock[Database.DB_01].getPreparedStatement(STATEMENT) ;
			pstmt.setInt(1, number) ;
			return getPreparedRecord(pstmt, method, 1)[0] ;	
		} catch(Exception e) {
			logger.logSevere(leadingName + method + PSTMT_NOK + e) ;
			return null ;
		} finally {
			// This will be executed even if an exception occurs in the catch block, or after
			// the method returned the StringBuffer array.
			boolean closed = mock[Database.DB_01].returnPreparedStatement(pstmt) ;
			if (boudoir.getConfiguration().VERBOSE) System.out.println(leadingName + method + "closed pstmt? " + closed) ;
		}
	}

} // End of class