/**
* Title: 		 SimpleLogic (legacy name: ServiceLogic)
* Description:	 Sample class for the business logic of an application using the Boudoir library
* Copyright:     Copyright (c) 2013-2014
* Company:       Morpheus Quality Technologies
* Creation date: 03-12-2013
* Last modified: 11-02-2014
* @author 		 Stefano Adriani
* @version 		 1.0
*/

package adriani.boudoir.test ;

import adriani.boudoir.* ;
import adriani.boudoir.db.* ;
import adriani.boudoir.app.*;
import adriani.boudoir.logging.WSLogger ;
import adriani.boudoir.test.wrappers.Wrapper ;
import adriani.boudoir.test.wrappers.CustomerWrapper ;
import adriani.boudoir.test.wrappers.EntityTimeWrapper ;
import adriani.boudoir.util.* ;

import java.util.* ;

/**
* Sample class for the business logic of an application using the Boudoir library.<br/><br/>
* 
* This example mimics a class which is responsible for the high level business logic, which regards
* interface methods, execution flow and use cases. This class access methods of the DAO objects
* without worrying about backend details such as database connections, statement preparation 
* and SQL issues: all these details should be delegated to an object like the <code>SimpleMacro</code> class.
*/
public class SimpleLogic extends GeneralApplication {

  	/** A reference to the <code>ServiceMacro</code> managing the DAO layer.								*/
	protected 					SimpleMacro 	simpleMacro ;
	
  	/** Static value identifying the 1st application method.												*/
	public static final			int  			GET_CUSTOMER		= 1 ;
  	/** Static value identifying the 2nd application method.												*/
	public static final			int  			GET_TIME			= 2 ;

  	/** The static value used to identify the INFO empty response.											*/
	protected static final		int  			SILENT 				= 1 ;
  	/** The static value used to identify the INFO empty response.											*/
	protected static final		int  			INFO 				= 2 ;
	/** The static value used to identify the WARNING empty response.										*/
	protected static final		int  			WARN 				= 3 ;
	/** The static value used to identify the ERROR empty response.											*/
	protected static final		int  			ERROR 				= 4 ;
	
	/** Time (in milliseconds) used to stop execution during test or debug.									*/
	public static	 			int				DELAY 				= 0 ;			// Default value
	
	/** 
	* Index of the automatic unit-test to be executed.<br/>
	* The value of this index is stepped by the <code>SimpleClient</code> class during
	* execution of the automatic test trying all exit conditions, in order to test
	* data collected by the Analyzer feature. 
	*/
	public 						int  			TEST_CASE 			= NO_TEST ;
	
	/** Reference constant value used to trigger automatic unit-test.										*/
	public static final			int  			NO_TEST				= 0 ;
	/** Reference constant value used to trigger automatic unit-test.										*/
	public static final			int  			NO_RESULT			= 1 ;
	/** Reference constant value used to trigger automatic unit-test.										*/
	public static final			int  			LOGIC_ERROR			= 2 ;
	/** Reference constant value used to trigger automatic unit-test.										*/
	public static final			int  			FATAL				= 3 ;
	/** Reference constant value used to trigger automatic unit-test.										*/
	public static final			int  			VALIDATION_ERROR	= 4 ;
	/** Reference constant value used to trigger automatic unit-test.										*/
	public static final			int  			MACRO_ERROR			= 5 ;
	/** Reference constant value used to trigger automatic unit-test.										*/
	public static final			int  			MOCKED				= 6 ;

	/** Retrieving the logger instance to be used for this class.											*/
	protected 		 			WSLogger		logger ; 
	
  	/** Beginning of each logging message produced by this class.											*/
	protected static			String 			leadingName 	= "SimpleLogic - " ;

	/**
	* Constructor, creating also the <code>ServiceMacro</code> object and storing it as class attribute.
	* 
	* @param boudoir the <code>Boudoir</code> instance "owning" this object 
    */
    public SimpleLogic(Boudoir boudoir) {

    	super(boudoir) ;
		this.simpleMacro 		= new SimpleMacro(boudoir) ;
		this.logger				= boudoir.getLogger().getLogger(SimpleLogic.class) ;
		//System.out.println(leadingName + "Set boudoir instance: " + boudoir) ;
	}

 // *************************************************************************************************************
 // **********************************		FIRST SAMPLE METHOD		*********************************************
 // *************************************************************************************************************

	/**
	* Sample dummy method for the main business logic of an hypothetical operation, in this case:
	* searching a customer in the database.
	*
	* @param input an array of string parameters (in this case the main <code>args</code> array)
	* @return an example of dummy wrapper for the operation's results
    */
    public Wrapper getCustomerDummy(StringBuffer[] input) {
    	
		// 1. Starting call
		ArrayList<CustomerWrapper> list = null ;
		if (boudoir.getConfiguration().ANALYZE) boudoir.getAnalyzer().startCall() ;
		if (logger.isInfoEnabled()) 			logger.logInfo(leadingName + "1st input param: " + input) ;
		
		// 2. Checking for remote manager commands
		if ( boudoir.getConfiguration().MANAGER && (input != null) && (input.length > 0) ) {
			if ( manageCommand(input[0].toString()) ) return getEmptyResponse(GET_CUSTOMER, SILENT, ErrorMessage.COMMAND) ;
		}
		
		// 3. Input validation: trivial example, just converting from StringBuffer to String
		String[] params = null ; 
		if ( (input != null) & input.length > 0 ) {
			params = new String[input.length] ;
			for (int i = 0 ; i < input.length ; i++) params[i] = input[i].toString() ;
		} else {
			logger.logSevere(leadingName + "Wrong input data (null)") ;
			return getEmptyResponse(GET_CUSTOMER, SILENT, ErrorMessage.BAD_DATA) ;
		}

		// 4. Accessing DAO layer via the SimpleMacro class
		try {
			list = simpleMacro.getCustomerDummy(params) ;					
		} catch(Exception e) {	
			return handleMacroError(GET_CUSTOMER, e, true) ;	
		}
		
		// Analysis Test 1: forces a NO_RESULT ERROR (see 6.1)
		if (TEST_CASE == NO_RESULT) list = null ; 
		
		// 5. Handling critical cases (list == null)
		int  status =  simpleMacro.getStatus() ;
		if ( status == SimpleMacro.NO_DATABASE	) 		return getEmptyResponse(GET_CUSTOMER, SILENT, ErrorMessage.NO_DB) ;
		// Only for this application, the NO_DATA means the NO_CUSTOMER case, because we have just one query.
		// In actual usage we could NO_CUSTOMER when the customer is not found in the database, and NO_DATA when 
		// the customer is found, but he hasn't data in the database.
		if ( status == SimpleMacro.NO_DATA		) {
			if (boudoir.getConfiguration().ANALYZE) ((SimpleAnalyzer) boudoir.getAnalyzer()).addNoCustomer() ;
			return getEmptyResponse(GET_CUSTOMER, SILENT, ErrorMessage.NO_RESULT) ;
		}
		
		// 5.1 Logging list size
		if (boudoir.getConfiguration().VERBOSE) 		System.out.println(leadingName + "Retrieved list: " + list) ;
		
		// 6. Preparing output object
		CustomerWrapper output = null ;
		if (list != null && list.size() > 0) {
			CustomerWrapper[] record = (adriani.boudoir.test.wrappers.CustomerWrapper[] ) list.toArray( new CustomerWrapper[list.size()] 	) ;
			output = record[0] ;
			if (logger.isInfoEnabled()) logger.logInfo(leadingName + "Retrieved data from database") ;
			//System.out.println(leadingName + "Set list size: " + list.size()) ;
		} else {
			// 6.1 This case should never happen, because if "list" is null or empty we should have already
			// handled it above (with the NO_DATA case) -> we log the situation with warning level.
			return getEmptyResponse(GET_CUSTOMER, WARN, ErrorMessage.NO_RESULT) ;
		}
		
		// Analysis Test 2: forces a LOGIC ERROR
		if (TEST_CASE == LOGIC_ERROR) output = null ;
		
		// 7. Adjusting output element
		if (output != null) {
			if (status == SimpleMacro.OK) {
				// Placeholder example: to be used if output object carries status information
				//output.setStatus(STATUS_OK) ;
				if (logger.isInfoEnabled()) logger.logInfo(leadingName + "Operation succesful") ;
			} else {
				// Placeholder example: to be used if output object carries status information				
				//output.setStatus(STATUS_WRONG) ;
				if (logger.isInfoEnabled()) logger.logInfo(leadingName + "Operation failed") ;
			}
			if (logger.isInfoEnabled()) logger.logInfo(leadingName + "Validating output data"	) ;
			
			try {	
				output.validate() ;													
			} catch(Exception e) {	
				return handleNoValidation(output, e) ;								
			} catch(Throwable t){
				if (boudoir.getConfiguration().DEBUG) t.printStackTrace() ;
				return getEmptyResponse(GET_CUSTOMER, ERROR, ErrorMessage.LOGIC) ;		
			}			
		}
		
		// 8. If we reach this block there's something wrong in the above logic
		if (output == null) {
			logger.logSevere(leadingName + "Found null output list") ;
			return getEmptyResponse(GET_CUSTOMER, ERROR, ErrorMessage.LOGIC) ;
		}

		// Analysis Test 3: forces an unexpected case (not handled by design)
		if (TEST_CASE == FATAL)	{	Object foo = null ;		foo.toString() ;	}

		// 9. Stopping the chronwatch and updating analysis counters
		if (boudoir.getConfiguration().ANALYZE) boudoir.getAnalyzer().endCall(ErrorMessage.OK) ;

		// 10. Returning result
		return output ;
    }
    
// *************************************************************************************************************
// **********************************		SECOND SAMPLE METHOD	********************************************
// *************************************************************************************************************
    
	/**
	* Sample dummy method for the main business logic of an hypothetical operation, in this case:
	* searching a customer in the database.
	*
	* @param input an array of string parameters (in this case the main <code>args</code> array)
	* @return an example of dummy wrapper for the operation's results
    */
    public Wrapper getEntityTimesDummy(StringBuffer[] input) {
    	
		// 1. Starting call
		ArrayList<EntityTimeWrapper> list = null ;
		if (boudoir.getConfiguration().ANALYZE) boudoir.getAnalyzer().startCall() ;
		if (logger.isInfoEnabled()) 			logger.logInfo(leadingName + "1st input param: " + input) ;
		
		// 2. Checking for remote manager commands
		// We don't want to bound the Remote Manager system to this method
		
		// 3. Input validation: checking that input string is a valid integer
		int param ; 
		if ( (input != null) & input.length > 0 ) {
			param = valueOfString(input[0].toString(), -1) ;
		} else {
			logger.logSevere(leadingName + "Wrong input data (not a number)") ;
			return getEmptyResponse(GET_TIME, SILENT, ErrorMessage.BAD_DATA) ;
		}

		// 4. Accessing DAO layer via the SimpleMacro class
		try {
			list = simpleMacro.getEntityTimesDummy(param) ;					
		} catch(Exception e) {	
			return handleMacroError(GET_TIME, e, true) ;	
		}
		
		// Analysis Test 1: forces a NO_RESULT ERROR (see 6.1)
		// We don't want to allow an analysis test for this method  
		
		// 5. Handling critical cases (list == null)
		int  status =  simpleMacro.getStatus() ;
		if ( status == SimpleMacro.NO_DATABASE	) 		return getEmptyResponse(GET_TIME, SILENT, ErrorMessage.NO_DB) ;
		// Only for this application, the NO_DATA means the NO_RESULT case, because we have just one query.
		// In actual usage we could NO_RESULT when the record is not found in the database, and NO_DATA when 
		// the main record is found, but he hasn't data in the database (e.g. in other tables).
		if ( status == SimpleMacro.NO_DATA		) {
			if (boudoir.getConfiguration().ANALYZE) ((SimpleAnalyzer) boudoir.getAnalyzer()).addNoCustomer() ;
			return getEmptyResponse(GET_TIME, SILENT, ErrorMessage.NO_RESULT) ;
		}
		
		// 5.1 Logging list size
		if (boudoir.getConfiguration().VERBOSE) 		System.out.println(leadingName + "Retrieved list: " + list) ;
		
		// 6. Preparing output object
		EntityTimeWrapper output = null ;
		if (list != null && list.size() > 0) {
			EntityTimeWrapper[] record = (adriani.boudoir.test.wrappers.EntityTimeWrapper[] ) list.toArray( new EntityTimeWrapper[list.size()] 	) ;
			output = record[0] ;
			if (logger.isInfoEnabled()) logger.logInfo(leadingName + "Retrieved data from database") ;
			//System.out.println(leadingName + "Set list size: " + list.size()) ;
		} else {
			// 6.1 This case should never happen, because if "list" is null or empty we should have already
			// handled it above (with the NO_DATA case) -> we log the situation with warning level.
			return getEmptyResponse(GET_TIME, WARN, ErrorMessage.NO_RESULT) ;
		}
		
		// Analysis Test 2: forces a LOGIC ERROR
		if (TEST_CASE == LOGIC_ERROR) output = null ;
		
		// 7. Adjusting output element
		if (output != null) {
			if (status == SimpleMacro.OK) {
				// Placeholder example: to be used if output object carries status information
				//output.setStatus(STATUS_OK) ;
				if (logger.isInfoEnabled()) logger.logInfo(leadingName + "Operation succesful") ;
			} else {
				// Placeholder example: to be used if output object carries status information				
				//output.setStatus(STATUS_WRONG) ;
				if (logger.isInfoEnabled()) logger.logInfo(leadingName + "Operation failed") ;
			}
			if (logger.isInfoEnabled()) logger.logInfo(leadingName + "Validating output data"	) ;
			
			try {	
				output.validate() ;													
			} catch(Exception e) {	
				return handleNoValidation(output, e) ;								
			} catch(Throwable t){
				if (boudoir.getConfiguration().DEBUG) t.printStackTrace() ;
				return getEmptyResponse(GET_TIME, ERROR, ErrorMessage.LOGIC) ;		
			}			
		}
		
		// 8. If we reach this block there's something wrong in the above logic
		if (output == null) {
			logger.logSevere(leadingName + "Found null output list") ;
			return getEmptyResponse(GET_TIME, ERROR, ErrorMessage.LOGIC) ;
		}

		// Analysis Test 3: forces an unexpected case (not handled by design)
		if (TEST_CASE == FATAL)	{	Object foo = null ;		foo.toString() ;	}

		// 9. Stopping the chronwatch and updating analysis counters
		if (boudoir.getConfiguration().ANALYZE) boudoir.getAnalyzer().endCall(ErrorMessage.OK) ;

		// 10. Returning result
		return output ;
    }

// *************************************************************************************************************
// **********************************		WRAPPERS FACTORY		********************************************
// *************************************************************************************************************
    
	/**
	* The <strong>factory</strong> creating the proper <code>XXXWrapper</code> object that will be returned
	* by the corresponding <code>getXXXDummy()</code> method. Every possible <code>XXXWrapper</code> class
	* must extend the parent <code>Wrapper</code> class.<br/>
	* This factory should return the proper <code>XXXWrapper</code> object corresponding to the input
	* <em>method</em> parameter, which describes one of the operations offered by the simple logic class.
	*
	* In current implementation this factory should be accessed only the the <code>getEmptyResponse()</code>
	* method.
	*
	* @param method the logic method that invoked this method (see class fields)
	* @return the proper <code>Wrapper</code> subclass for the corresponding logic method
    */
    public Wrapper getWrapper(int method) {

		// 1. Preparing the proper output object
		Wrapper wrapper = null ;
		switch(method) {
			case	GET_CUSTOMER	:	wrapper = new CustomerWrapper() ;			break ;
			case	GET_TIME		:	wrapper = new EntityTimeWrapper() ;			break ;
		}

		// 2. Eventually logging creation failure
		if (wrapper == null) {
			logger.logSevere(leadingName + "WrapperOut factory failed for method " + method) ;
		}

		return wrapper ;
	}
    
// *************************************************************************************************************
// **********************************		ERRORS MANAGEMENT		********************************************
// *************************************************************************************************************

	/**
	* Provides a valid <code>Object</code> output when the exception comes from a lower level, like the
	* <code>SimpleMacro</code> class or the DAO layer. In the last case the error message will be
	* obtained by reading the necessary <code>StackTraceElement</code> objects.<br/><br/>
	*
	* This method could be necessary when the application requires a formally corrected output in any scenario.
	* When the <code>SimpleMacro</code> is interrupted before producing any output, this method could be
	* invoked to produce the proper output object. If such object contains a reference to the execution  
	* error, such error will be something like "Error at macro level".
	*
	* @param method the logic method that invoked this method (see class fields)
	* @param exception the <code>Exception</code> thrown from a lower level (<code>null</code> otherwise)
	* @param last if <code>true</code> the <code>getEmptyResponse()</code> method is invoked
	* @return an example of dummy wrapper for the operation's results
    */
    protected Wrapper handleMacroError(int method, Exception exception, boolean last) {

		// 1. Closing connections to databases (we don't know if they have been left opened)
    	// Usually they should have already been closed by the ServiceMacro.handleNoDatabase() method
    	if (boudoir.getConfiguration().DEBUG) System.out.println(leadingName + "Handling macro error") ;
		SimpleDAO simpleDAO = simpleMacro.getServiceDAO() ;
		if (simpleDAO != null) {
			simpleDAO.closeConnection(Database.DB_01) ;
			// Placeholder: we could have more than one database
			//serviceDAO.closeConnection(Database.DB_02) ;
			logger.logWarning(leadingName + "Forced connections check in for all databases") ;
		}

		// 2. Retrieving exception message
		String error = null ;
		if (exception != null) try {
			exception.printStackTrace() ;
			StackTraceElement[] stackTrace = exception.getStackTrace() ;
			error = stackTrace[0].getClassName() + "." + stackTrace[0].getMethodName() + " " + exception ;
		}
		catch(Exception e) {
			e.printStackTrace() ;
			error = "STACKTRACE ERROR (" + e.getMessage() + ") for " + exception.getMessage() ;
		}

		// 3. Logging error message
		String message = boudoir.getMessages().getMessage(ErrorMessage.MACRO) ;
		logger.logSevere(leadingName + message + ": " + error) ;

		// 4. If this is the last macro operation, we create the output object in order to provedi a
		// complete response for the main method (the main logic will quit processing).
		if (last) 	return getEmptyResponse(method, SILENT, ErrorMessage.MACRO) ;
		else 		return null ;
	}

	/**
	* Provides a valid <code>Object</code> output when the output object requires some formal validation.
	* For instance, this could happen when the output object is a SOAP element or a DTO requiring a
	* consistency check.<br/><br/>
	*
	* This method could be necessary when the application requires a formally corrected output in any scenario.
	* When the <code>SimpleLogic</code> is interrupted because we found an invalid output, this method could be
	* invoked to produce the proper output object. If such object contains a reference to the execution  
	* error, such error will be something like "Validation error".
	*
	* @param output the <code>Object</code> that failed validation (could need to be adjusted)
	* @param exception the <code>Exception</code> thrown from a lower level (<code>null</code> otherwise)
	* @return an example of dummy wrapper for the operation's results
    */
    protected Wrapper handleNoValidation(Wrapper output, Exception e) {

		// Logging only the validation error message due to the <code>Wrapper</code> inner wrappers,
		// because any other error (if any) has already been logged.
		String invalid = boudoir.getMessages().getMessage(ErrorMessage.VALIDATION) ;
		String message = invalid + ": " + e.getMessage() ;
		logger.logSevere(leadingName + message) ;

		// Placeholder example: to be used if output object carries status information
		//output.setStatus(VALIDATION_ERROR) ;

		// Eventually ending analysis
		if (boudoir.getConfiguration().ANALYZE) boudoir.getAnalyzer().endCall(ErrorMessage.VALIDATION) ;
		return output ;
	}

	/**
	* Prepare a valid output <code>object</code> when it's not possible to produce a valid result.<br/>
	* This method can be called in many different situations, each situation corresponding to a different
	* message to be logged while producing the empty response. These situations are identified using the
	* <code>state</code> param, which should be one of the integer values mapped by the
	* <code>ErrorMessage</code> class. In current implementation this method is called in following
	* situations: NO_DB, NO_RESULT, ERROR_LOGIC, ERROR_MACRO.<br/><br/>
	*
	* The <code>level</code> parameter can be: SILENT, INFO, WARN or ERROR.<br/>
	* If the SILENT constant it's used when no log message must be produced here, because we assume the caller
	* has already logged a message describing the situation. Otherwise we produce a log message with
	* the specified level.
	*
	* @param method a reference to one the SOAP methods exposed in the WSDL (see class fields)
	* @param level the integer value defining the level of the logged message
	* @param state the code of the message as described by the <code>ErrorMessage</code> class
	* @return an example of dummy wrapper for the operation's results
	* @see #getWrapperOut(int method)
    */
    protected Wrapper getEmptyResponse(int method, int level, int state) {

		// 1. Ending analysis and getting error message
		if (boudoir.getConfiguration().ANALYZE) boudoir.getAnalyzer().endCall(state) ;
		String message = boudoir.getMessages().getMessage(state) ;

		// 2. Eventually logging the error message
		switch(level) {
			case 	SILENT	:	break ;
			case 	INFO	:	if (logger.isInfoEnabled()) logger.logInfo(leadingName + message) ;		break ;
			case 	WARN	:	logger.logWarning(leadingName + message) ;								break ;
			case 	ERROR	:	logger.logSevere(leadingName + message) ;								break ;
			default			:	if (logger.isInfoEnabled()) logger.logInfo(leadingName + "EMPTY RESPONSE") ;
		}
		if (boudoir.getConfiguration().DEBUG) System.out.println(leadingName + "Producing empty response (" + message + ")") ; 

		// 3. Producing a default response (without data)
		Wrapper output = getWrapper(method) ;

		// 4. Returning output wrapper
		return output ;
	}
    
// *************************************************************************************************************
// **********************************			ACCESSORS			********************************************
// *************************************************************************************************************
    
	/** Getter method																						   */
    public SimpleMacro getSimpleMacro() 			{			return this.simpleMacro ;					   }

} // End of class