/**
* Title: 		 SimpleMacro (legacy name: ServiceMacro)
* Description:	 Sample layer between the SimpleLogic and the DAO layer
* Copyright:     Copyright (c) 2013-2014
* Company:       Morpheus Quality Technologies
* Creation date: 03-12-2013
* Last modified: 07-01-2014
* @author 		 Stefano Adriani
* @version		 1.0
*/

package adriani.boudoir.test ;

import adriani.boudoir.* ;
import adriani.boudoir.db.* ;
import adriani.boudoir.app.* ;
import adriani.boudoir.test.model.CustomerDTO ;
import adriani.boudoir.test.model.EntityTimeDTO ;
import adriani.boudoir.test.wrappers.CustomerWrapper ;
import adriani.boudoir.test.wrappers.EntityTimeWrapper ;
import adriani.boudoir.util.* ;
import adriani.boudoir.logging.WSLogger ;

import java.util.* ;

/**
* Sample class for the layer between the main business logic of the application and the DAO layer.<br/><br/> 
*
* A macro class should take care of preparing all the calls to the necessary DAO.<br/>
* The "interface policy" of such class could be the following:<br/><br/>
*
* EXCEPTION   : an unexpected exception is thrown, hence nothing is returned<br/>
* DATA OK     : the method return an <code>ArrayList</code> containing a set of output <code>Object</code><br/>
* NO DATABASE : the method return a <code>null</code> object (database not available)<br/>
* NO DATA     : the method return an <code>ArrayList</code> containing a set of output <code>Object</code> (could be empty)<br/>
*
* The specific case can be acknowledged by calling the <code>getStatus()</code> method after execution.<br/><br/>
*
* Implementation guidelines:<br/><br/>
*
* To grant loose coupling with the application interface and the database, all the database
* issues should be handled by the <code>SimpleDAO</code> class, and all the interface issues
* should be handled by a custom modeling package (not provided here since since it strongly depends
* by the application using the library). When using an actual modeling package, all the reference to 
* the <code>StringBuffer</code> objects should be substituted with the proper DTO objects. 
*/
public class SimpleMacro extends GeneralApplication {

	/** A reference to the <code>ServiceDAO</code> class accessing the necessary database's.				*/
	protected 				SimpleDAO 		simpleDAO ;

  	/** An integer code describing the status of this macro after the last operation.						*/
	protected 				int 			status ;

  	/** A static code describing the status of this macro after the last operation.							*/
	protected static final 	int  			OK	 				= 1 ;
  	/** A static code describing the status of this macro after the last operation.							*/
	protected static final 	int  			NO_DATA				= 2 ;
  	/** A static code describing the status of this macro after the last operation.							*/
	protected static final 	int  			NO_DATABASE			= 3 ;
	
	/** 
	* Index of the automatic unit-test to be executed.<br/>
	* The value of this index is stepped by the <code>SimpleClient</code> class during
	* execution of the automatic test trying all exit conditions, in order to test
	* data collected by the Analyzer feature. 
	*/
	public 					int  			TEST_CASE 			= SimpleLogic.NO_TEST ;

	/** Retrieving the logger instance to be used for this class.											*/
	protected 		 		WSLogger		logger 				= boudoir.getLogger().getLogger(SimpleMacro.class) ;
	
  	/** Beginning of each logging message produced by this class.											*/
	protected static		String 		 	leadingName 		= "SimpleMacro - " ;

	/**
	* Standard constructor, creating also the <code>ServiceDAO</code> object and storing it as class attribute.
	* 
	* @param boudoir the <code>Boudoir</code> instance "owning" this object 
    */
    public SimpleMacro(Boudoir boudoir) {

    	super(boudoir) ;
		this.simpleDAO 			= new SimpleDAO(this) ;
	}

	/**
	* First example method of macro implementation.<br/><br/> 
	* 
	* <strong>Remark</strong>: if this method opens a connection to a database, it's strongly suggested to make 
	* sure to not return from this method without closing the opened connection!
	* 
	* @param input the input parameters as passed by the logic layer
	* @return an array list containing all the found records (wrapped by a proper wrapper)
    */
    public ArrayList<CustomerWrapper> getCustomerDummy(String[] input) {

		// 1. Opening connections to the database
		if (! simpleDAO.openConnection(Database.DB_01) ) return handleNoDatabase(Database.DB_01) ;
		
		// Mocked connection: allows automatic test of the Mocked Connections feature
		if (TEST_CASE == SimpleLogic.MOCKED) System.out.println(leadingName + " !!! " + simpleDAO.testPreparedStatement(1)) ;

		// 2. Preparing call environment
		this.status = NO_DATA ;
		ArrayList<CustomerWrapper> list = null ;
		if (logger.isInfoEnabled()) logger.logInfo(leadingName + "Entering Macro for: " + input[0] ) ;
		System.out.println (leadingName + "Entering Macro for: " + input[0] ) ;

		// 3. Retrieving records from the database
		CustomerDTO[]	records 	= simpleDAO.getCustomerData(input[0], null) ;

		// 4. PLACEHOLDER: if necessary write some "macro logic" here

		// 5. Preparing return object
		if (records != null) {
			
			// Analysis Test 4: forces a VALIDATION ERROR
			if (TEST_CASE == SimpleLogic.VALIDATION_ERROR) records[0].setName( new StringBuffer("42") ) ;
			
			list = new ArrayList<CustomerWrapper>() ;
			
			// Complex nested output wrapper
			for (int n = 0 ; n < records.length ; n++) {
				CustomerWrapper output = new CustomerWrapper(boudoir) ;
				output.setName		( records[n].getName()	) ;
				output.setSurname	( records[n].getSurname()	) ;				
				list.add(output) ;
				System.out.println(leadingName + "Retrieved data: " + records[n].getName() + "-" + records[n].getSurname()) ;
			}
			
			if (logger.isInfoEnabled()) logger.logInfo(leadingName + "Fetched: " + list.size() + " records from database") ;			
			this.status = OK ;
		} else {
			this.status = NO_DATA ;
		}
		// If necessary we can quit execution of this method
		//else return handleNoRecord(input, NO_DATA) ;
		
		// 6. If necessary stopping current thread to easy concurrent benchmark test
		if (SimpleLogic.DELAY > 0) {
			if (logger.isInfoEnabled()) logger.logInfo(leadingName + "Stopping current thread for " + SimpleLogic.DELAY + " ms") ;
			delay(SimpleLogic.DELAY) ; 
		}
		
		// Analysis Test 5: forces an ERROR AT MACRO LEVEL
		if (TEST_CASE == SimpleLogic.MACRO_ERROR) 		{	Object foo = null ;		foo.toString() ;	}

		// 7. Closing connections to databases
		simpleDAO.closeConnection(Database.DB_01) ;

		// 8. Returning result
		return list ;
	}
    
	/**
	* Second example method of macro implementation.<br/><br/> 
	* 
	* <strong>Remark</strong>: if this method opens a connection to a database, it's strongly suggested to make 
	* sure to not return from this method without closing the opened connection!
	* 
	* @param input the integer input parameter as passed by the logic layer
	* @return an array list containing all the found records (wrapped by a proper wrapper) 
    */
    public ArrayList<EntityTimeWrapper> getEntityTimesDummy(int input) {

		// 1. Opening connections to the database
		if (! simpleDAO.openConnection(Database.DB_01) ) return handleNoDatabase(Database.DB_01) ;
		
		// Mocked connection: allows automatic test of the Mocked Connections feature
		if (TEST_CASE == SimpleLogic.MOCKED) System.out.println(leadingName + " !!! " + simpleDAO.testPreparedStatement(1)) ;

		// 2. Preparing call environment
		this.status = NO_DATA ;
		ArrayList<EntityTimeWrapper> 	list = null ;
		if (logger.isInfoEnabled()) 	logger.logInfo(leadingName + "Entering Macro for: " + input ) ;

		// 3. Retrieving records from the database
		EntityTimeDTO	record		 	= simpleDAO.getEntityTime(input) ;

		// 4. PLACEHOLDER: if necessary write some "macro logic" here

		// 5. Preparing return object
		if (record != null) {
			
			// Analysis Test 4: forces a VALIDATION ERROR
			// No validation required for this method
			
			list = new ArrayList<EntityTimeWrapper>() ;
			
			// Complex nested output wrapper
			EntityTimeWrapper output = new EntityTimeWrapper(boudoir) ;
			output.setTime		( record.getEndTime()	) ;
			list.add(output) ;
			System.out.println(leadingName + "Retrieved data: " + record.getEndTime()) ;
			
			if (logger.isInfoEnabled()) logger.logInfo(leadingName + "Fetched: " + list.size() + " record from database") ;			
			this.status = OK ;
		} else {
			this.status = NO_DATA ;
		}
		// If necessary we can quit execution of this method
		//else return handleNoRecord(input, NO_DATA) ;
		
		// Analysis Test 5: forces an ERROR AT MACRO LEVEL
		// No unit-testing required for this method

		// 7. Closing connections to databases
		simpleDAO.closeConnection(Database.DB_01) ;

		// 8. Returning result
		return list ;
	}
    
// *************************************************************************************************************
// **********************************		ERRORS MANAGEMENT		********************************************
// *************************************************************************************************************

	/**
	* Prepare a null <code>ArrayList</code> object when the service logic fails because there is no
	* connection to one of the required databases. The output message will contain a message like
	* "No connection to the database".
	*
	* @param database a reference to the code number of the missing database
	* @return a null <code>ArrayList</code> object with a message describing this situation
	* @see Database
    */
    protected ArrayList handleNoDatabase(int database) {

		// Trying to close connections (we can have more than a database)
		simpleDAO.closeConnection(Database.DB_01) ;
		if (boudoir.getConfiguration().DEBUG) System.out.println(leadingName + "Forced connections check in for all databases") ;

		// Writing error on the log system
		this.status = NO_DATABASE ;
		String message = boudoir.getMessages().getMessage(ErrorMessage.NO_DB) + " (" + database + ')' ;
		logger.logSevere(leadingName + message) ;
		return null ;
	}

// *************************************************************************************************************
// **********************************			MUTATORS			********************************************
// *************************************************************************************************************

	/**
	* Accessor method.
	*/
	public void setStatus(int value)				{	this.status = value ;									}

// *************************************************************************************************************
// **********************************			ACCESSORS			********************************************
// *************************************************************************************************************

	/**
	* Accessor method.
	*/
	public int getStatus() 							{	return this.status ;									}

	/**
	* Accessor method.
	*/
	public SimpleDAO getServiceDAO() 				{	return this.simpleDAO ;								}

 } // End of class